#include <mex.h>

/*
 * function [e] = addpathsf(V, R, Vi, tol)
 *
 * Choose columns of Vi to extend V and update the Cholesky factor R.
 * (Single precision block packed storage)
 */


/*
 * SUBROUTINE TRTRS( UPLO, TRANS, DIAG, N, NRHS, A, LDA,  B,
 *                   LDB, INFO )
 */
void strtrs_(char* uplo, char* trans, char* diag,
             int* n, int* nrhs,
             float* a, int* lda,
             float* b, int* ldb,
             int* info);

/*
 * SUBROUTINE GEMM ( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA,
 *                   B, LDB, BETA, C, LDC )
 */
void sgemm_(char* transA, char* transB, 
            int* m, int* n, int* k, float* alpha,
            float* A, int* ldA,
            float* B, int* ldB, float* beta, 
            float* C, int* ldC);

/*
 * SUBROUTINE GEQP3( M, N, A, LDA, JPVT, TAU,  WORK,  LWORK, INFO )
 *  -- QR with column pivoting.  Level 3 BLAS.
 */
void sgeqp3_(int* m, int* n, float* a, int* lda, int* jpvt, float* tau,
             float* work, int* lwork, int* info);

/*
 * SUBROUTINE POTRF( UPLO, N, A, LDA, INFO )
 */
void spotrf_(char* uplo, int* n, float* A, int* ldA, int* info);


/* Compute overlap for two path vectors
 *  p1 -- l1 edge numbers in sorted order
 *  p2 -- l2 edge numbers in sorted order
 */
int path_overlap(int* p1, int l1, 
                 int* p2, int l2)
{
    int i1      = 0;
    int i2      = 0;
    int overlap = 0;

    while (i1 < l1 && i2 < l2) {
        if (p1[i1] < p2[i2]) {
            ++i1;
        } else if (p1[i1] > p2[i2]) {
            ++i2;
        } else {
            ++overlap;
            ++i1;
            ++i2;
        }
    }

    return overlap;
}


/* Compute overlap contributions:
 *   R(1:k,     k+1:k+m) = V' *Vi
 *   R(k+1:k+k, k+1:k+m) = Vi'*Vi
 */
void compute_overlaps(float* R12, float* R22, 
                      int k, int* irV,  int* jcV,
                      int m, int* irVi, int* jcVi)
{
    int i, j;
    
    for (j = 0; j < m; ++j) {
        for (i = 0; i < k; ++i) {
            R12[j*k+i] = path_overlap(irV  + jcV [i], jcV [i+1] - jcV [i],
                                      irVi + jcVi[j], jcVi[j+1] - jcVi[j]);
        }
        for (i = 0; i < m; ++i) {
            R22[j*m+i] = path_overlap(irVi + jcVi[i], jcVi[i+1] - jcVi[i],
                                      irVi + jcVi[j], jcVi[j+1] - jcVi[j]);
        }
    }
}


/*
 * Blocked backsolve with R^T
 */
void block_trs(float* R, int k, int kb, float* b, int m)
{
    int n, J, info;
    float one     =  1;
    float neg_one = -1;

    for (J = 0; J <= (k-1)/kb; ++J) {

        int Jkb    = J*kb;
        int Jkbp   = Jkb + kb;
        int blockn = (Jkbp < k) ? kb : (k - Jkb);

        sgemm_("T", "N", &blockn, &m, &Jkb,
               &neg_one,
               R    + (Jkb*Jkbp)/2,  &Jkbp,
               b,                    &k,
               &one,
               b    + Jkb,           &k);
        
        strtrs_("U", "T", "N", &blockn, &m,
                R    + Jkb + (Jkb*Jkbp)/2,  &Jkbp,
                b    + Jkb,                 &k,
                &info);
    }
}


/* Choose which vectors we should keep via QR with column pivots
 */
int add_paths(int* jpvt, float* R,
              int k, int* irV,  int* jcV,
              int m, int* irVi, int* jcVi,
              float tol)
{

    int i, j, J, nchosen, lwork, info;
    float twork;
    float* work;
    float* scratch;
    float* R12t;
    float* R22t;
    float* a;
    float neg_one = -1;
    float one     =  1;

    int kb = 100;


    /* 
     * Workspace query and allocations:
     *   k * m     -- scratch for R12
     *   m * m     -- scratch for R22
     *   m * m     -- scratch for R22 factorization
     *   lwork + m -- scratch for pivoted QR
     */
    lwork = -1;
    sgeqp3_(&m, &m, NULL, &m, NULL, NULL, &twork, &lwork, &info);
    lwork = (int) twork;

    scratch = mxMalloc((lwork + m + (k+m+m)*m) * sizeof(float));
    R12t    = scratch;
    R22t    = R12t + k*m;
    a       = R22t + m*m;
    work    = a    + m*m;

    /* Extend R with columns for V'*Vi and Vi'*Vi
     */
    compute_overlaps(R12t, R22t,
                     k, irV,  jcV,
                     m, irVi, jcVi);

    /* Do the first step of Cholesky updating:
     *   R12 := R11^-T R12
     *   R22 := R22 - R12'*R12
     */
    if (k != 0) {

        block_trs(R, k, kb, R12t, m);

        sgemm_("T", "N", &m, &m, &k, 
               &neg_one,
               R12t, &k,
               R12t, &k, 
               &one,
               R22t, &m);

    }

    /* Pivoted QR factorization on copy of R22 */
    memset(jpvt, 0, m  *sizeof(int));
    memcpy(a, R22t, m*m*sizeof(float));
    sgeqp3_(&m, &m, a, &m, jpvt, work, work + m, &lwork, &info);

    /* Mark selected columns */
    nchosen = 0;
    for (i = 0; i < m; ++i) {
        if (a[i*(m+1)] > tol || a[i*(m+1)] < -tol) {
            jpvt[nchosen++] = jpvt[i]-1;
        }
    }

    /* Subset chosen rows and columns of R22 into a */
    for (i = 0; i < nchosen; ++i)
        for (j = 0; j < nchosen; ++j)
            a[j*m + i] = R22t[jpvt[j]*m + jpvt[i]];

    /* Run Cholesky on the trailing nchosen-by-nchosen block */
    spotrf_("U", &nchosen, a, &m, &info);

    /* Copy R12 and R22 blocks into place */
    for (j = 0; j < nchosen; ++j) {

        int J    = (k+j)/kb;
        int Jkb  = J*kb;
        int Jkbp = Jkb + kb;

        for (i = 0; i < k; ++i)
            R[i   + ((2*(k+j) - Jkb)*Jkbp) / 2] = R12t[jpvt[j]*k + i];
        for (i = 0; i <= j; ++i)
            R[k+i + ((2*(k+j) - Jkb)*Jkbp) / 2] = a[j*m + i];

    }

    mxFree(scratch);
    return nchosen;
}


/*
 * function [e] = addpaths(V, R, Vi, tol)
 *
 * Choose columns of Vi to extend V and update the Cholesky factor R.
 */
void mexFunction(int nlhs, mxArray** plhs,
                 int nrhs, const mxArray** prhs)
{
    int     i, k, m, nchosen, ldR;
    int*    jpvt;
    int*    irV;
    int*    jcV;
    int*    irVi;
    int*    jcVi;
    float*  R;
    float   tol;

    if (nrhs != 4)
        mexErrMsgTxt("Wrong number of arguments");

    if (!mxIsSparse(prhs[0]) && mxGetN(prhs[0]) != 0)
        mexErrMsgTxt("U must be sparse");
    if (!mxIsNumeric(prhs[1]))
        mexErrMsgTxt("R must be dense");
    if (!mxIsSparse(prhs[2]))
        mexErrMsgTxt("V must be sparse");

    k    = mxGetN (prhs[0]);
    irV  = mxGetIr(prhs[0]);
    jcV  = mxGetJc(prhs[0]);
    
    R    = (float*) mxGetPr(prhs[1]);

    m    = mxGetN (prhs[2]);
    irVi = mxGetIr(prhs[2]);
    jcVi = mxGetJc(prhs[2]);
    
    tol  = (float) *mxGetPr(prhs[3]);

    jpvt = (int*) mxMalloc(m * sizeof(int));
    nchosen = add_paths(jpvt, R, k, irV,  jcV, m, irVi, jcVi, tol);

    plhs[0] = mxCreateDoubleMatrix(nchosen,1, mxREAL);
    for (i = 0; i < nchosen; ++i) {
        mxGetPr(plhs[0])[i] = jpvt[i]+1;
    }
    mxFree(jpvt);
}
