% [V, Vinfo, k, percent, tall] = scanpaths(n, adj, A, ne, verb);
%
% Try to compute an orthonormal basis V for the range space of
% all pairs of shortest-path vectors between the nodes in n.
% Return also the rank k of the generated basis.
%
% Inputs:
%   n    - list of nodes used
%   adj  - adjacency matrix of edge weights
%   A    - adjacency matrix of edge indices
%   ne   - number of edges (compute from A if not provided)
%   verb - verbosity flag (default is 1, verbose)
%
% Outputs:
%   V       - reduced path matrix (transposed)
%   Vinfo   - [source; destination; distance] for columns of V
%   k       - rank (number of columns in V)
%   percent - percentage of total paths kept
%   tall    - total time

function [V, Vinfo, k, percent, tall] = scanpaths(n, adj, A, ne, verb);

if nargin < 4,   ne   = [];         end
if nargin < 5,   verb = 1;          end
if isempty(ne),  ne = max(max(A));  end

m     = length(n);
V     = [];
Vinfo = [];
nconsidered = 0;

if verb, fprintf('i:\tk\t%%\tt_step\tt_all\n'); end

t0 = clock;
for i = 1:m-1
  t1 = clock;

  Vi     = [];
  Vinfoi = [];

  % Compute paths to extend basis
  [d, p] = dijkstra(adj, n(i));
  for j = i+1:m
    Vi     = [Vi, pathv(n(j), p, A, ne)];
    Vinfoi = [Vinfoi, [i; j; d(n(j))]];
  end
  nconsidered = nconsidered + m-i;

  % Compute index vector e of new independent paths:
  %  - Gram-Schmidt to eliminate V directions from Vi
  %  - Rank revealing QR on remainder
  %
  if isempty(V)
    Vp    = full(Vi'*Vi);
    e     = qrchoose(Vp, 1e-8);
  else
    P     = colamd(V);
    [C,R] = qr(V(:,P), Vi, 0);
    Vp    = full(Vi'*Vi - C'*C); 
    e     = qrchoose(Vp, 1e-5);
  end

  % Update path record
  V     = [ V,     Vi(:,e)     ];
  Vinfo = [ Vinfo, Vinfoi(:,e) ];

  % Print current info
  if verb
    fprintf('\r                                           ');
    fprintf('\r%d:\t%d\t%3.2f\t%3.2f\t%g', i, size(V,2), ...
            100*size(V,2) / nconsidered, ...
            etime(clock,t1), etime(clock,t0));
  end

end

% Convert local node labels to real labels
Vinfo(1:2,:) = n(Vinfo(1:2,:));

% Compute statistics
percent = 100*size(V,2) / nconsidered;
k       = size(V,2);
tall    = etime(clock,t0);

if verb, fprintf('\n'); end
