 /**
 *OBJECT:	This class provides methods to invoke every
 *		all-pairs shortest path algorithm implemented.
 *
 * @version    $Id: APSP.java,v 1.1 12/26/2001
 *
 * @author    Mark Rabin
 *
 * Revisions:
 *     $Log: APSP.java,v 1.1$
 *     Revision 1.1  8/20/2002
 *     Initial revision
 *
 */


// package LossyLink;


public class APSP {
	/**
	 * Run Dijkstra algorithm using a linear array priority queue.
	 *
   *
   * @param testGraph the graph in adjacency list form
	 * @param source the source node
	 * @return a ShortestPath object conatining the solved shortest paths
	 */
    public static ShortestPath linearDijkstra(  Node [] graph, int source ) {
        return new LinearDijkstra( graph, source, null, null );
    }
    public static ShortestPath linearDijkstra(  Node [] graph, int source, LossyLinkPathInfo pathInfo, LossyLinkData data ) {
    	return new LinearDijkstra( graph, source, pathInfo, data );
    }

	/**
	 * Run Dijkstra algorithm using a binary heap priority queue.
	 *
   *
   * @param testGraph the graph in adjacency list form
	 * @param source the source node
	 * @return a ShortestPath object conatining the solved shortest paths
	 */
    public static ShortestPath binaryDijkstra(  Node [] graph, int source ) {
    	return new BinaryDijkstra( graph, source );
    }

   /**
	 * Run Dijkstra algorithm using a fibonacci heap priority queue.
	 *
   *
   * @param testGraph the graph in adjacency list form
	 * @param source the source node
	 * @return a ShortestPath object conatining the solved shortest paths
	 */
    public static ShortestPath fibDijkstra(  Node [] graph, int source ) {
    	return new FibDijkstra( graph, source );
    }

   /**
	 * Run BellmanFord algorithm.
	 *
   *
   * @param testGraph the graph in adjacency list form
	 * @param source the source node
	 * @return a ShortestPath object conatining the solved shortest paths
	 */
    public static ShortestPath bellmanFord(  Node [] graph, int source ) {
    	return new BellmanFord( graph, source, 1 );
    }

	/**
   * Run Johnson algorithm using a linear array priority queue.
   *
   * @param testGraph the graph in adjacency list form
   * @param euclid is the test graph Euclidian
   * @return a ShortestPath array conatining the solved shortest paths
   */
    public static ShortestPath[] linearJohnson( Node [] testGraph, boolean euclid )
    {
      return linearJohnson( testGraph, euclid );
    }
    public static ShortestPath[] linearJohnson( Node [] testGraph, boolean euclid, LossyLinkPathInfo pathInfo, LossyLinkData data )
    {
	//add zero node with zero weight edges to all verticies
	if( euclid )
	    {
		testGraph[0] = new EuclidNode( 0, -1, -1 );
		Node cursor = (EuclidNode)testGraph[0];
		for( int i = 1; i < testGraph.length; i++ )
		    {
			cursor.setNext( new EuclidNode( i, ((EuclidNode)testGraph[i]).x,
							((EuclidNode)testGraph[i]).y ), 0 );
			cursor = cursor.next;
		    }
	    }
	else
	    {
		testGraph[0] = new Node( 0 );
		Node cursor = testGraph[0];
		for( int i = 1; i < testGraph.length; i++ )
		    {
			cursor.setNext( new Node( i ), 0 );
			cursor = cursor.next;
		    }
	    }
	return (new Johnson( testGraph, 'l', pathInfo, data )).allPairs;
    }

	/**
   * Run Johnson algorithm using a binary heap priority queue.
   *
   * @param testGraph the graph in adjacency list form
   * @param euclid is the test graph Euclidian
   * @return a ShortestPath array conatining the solved shortest paths
   */
	public static ShortestPath[] binaryJohnson( Node [] testGraph, boolean euclid ) {

		//add zero node with zero weight edges to all verticies
    if( euclid ) {
			testGraph[0] = new EuclidNode( 0, -1, -1 );
    	Node cursor = (EuclidNode)testGraph[0];
    	for( int i = 1; i < testGraph.length; i++ ) {
    	    cursor.setNext( new EuclidNode( i, ((EuclidNode)testGraph[i]).x,
    	    	((EuclidNode)testGraph[i]).y ), 0 );
    	    cursor = cursor.next;
    	}
		}
		else {
    	testGraph[0] = new Node( 0 );
    	Node cursor = testGraph[0];
    	for( int i = 1; i < testGraph.length; i++ ) {
    	    cursor.setNext( new Node( i ), 0 );
    	    cursor = cursor.next;
    	}
    }
    return (new Johnson( testGraph, 'b' )).allPairs;
	}

	/**
   * Run Johnson algorithm using a Fibonacci heap priority queue.
   *
   * @param testGraph the graph in adjacency list form
   * @param euclid is the test graph Euclidian
   * @return a ShortestPath array conatining the solved shortest paths
   */
	public static ShortestPath[] fibJohnson( Node [] testGraph, boolean euclid  ) {

		//add zero node with zero weight edges to all verticies
    if( euclid ) {
			testGraph[0] = new EuclidNode( 0, -1, -1 );
    	Node cursor = (EuclidNode)testGraph[0];
    	for( int i = 1; i < testGraph.length; i++ ) {
    	    cursor.setNext( new EuclidNode( i, ((EuclidNode)testGraph[i]).x,
    	    	((EuclidNode)testGraph[i]).y ), 0 );
    	    cursor = cursor.next;
    	}
		}
		else {
    	testGraph[0] = new Node( 0 );
    	Node cursor = testGraph[0];
    	for( int i = 1; i < testGraph.length; i++ ) {
    	    cursor.setNext( new Node( i ), 0 );
    	    cursor = cursor.next;
    	}
    }
    return (new Johnson( testGraph, 'f' )).allPairs;
	}

	/**
   * Run FloydWarshall algorithm
   *
   * @param testGraph the graph in adjacency list form
   * @return a FloydWarshall object conatining the solved shortest paths
    */
	public static FloydWarshall floydWarshall( double[][] graphAM ) {
		return new FloydWarshall( graphAM );
	}

	/**
   * Run A* algorithm |V|^2 times to sove the APSP problem
   * using a linear array priority queue.
   *
   * @param testGraph the graph in adjacency list form
   * @param nodes n dimension in an n x n graph
   * @return a ShortestPath object conatining the solved shortest paths
   */
	public static ShortestPath linearAStar( Node [] testGraph, int nodes ) {
		//System.out.println("hi again");
		int size = (int)Math.pow( nodes, 2 );
		for( int i = 1; i <= size; i++ )
			for( int j = i + 1; j <= size; j++ )
				if( i != j )
					if( ((EuclidNode)testGraph[i]).relevantNode &&
						((EuclidNode)testGraph[j]).relevantNode )
						new LinearAStar( (EuclidNode [])testGraph, i, j );
    return null;
	}

	/**
   * Run A* algorithm |V|^2 times to sove the APSP problem
   * using a binary heap priority queue.
   *
   * @param testGraph the graph in adjacency list form
   * @param nodes n dimension in an n x n graph
   * @return a ShortestPath object conatining the solved shortest paths
   */
	public static ShortestPath binaryAStar( Node [] testGraph, int nodes ) {
		int size = (int)Math.pow( nodes, 2 );
		for( int i = 1; i <= size; i++ )
			for( int j = i + 1; j <= size; j++ )
				if( i != j )
					if( ((EuclidNode)testGraph[i]).relevantNode &&
						((EuclidNode)testGraph[j]).relevantNode )
						new BinaryAStar( (EuclidNode [])testGraph, i, j );
    return null;
	}

	/**
   * Run A* algorithm |V|^2 times to sove the APSP problem
   * using a Fibonacci heap priority queue.
   *
   * @param testGraph the graph in adjacency list form
   * @param nodes n dimension in an n x n graph
   * @return a ShortestPath object conatining the solved shortest paths
   */
	public static ShortestPath fibAStar( Node [] testGraph, int nodes ) {
		int size = (int)Math.pow( nodes, 2 );
		for( int i = 1; i <= size; i++ )
			for( int j = i + 1; j <= size; j++ ) {
				//System.out.println( size);
				//System.out.println( "Set: " + i + " " + ", " + j );
				if( i != j )
					if( ((EuclidNode)testGraph[i]).relevantNode &&
						((EuclidNode)testGraph[j]).relevantNode )
						new FibAStar( (EuclidNode [])testGraph, i, j );
				//System.out.println( "booya " + i + " " + ", " + j );
			}
    return null;
	}
}


