/**
 *OBJECT:	This class computes the shortest path
 *				from a source node to a goal node in
 *				a Euclidian graph.  The algorithm is used
 *				is the A* method with a Binary heap
 *				priority queue.
 *
 * @version    $Id: BinaryAStar.java,v 1.1 12/26/2001
 *
 * @author    Mark Rabin
 *
 * Revisions:
 *     $Log: BinaryAStar.java,v 1.1$
 *     Revision 1.1  12/26/2001
 *     Initial revision
 *
 */


// package LossyLink;



public class BinaryAStar extends BinaryDijkstra {

 	  /**
     * Constructor serves to instantiate all needed arrays and run the
     * algorithm to completion, filling the S array with the final
     * shortest path weights and the pi array with the vertices' predecessors.
     * The constructor runs the A* algorithm until either the
	   * goal node is found or all nodes connected to the source
	   * have been searched.
     *
     * @param graph the graph in adjacency list form
     * @param source the vertex which the shortest path will eminate from
     * @param goal the id number of the goal node
     */
    public BinaryAStar( EuclidNode [] graph, int source, int goal ) {
      super( graph );
      initializeSingleSource( graph, source );

		  int min = extractMin( );
        while( min != goal ) {
        		if( min == -1 ) {
        			System.out.println( "Goal not reachable from source." );
        			break;
        		}
            S[min] = minWeight;
            for( EuclidNode curNode = graph[min]; curNode.next != null;
            	curNode = (EuclidNode)curNode.next ) {
            		//System.out.println( "relaxing " + graph[min].id +
            		//" to " + curNode.next.id );

                if( nodePos[ curNode.next.id ] != -1 &&
                		( (AStarBHNode)d[ nodePos[ curNode.next.id ] ] ).goalEstVal
                			== 0 ) {
                	( (AStarBHNode)d[ nodePos[ curNode.next.id ] ] ).goalEstVal =
                		( (EuclidNode)curNode.next ).goalEst( graph[goal] );
                }
                relax( graph[min].id, curNode.next.id, curNode.edgeWeight );
            }
            min = extractMin( );
        }
        S[min] = minWeight;
        d = null;
        nodePos = null;
    }

	  /**
     * Heapify is a process that causes an input node to be swapped with
     * the smaller of its children continually until either it reaches the
     * bottom of the tree or is less than or equal to its children. Once
     * the node "runs down" the tree, the heap property is restored
     * for that node with respect to the node's original position and its
     * new children, if any.
     *
     * @param i the index in the binary heap array of the node to
     *		 	   perform heapify on
     */
    public void heapify( int i ) {
    	//System.out.print( "Before heapify: ");
    	//debugHeap( );
    	int l = i << 1;
    	//System.out.println( "i = " + i + ", l = " + l);
    	int r = l + 1;
    	int smallest;
    		//if( l <= heapSize )
    			//System.out.println( "lVal = " + d[l].value + ",
    				//iVal = " + d[i].value );
    	if( l <= heapSize && ( (AStarBHNode)d[l] ).keyValue() <
    		( (AStarBHNode)d[i] ).keyValue() )
    	    smallest = l;
    	else
    	    smallest = i;
    	if( r <= heapSize && ( (AStarBHNode)d[r]).keyValue() <
    		( (AStarBHNode)d[smallest] ).keyValue() )
    	    smallest = r;
    	if( smallest != i ) {
    	    swap( i , smallest );
    	    heapify( smallest );
    	}
    	//System.out.print( "After heapify: ");
    	//debugHeap( );

    }


	  /**
     * Initialize the shortest-path estimates and predecessor function.
     * The predecessor function is left at the default of all zeros.
     * Because there is no zero node in the graph, zero denotes a null
     * value for a predecessor. The source node's d value is initially
     * set to weight zero and all other nodes are set to have weight equal
     * to the maximum double floating point precision value to represent
     * infinity.
     *
     * @param graph the input graph in adjacency-list form
     * @param source the number identifier of the source node
     */
    public void initializeSingleSource( Node [] graph, int source ) {
    	//System.out.println( "initializeSingleSource" );

    	//instantiate arrays
    	d = new AStarBHNode[graph.length];
    	nodePos = new int[graph.length];

    	d[1] = new AStarBHNode( source, 0 );
    	nodePos[source] = 1;

    	//BHNodes greater than source
    	int i = 2;
    	for( int nextNodeNum = source + 1;
    	     nextNodeNum < graph.length; i++, nextNodeNum++ ) {

    	    d[i] = new AStarBHNode( nextNodeNum, Double.MAX_VALUE );
    	    nodePos[nextNodeNum] = i;
    	}

    	//BHNodes less than source
    	for( int nextNodeNum = 1;
    	     nextNodeNum < source; i++, nextNodeNum++ ) {

    	    d[i] = new AStarBHNode( nextNodeNum, Double.MAX_VALUE );
    	    nodePos[nextNodeNum] = i;
    	}
    	heapSize = d.length - 1;
    	//debugHeap( );
    }

    /**
		 * This method lowers the value of a node in the binary heap and restores
		 * the heap property.  The node is found in the heap using the nodePos
		 * array.
		 *
		 * @param nodeNum the number of the node to have its key decreased
		 * @param newVal the new key value
		 */
    public void decreaseKey( int nodeNum, double newVal ) {
    	int y = nodePos[ nodeNum ];
    	d[y].value = newVal;
    	int parent = y >> 1;
    	//System.out.println("childVal: " + d[y].value + ", parentVal: "
    		//+ d[parent].value);
    	while( parent != 0 && ( (AStarBHNode)d[y] ).keyValue() <
    		( (AStarBHNode)d[parent] ).keyValue() ){
    	    swap( y, parent );
    	    y = parent;
    	    parent = y >> 1;
    	}
    }

}