/**
 *OBJECT:	This class computes the shortest path
 *				from a source node to all other nodes in
 *				a directed graph.  The algorithm is used
 *				is the Dijkstra method with a Binary heap
 *				priority queue.
 *
 * @version    $Id: BinaryDijkstra.java,v 1.1 12/26/2001
 *
 * @author    Mark Rabin
 *
 * Revisions:
 *     $Log: BinaryDijkstra.java,v 1.1$
 *     Revision 1.1  12/26/2001
 *     Initial revision
 *
 */


// package LossyLink;


public class BinaryDijkstra extends ShortestPath{

  int heapSize;


  //array of BHNodes containing information on the upper bounds of the
  //weight of a shortest path from source s to a node in d
  BHNode [] d;

  //positions of nodes in the heap -- needed for decreaseKey
  int [] nodePos;

  /**
  * Constuctor for use by subclasses.
  *
  * @param graph the graph in adjacency list form
  */
  public BinaryDijkstra( Node [] graph ) {
  	super( graph );
  }

  /**
  * Constructor serves to instantiate all needed arrays and run the
  * algorithm to completion, filling the S array with the final shortest
  * path weights and the pi array with the vertices' predecessors.
  *
  * @param graph the graph in adjacency list form
  * @param source the vertex which all paths will eminate from
  */
  public BinaryDijkstra( Node [] graph, int source ) {
      super( graph, source );

      d = null;
      nodePos = null;
  }

  /**
  * Heapify is a process that causes an input node to be swapped with
  * the smaller of its children continually until either it reaches the
  * bottom of the tree or is less than or equal to its children. Once
  * the node "runs down" the tree, the heap property is restored
  * for that node with respect to the node's original position and its
  * new children, if any.
  *
  * @param i the index in the binary heap array of the node to
  *		 	   perform heapify on
  */
  public void heapify( int i ) {
      //System.out.print( "Before heapify: ");
      //debugHeap( );
      int l = i << 1;
      //System.out.println( "i = " + i + ", l = " + l);
      int r = l + 1;
      int smallest;
      	//if( l <= heapSize )
      		//System.out.println( "lVal = " + d[l].value + ",
      			//iVal = " + d[i].value );
      if( l <= heapSize && d[l].value < d[i].value )
          smallest = l;
      else
          smallest = i;
      if( r <= heapSize && d[r].value < d[smallest].value )
          smallest = r;
      if( smallest != i ) {
          swap( i , smallest );
          heapify( smallest );
      }
      //System.out.print( "After heapify: ");
      //debugHeap( );

  }

	/**
	* The method is used by heapify and decreaseKey to switch the positions
	* of two nodes in the binary heap.
	*
	* @param a index of node to be swapped in the binary heap array
	* @param b index of other node to be swapped in the binary heap array
	*/
    public void swap( int a, int b ) {
        BHNode temp = d[a];
        d[a] = d[b];
        d[b] = temp;
        int tempInt = nodePos[ d[a].id ];
        nodePos[ d[a].id ] = nodePos[ d[b].id ];
        nodePos[ d[b].id ] = tempInt;
    }
	/**
	 * This method removes the minimum valued node form the binary heap,
	 * shifts the node at the bottom of the heap to the top and calls
	 * heapify on that node.  If the heap is empty, -1 is returned.
	 *
	 * @return the node number with the minimum shortest path estimate
	 */
  public int extractMin( ) {
      BHNode min;
      if( heapSize < 1 ) return -1;
      min = d[1];
      nodePos[ d[1].id ] = -1; //node removed from node positions
      d[1] = d[heapSize];
      nodePos[ d[1].id ] = 1;  //largest value node now in position one
      d[heapSize] = null;  //remove node
      heapSize--;
      heapify( 1 );
      minWeight = min.value;
      return min.id;
  }


  /**
   * Initialize the shortest-path estimates and predecessor function.
   * The predecessor function is left at the default of all zeros.
   * Because there is no zero node in the graph, zero denotes a null
   * value for a predecessor. The source node's d value is initially
   * set to weight zero and all other nodes are set to have weight equal
   * to the maximum double floating point precision value to represent
   * infinity.
   *
   * @param graph the input graph in adjacency-list form
   * @param source the number identifier of the source node
   */
  public void initializeSingleSource( Node [] graph, int source ) {
      //System.out.println( "initializeSingleSource" );

      //instantiate arrays
      d = new BHNode[graph.length];
      nodePos = new int[graph.length];

      d[1] = new BHNode( source, 0 );
      nodePos[source] = 1;

      //BHNodes greater than source
      int i = 2;
      for( int nextNodeNum = source + 1;
           nextNodeNum < graph.length; i++, nextNodeNum++ ) {

          d[i] = new BHNode( nextNodeNum, Double.MAX_VALUE );
          nodePos[nextNodeNum] = i;
      }

     //BHNodes less than source
      for( int nextNodeNum = 1;
           nextNodeNum < source; i++, nextNodeNum++ ) {

          d[i] = new BHNode( nextNodeNum, Double.MAX_VALUE );
          nodePos[nextNodeNum] = i;
      }
      heapSize = d.length - 1;
      //debugHeap( );
  }

  /**
   * Test if shortest path to v can be improved by going through u, and
   * if so, updating d[v] and pi[v]. w is the weight from u to v.
   * if nodePos[v] is -1, then the shortest path to v has already been
   * found and there is no need to do the relaxation.
   *
   * @param u vertex relaxing from
   * @param v vertex relxing to
   * @param w the weight from u to v
   */
  public void relax( int u, int v, double w ) {
      //System.out.println( "relaxing" + u + v + w );
      //System.out.println( "NodePos[v]: " + nodePos[v]
      	//+ " S[u]: " + S[u] );
      if( nodePos[v] == -1 ) return;
      if( d[ nodePos[v] ].value > S[u] + w ) {
          decreaseKey( v, S[u] + w ); //d[ nodePos[v] ].value = S[u] + w;
          pi[v] = u;
          //System.out.println( "updating d" + d[ nodePos[v] ].value );
          heapify( nodePos[v] );
          // try {
          //  SimpleInput.readInt();} catch(Exception e){}
      }
  }

	/**
	* This method lowers the value of a node in the binary heap and restores
	* the heap property.  The node is found in the heap using the nodePos
	* array.
	*
	* @param nodeNum the number of the node to have its key decreased
	* @param newVal the new key value
	*/
  public void decreaseKey( int nodeNum, double newVal ) {
      int y = nodePos[ nodeNum ];
      d[y].value = newVal;
      int parent = y >> 1;
      //System.out.println("childVal: " + d[y].value + ",
      	//parentVal: " + d[parent].value);
      while( parent != 0 && d[y].value < d[parent].value ){
          swap( y, parent );
          y = parent;
          parent = y >> 1;
      }
  }

	/**
	* Print out the heap, the size of the heap, and nodePos array for
	* debugging.
	*/
  public void debugHeap( ) {
      System.out.println( "heapSize = " + heapSize );
      System.out.print( "d = " );
      for( int j = 1; j < d.length; j++ )
          System.out.print( " " + d[j] );
      System.out.println();
      System.out.print( "nodePos = " );
      for( int j = 1; j < nodePos.length; j++ )
          System.out.print( " " + nodePos[j] );
      System.out.println();
  }

}
