/**
 *OBJECT:	This class computes the shortest path
 *				from a source node to a goal node in
 *				a Euclidian graph.  The algorithm is used
 *				is the A* method with a Fibonacci heap
 *				priority queue.
 *
 * @version    $Id: FibAStar.java,v 1.1 12/26/2001
 *
 * @author    Mark Rabin
 *
 * Revisions:
 *     $Log: FibAStar.java,v 1.1$
 *     Revision 1.1  12/26/2001
 *     Initial revision
 *
 */


// package LossyLink;


public class FibAStar extends FibDijkstra {

	/**
	 * Constructor serves to instantiate all needed arrays and run the
   * algorithm to completion, filling the S array with the final
   * shortest path weights and the pi array with the vertices' predecessors.
   * The constructor runs the A* algorithm until either the
	 * goal node is found or all nodes connected to the source
	 * have been searched.
	 *
	 * @param graph the graph in adjacency list form
   * @param source the source vertex in the graph
   */
	public FibAStar( EuclidNode [] graph, int source, int goal ) {
   	super( graph );
		initializeSingleSource( graph, source );

      int min = extractMin( );
      //debugHeap();
        while( min != goal ) {
        		if( min == -1 ) {
        			System.out.println( "Goal not reachable from source." );
        			break;
        		}
            S[min] = minWeight;
            for( EuclidNode curNode = graph[min]; curNode.next != null;
            	curNode = (EuclidNode)curNode.next ) {
            		//System.out.println( "relaxing " + graph[min].id + " to " +
            		//	curNode.next.id );


                if(
                	( (AStarFibNode)nodePos[ curNode.next.id ] ).goalEstVal ==
                	 0 )
                	( (AStarFibNode)nodePos[ curNode.next.id ] ).goalEstVal =
                		( (EuclidNode)curNode.next ).goalEst( graph[goal] );

                relax( graph[min].id, curNode.next.id, curNode.edgeWeight );
                //System.out.println("After relax:  ");
                //debugHeap();
            }
            //System.out.println("begin extract min " );
            //if( heapSize > 0 )debugHeap();
            min = extractMin( );
            //if( heapSize > 0 ) debugHeap();
            //System.out.println("sucessful extract min: " + min );
        }
        S[min] = minWeight;

      //release resources that are no longer needed
      nodePos = null;
      minNode = null;
      originalRootList = null;
    }

	/**
	 * Consolidation works by repeatedly finding two roots x and y in the
	 * root list with the same degree, where the key of x is less than or
	 * equal to the key of y.  Then y is removed from the root list and
	 * becomes a child of x.  The degree of x is incremented and the mark
	 * on y is set to false.  An auxiliary array is used to keep track of
	 * which root node has a claim on which degree.  This auxiliary array
	 * runs from 0 to the maximum degree of a node.
	 */
  public void consolidate() {
      //compute an upper bound on the maximum number of children a root
      //node can have
      int maxDegree = (int)Math.ceil( Math.log(heapSize)/Math.log(1.62) );
      //System.out.println( maxDegree );
      //System.out.println("Consolidate:  ");

      FibNode [] A = new AStarFibNode[ maxDegree + 1 ];
      FibNode x = minNode;
      int d = minNode.degree;

      originalRootList[0] = minNode;
      int i = 1;
      for( FibNode cursor = minNode.right; cursor != minNode;
           cursor = cursor.right, i++ )
          originalRootList[i] = cursor;
	   originalRootList[i] = null;

      //loop through the original root list ignoring any nodes
      //that have since left the root list
      i = 0;
      for( FibNode cursor = originalRootList[i]; cursor != null;
           cursor = originalRootList[++i] ) {

          //ignore nodes not in the root list
          if( cursor.parent != null ) continue;

          x = cursor;
          d = cursor.degree;
          //System.out.println("d = " + d );
          while( A[d] != null ) {
              //printArray(A);
              FibNode y = A[d];
              //System.out.println( "y = " + y + " x = " + x);
              if( ( (AStarFibNode)x ).keyValue() > ( (AStarFibNode)y ).keyValue() )
                  { FibNode t = x; x = y; y = t; }  //swap
              fibHeapLink( y, x );
              //minNode = x;  //for debuging only
              //debugHeap();
              //freeze();
              A[d] = null;
              d++;
              //System.out.println("\td = " + d );
          }
          //printArray( A );
          A[d] = x;
          //System.out.println( "cursor = " + cursor );
          //System.out.println( "nextCursor = " + originalRootList[i+ 1] );
          //if( cursor == lastRoot ) break;  //old method -- doesn't work
      }

      //rebuild root list
      minNode = null;
      for( i = 0; i <= maxDegree; i++ ) {
          if( A[i] != null ) {
              addToRootList( A[i] );
              if( minNode == null || ( (AStarFibNode)A[i] ).keyValue() <
              	( (AStarFibNode)minNode ).keyValue() )
                  minNode = A[i];
          }
      }
      //debugHeap();
      //freeze();
  }

	/**
   * Initialize the shortest-path estimates and predecessor function.
   * The predecessor function is left at the default of all zeros.
   * Because there is no zero node in the graph, zero denotes a null
   * value for a predecessor. The source node's value is initially
   * set to weight zero and all other nodes are set to have weight equal
   * to the maximum double floating point precision value to represent
   * infinity.
   *
   * @param graph the graph in adjacency list form
   * @param source the source vertex in the graph
   *
   */
  public void initializeSingleSource( Node [] graph, int source ) {
      //System.out.println( "initializeSingleSource" );
      nodePos = new AStarFibNode[graph.length];
      originalRootList = new AStarFibNode[graph.length];

      for( int i = 1; i < graph.length; i++ ) {

          nodePos[i] = new AStarFibNode( i, Double.MAX_VALUE );
          addToRootList( nodePos[i] );
      }

      nodePos[source].value = 0;
      minNode = nodePos[source];
      heapSize = graph.length - 1;
      //debugHeap( );
  }





	/**
	* This method lowers the value of a node in the Fibonacci heap and
	* restores the heap property.  The node is found in the heap using
	* the nodePos array.
	*
	* @param nodeNum the number of the node to have its key decreased
	* @param newVal the new key value
	*/
  public void decreaseKey( int nodeNum, double newVal ) {
      AStarFibNode x = (AStarFibNode)nodePos[nodeNum];
      x.value = newVal;
      AStarFibNode y = ( (AStarFibNode)( x.parent ) );
      //System.out.print("Decrease Key:  ");
      //debugHeap();
      if( y != null && x.keyValue() < y.keyValue() ) {
          cut( x, y );
          cascadingCut( y );
      }
      //debugHeap();
      if( x.keyValue() < ( (AStarFibNode)minNode ).keyValue() )
          minNode = x;
      //System.out.println("look end of decrease key: ");
      //debugHeap();
  }

}