/**
 *OBJECT:	This class computes the shortest path
 *				from a source node to all other nodes in
 *				a directed graph.  The algorithm is used
 *				is the Dijkstra method with a Fibonacci heap
 *				priority queue.
 *
 * @version    $Id: FibDijkstra.java,v 1.1 12/26/2001
 *
 * @author    Mark Rabin
 *
 * Revisions:
 *     $Log: FibDijkstra.java,v 1.1$
 *     Revision 1.1  12/26/2001
 *     Initial revision
 *
 */


// package LossyLink;



public class FibDijkstra extends ShortestPath {

	int heapSize;

	//circular, doubly linked list of upper bounds on the
	//weight of a shortest path from source s to node: FibNode.id
	FibNode minNode;

	//positions of nodes in the heap -- needed for decreaseKey
	FibNode [] nodePos;

	//This node array  is used to record the root list's configuration
	//before the root list is altered during consolidation.
	FibNode [] originalRootList;

	/**
	 * Constuctor for use by subclasses.
	 *
	 * @param graph the graph in adjacency list form
	 */
	public FibDijkstra( Node [] graph ) {
		super( graph );
	}

  /**
	 * This constructor runs the Dijkstra algorithm until
	 * all nodes connected to the source have been searched.
	 *
	 * @param graph the graph in adjacency list form
   * @param source the source vertex in the graph
   */
  public FibDijkstra( Node [] graph, int source ) {
    super( graph, source );

		//test heap used for debugging
        /*initializeTestHeap( );
        debugHeap();
        extractMin();
        debugHeap();
        decreaseKey( 12, 15 );
        debugHeap();
        decreaseKey( 15, 5 );
        debugHeap();*/

    //release resources that are no longer needed
  	nodePos = null;
  	minNode = null;
  	originalRootList = null;
  }


  /**
	* This method removes the minimum valued node form the Fibonacci heap,
	* moves all children of the removed node to the root list, then
	* consolidates the heap. Linking roots of equal degree until at most
	* one root remains of each degree consolidates the root list.
	* If the heap is empty, null is returned.
	*
	* @return the node with the minimum shortest path estimate
	*/
  public int extractMin( ) {
  	//System.out.println("Heap size = " + heapSize + " minNode: " + minNode);
  	FibNode min = null;
  	if( heapSize < 1 ) return -1;
  	min = minNode;

	  //add all children of the minimum node to the root list
    if( min.child != null ) {

    	FibNode firstborn = min.child;
    	FibNode lastborn = firstborn.left;

    	//remove parent pointers
    	firstborn.parent = null;
    	for( FibNode cursor = firstborn.left; cursor != firstborn;
    	     cursor = cursor.left )
    	    cursor.parent = null;

    	//connect left end of child list to the root list
    	firstborn.left = min.left;
    	min.left.right = firstborn;

    	//connect right end of child list to the root list
    	lastborn.right = min;
    	min.left = lastborn;

    }

    FibNode newHead = min.right;
    removeFromRootList( min );

	  //check to see if heap will be empty after extractMin
    if( min == newHead ) {
    	minNode = null;
    }
    //set minNode to arbitrary position and consolidate heap
    else {
    	minNode = newHead;
    	//debugHeap();

    	//at size 2 there will only be 1 node left after extractMin
    	//so nothing needs to be consolidated
    	if( heapSize != 2 )
            consolidate();
    }
    heapSize--;
    minWeight = min.value;
    return min.id;
  }


	/**
	 * Consolidation works by repeatedly finding two roots x and y in the
	 * root list with the same degree, where the key of x is less than or
	 * equal to the key of y.  Then y is removed from the root list and
	 * becomes a child of x.  The degree of x is incremented and the mark
	 * on y is set to false.  An auxiliary array is used to keep track of
	 * which root node has a claim on which degree.  This auxiliary array
	 * runs from 0 to the maximum degree of a node.
	 */
  public void consolidate() {
  	//compute an upper bound on the maximum number of children a root
  	//node can have
  	int maxDegree = (int)Math.ceil( Math.log(heapSize)/Math.log(1.62) );
  	//System.out.println( maxDegree );

  	FibNode [] A = new FibNode[ maxDegree + 1 ];
  	FibNode x = minNode;
  	int d = minNode.degree;

  	originalRootList[0] = minNode;
  	int i = 1;
  	for( FibNode cursor = minNode.right; cursor != minNode;
  	     cursor = cursor.right, i++ )
  	    originalRootList[i] = cursor;
	  originalRootList[i] = null;

    //loop through the original root list ignoring any nodes
    //that have since left the root list
    i = 0;
    for( FibNode cursor = originalRootList[i]; cursor != null;
         cursor = originalRootList[++i] ) {

      	//ignore nodes not in the root list
      	if( cursor.parent != null ) continue;

      	x = cursor;
      	d = cursor.degree;
      	//System.out.println("d = " + d );
      	while( A[d] != null ) {
      		//printArray(A);
      		FibNode y = A[d];
      		//System.out.println( "y = " + y + " x = " + x);
      		if( x.value > y.value )
      		    { FibNode t = x; x = y; y = t; }  //swap
      		fibHeapLink( y, x );
      		//minNode = x;  //for debuging only
      		//debugHeap();
      		//freeze();
      		A[d] = null;
      		d++;
      		//System.out.println("\td = " + d );
      	}
      	//printArray( A );
      	A[d] = x;
      	//System.out.println( "cursor = " + cursor );
      	//System.out.println( "nextCursor = " + originalRootList[i+ 1] );
      	//if( cursor == lastRoot ) break;  //old method -- doesn't work
      }

      //rebuild root list
      minNode = null;
      for( i = 0; i <= maxDegree; i++ ) {
      	if( A[i] != null ) {
        	addToRootList( A[i] );
          if( minNode == null || A[i].value < minNode.value )
          	minNode = A[i];
        }
      }
      //debugHeap();
      //freeze();
  }

  /**
   * This method removes y from the root list, makes y a child of x, and
   * unmarks y if it was previously marked.
   *
   * @param y the new child
   * @param x the parent
   *
   */
  public void fibHeapLink( FibNode y, FibNode x ) {
      removeFromRootList( y );
      makeChild( y, x );
      y.mark = false;
  }


  /**
   * Establish the parent and child relationship for two nodes.  If the
   * parent doesn't have any other children, then simply set the parent's
   * child pointer to the child and the child's parent pointer to the
   * parent.  If the parent already has at least one child, then
   * concatenate the child onto the parents circular, doubly linked
   * child list and set the child's parent pointer to the new parent.
   * In either case, increment the degree of the parent.
   *
   * @param child the node to be made a child of the parent
   * @param parent the node to be made the parent of the child
   *
   */
  public FibNode makeChild( FibNode child, FibNode parent ) {
      if( parent.degree == 0 ){
          parent.child = child;
          child.left = child.right = child;
      }
      else {
          FibNode head = parent.child;
          FibNode tail = head.left;
          child.left = tail;
          child.right = head;
          tail.right = child;
          head.left = child;
      }
      child.parent = parent;
      parent.degree++;
      return child;
  }


  /**
   * Initialize the shortest-path estimates and predecessor function.
   * The predecessor function is left at the default of all zeros.
   * Because there is no zero node in the graph, zero denotes a null
   * value for a predecessor. The source node's value is initially
   * set to weight zero and all other nodes are set to have weight equal
   * to the maximum double floating point precision value to represent
   * infinity.
   *
   * @param graph the graph in adjacency list form
   * @param source the source vertex in the graph
   *
   */
  public void initializeSingleSource( Node [] graph, int source ) {
      //System.out.println( "initializeSingleSource" );
      nodePos = new FibNode[graph.length];
      originalRootList = new FibNode[graph.length];

      for( int i = 1; i < graph.length; i++ ) {

          nodePos[i] = new FibNode( i, Double.MAX_VALUE );
          addToRootList( nodePos[i] );
      }

      nodePos[source].value = 0;
      minNode = nodePos[source];
      heapSize = graph.length - 1;
      //debugHeap( );
  }


  /**
   * concatenate the input FibNode to the root list
   *
   * @param addMe the node to be addded to the root list
   */
  public FibNode addToRootList( FibNode addMe ) {
      if( minNode != null ) {
          addMe.right = minNode;
          addMe.left = minNode.left;
          minNode.left.right = addMe;
          minNode.left = addMe;
      }
      else
          addMe.right = addMe.left = minNode = addMe;
      return addMe;
  }

  /**
   * remove the input FibNode from the root list
   *
   * @param removeMe the node to be removed from the root list
   */
  public void removeFromRootList( FibNode removeMe ) {
      removeMe.left.right = removeMe.right;
      removeMe.right.left = removeMe.left;
      removeMe.right = removeMe.left = removeMe;
  }

  /**
   * Test if shortest path to v can be improved by going through u, and
   * if so, updating d[v] and pi[v]. w is the weight from u to v.
   *
   * @param u vertex relaxing from
   * @param v vertex relxing to
   * @param w the weight from u to v
   */
  public void relax( int u, int v, double w ) {
  		//if( heapSize > 0 ) debugHeap();
      //System.out.println( "relaxing" + u + ", " + v + " w = " + w );
      //System.out.println( " S[u]: " + S[u] );

      if( nodePos[v].value > S[u] + w ) {
          decreaseKey( v, S[u] + w ); //will set nodePos[v].value=S[u]+w
          pi[v] = u;
          //System.out.println( "updating FibNode" + nodePos[v].value );
          //if( heapSize > 0 ) debugHeap();
      }
  }

	/**
	 * This method lowers the value of a node in the Fibonacci heap and
	 * restores the heap property.  The node is found in the heap using
	 * the nodePos array.
	 *
	 * @param nodeNum the number of the node to have its key decreased
	 * @param newVal the new key value
	 */
  public void decreaseKey( int nodeNum, double newVal ) {
      FibNode x = nodePos[nodeNum];
      x.value = newVal;
      FibNode y = x.parent;
      if( y != null && x.value < y.value ) {
          cut( x, y );
          cascadingCut( y );
      }
      if( x.value < minNode.value )
          minNode = x;
  }

	/**
	 * This method is used by the decreaseKey method to remove x from the
	 * child list of y and add x to the root list.
	 *
	 * @param x the child to be cut
	 * @param y the parent about to lose its child
	 */
  public void cut( FibNode x, FibNode y ) {
      //remove x from the child list of y
      if( y.degree == 1 )
          y.child = null;
      else {
          if( y.child == x )
              y.child = x.right;
          x.left.right = x.right;
          x.right.left = x.left;
      }

      y.degree--;
      addToRootList( x );
      x.parent = null;
      x.mark = false;
  }

	/**
	 * Recursive cutting of nodes continuing up the tree until an unmarked
	 * node is found or the root is reached.
	 *
	 * @param y the node to start the cut at
	 */
  public void cascadingCut( FibNode y ) {
      FibNode z = y.parent;
      if( z != null ) {
          if( !y.mark )
              y.mark = true;
          else {
              cut( y, z );
              cascadingCut( z );
          }
      }
  }

/*********************   Debugging Methods  *****************************/

 	/**
   * Initial heap as in Cormen figure 21.3 for testing first the extractMin
   * and then the decreaseKey operation
   */
  public void initializeTestHeap( ) {
      //System.out.println( "initializeSingleSource" );
      nodePos = new FibNode[16];
      pi = new int[16];
      nodePos[1] = minNode = new FibNode( 1, 3 );
      nodePos[2] = addToRootList( new FibNode( 2, 17 ) );
      nodePos[3] = addToRootList( new FibNode( 3, 24 ) );
      nodePos[4] = addToRootList( new FibNode( 4, 23 ) );
      nodePos[5] = addToRootList( new FibNode( 5, 7 ) );
      nodePos[6] = addToRootList( new FibNode( 6, 21 ) );
      nodePos[7] = makeChild( new FibNode( 7, 18 ), minNode );
      nodePos[7].mark = true;
      nodePos[8] = makeChild( new FibNode( 8, 52 ), minNode );
      nodePos[9] = makeChild( new FibNode( 9, 38 ), minNode );
      nodePos[10] = makeChild( new FibNode( 10, 30 ), minNode.right ) ;
      nodePos[11] = makeChild( new FibNode( 11, 26 ), minNode.right.right );
      nodePos[11].mark = true;
      nodePos[12] = makeChild( new FibNode( 12, 46 ), minNode.right.right );
      nodePos[13] = makeChild( new FibNode( 13, 39 ), minNode.child );
      nodePos[13].mark = true;
      nodePos[14] = makeChild( new FibNode( 14, 41 ), minNode.child.left );
      nodePos[15] = makeChild( new FibNode( 15, 35 ),
                               minNode.right.right.child );
      heapSize = 15;
      //debugHeap( );
  }


  /**
   * This method is used in debugging to print out the auxiliary array used
   * in the consolidate function called during the extractMin operation.
   *
   * @param A the auxiliary array of references to nodes of degree equal
   *          the index of the array
   */
  public void printArray( FibNode [] A ) {
  	for( int i = 0; i < A.length; i++ )
  		System.out.print( "A[" + i + "] = " + A[i] + ", " );
  	System.out.println();
  }

  /**
   * This method is used in debugging to pause the program and/or exit
   */
  public void freeze() {
  	try {
  		if( SimpleInput.readString().equals("q") )
  	  	System.exit(0);

  	} catch( Exception e ) {}
  }

  /**
   * This method is used to ouput the current state of the heap
   */
  public void debugHeap( ) {
      System.out.println( "heapSize = " + heapSize );
      System.out.println( "min = " + minNode );
      minNode.printHeap();
  }

}
