/**
 *OBJECT:	This class creates random graphs and
 *				calls upon different algorithms to
 *				compute the shortest paths.
 *
 * @version    $Id: Graph.java,v 1.1 12/26/2001
 *
 * @author    Mark Rabin
 *
 * Revisions:
 *     $Log: Graph.java,v 1.1$
 *     Revision 1.1  12/26/2001
 *     Initial revision
 *
 */


// package LossyLink;

import java.util.*;


public class Graph {

	/** total number of nodes for non-Euclidian graph
	 *  n dimension of n x n grid for Euclidian graph */
  public static final int NODES = 9;

  /** the probabilty of an edge existing from one node to another
  	* edges restricted to adjacent nodes in Euclidian graph */
  public static final double PROB = .75;

	/** minimum value for an edge weight */
	public static final double WEIGHTLOW = 1;

	/** maximum value for an edge weight */
  public static final double WEIGHTHIGH = 1;

  /** number used as a seed in random number generator */
  public static final long SEED = 1;

  /** should the generated graph be a Euclidian graph */
  public static final boolean EUCLID = false;

  /** how many different algorithms or algorithm variations
   *	 will be tested - if all algorihtms are used there are 14 */
  public static final int NUMALGS = 7;

  /** number of times each algorithm is run (for averaging) */
  public static final int NUM_TIMES = 5;

  /** if true, integer weights only */
	public static final boolean INT_ONLY = true;

	/** number of nodes to be solved for in the "many pairs" problem */
	public static final int NUM_NODES_AS_SOURCE  = 4;

	/**
 	* The main method create a random graph and
 	* calls upon different algorithms to
 	* compute the shortest paths.
 	*/
	public static void main( String args[] ) {
		//Node testGraph [] = genRandGraph(
		//  	NODES, PROB, WEIGHTLOW, WEIGHTHIGH );
		//Node testGraph [] = genRandGraphMaxOrMinOnly(
		//  NODES, PROB, WEIGHTLOW, WEIGHTHIGH );
		//EuclidNode testGraph [] = genRandEuclid( NODES, PROB, NUM_NODES_AS_SOURCE );
    //Node testGraph [] = genTestGraph();
    Node testGraph [] = genJohnsonTestGraph();
    printGraph( testGraph );

    //double[][] graphAM = convertToMatrix( testGraph );
    //printMatrix( graphAM );

    //System.out.println( ( new FibAStar( testGraph, 1, 9 ) ).allVals() );
    //System.exit( 0 );

    //    if( args.length < 1 ) {
    //		runAllAlgs( args, testGraph, graphAM );
    // }
    //		else {
    	runShortestPath( args, testGraph, 1, EUCLID );
	//}
	}

	/**
	* Compute all pairs shortest path using all available algorithms.
	* Test to see that results of all algorithm match
	*
	* @param args command-line arguements
	* @param testGraph the graph in adjacency list form
	* @param graphAM the graph in adjacency matrix form
	*/
	public static void runAllAlgs(
		String args [], Node [] testGraph, double[][] graphAM ) {

		ShortestPath allAlgs [][] = new ShortestPath[NUMALGS][NODES + 1];
		String arg[][] = new String[NUMALGS][2];

		//Dijkstra with linear array priority queue
		arg[0][0] = "1";
		arg[0][1] = "l";
		//Dijkstra with binary heap priority queue
		arg[1][0] = "1";
		arg[1][1] = "b";
		//Dijkstra with fibinacci heap priority queue
		arg[2][0] = "1";
		arg[2][1] = "f";
		//Belman Ford
		arg[3][0] = "2";
		//Johnson with linear array priority queue
		arg[4][0] = "3";
		arg[4][1] = "l";
		//Johnson with binary heap priority queue
		arg[5][0] = "3";
		arg[5][1] = "b";
		//Johnson with fibinacci heap priority queue
		arg[6][0] = "3";
		arg[6][1] = "f";

		FloydWarshall fw = null;
		//run all algorithms on the same test graph
		for( int count = 0; count < NUM_TIMES; count++ ) {
			for( int i = 0; i < NUMALGS; i++ ) {
				if( i < 4 )
					for( int j = 1; j <= NODES; j++ )
						allAlgs[i][j] = runShortestPath( arg[i], testGraph, j, EUCLID );
				else
					allAlgs[i] = runJohnson( arg[i], testGraph, EUCLID );
			}
			fw = APSP.floydWarshall( graphAM );

			if( EUCLID ) {
				APSP.linearAStar( testGraph, NODES );
				APSP.binaryAStar( testGraph, NODES );
				APSP.fibAStar( testGraph, NODES );
			}
		}

		//check to see that shortest path weights match for all algorithms
		for( int i = 1; i < NUMALGS; i++ ){
			System.out.println("--" + i + "--");
			for( int j = 1; j <= NODES; j++ )
				if( !allAlgs[0][j].equals( allAlgs[i][j] ) )
					System.out.println( "source: " + j + "\n" +
						allAlgs[0][j] + "\n\n" + allAlgs[i][j] );
		}
		//check Floyd-Warshall
		for( int i = 1; i <= NODES; i++ )
			for( int j = 1; j <= NODES; j++ )
				if( allAlgs[0][i].S[j] != fw.D[i][j] )
					System.out.println( "source: " + i + "\nnode: " +
						j + "\n" + allAlgs[0][i].S[j] + "\n\n" + fw.D[i][j] );


	}


/**
* Run one shortest path algorithm.
*
* @param args command-line arguements decide which
*		algorithm to use and priority queue type
*		Proper usage: java Graph <algorithm number> <priority queue type>
*		1.Dijkstra, 2.Bellman-Ford, or 3.Johnson
*		l for linear array priority queue
*		b for binary heap priority queue
*		f for fibonacci heap priority queue
* @param testGraph the graph in adjacency list form
* @param source the source node
*/
	public static ShortestPath runShortestPath(
		String args[], Node [] testGraph, int source, boolean euclid ) {

  	ShortestPath paths = null;

    switch( Integer.parseInt( args[0] ) ) {
    case 1:
        paths = choosePQ( args[1].charAt(0), testGraph, source );
        break;
    case 2:
        paths = APSP.bellmanFord( testGraph, source );
        break;
    case 3:
        paths = runJohnson( args, testGraph, euclid)[source];
        break;
    default:
        System.out.println( args[0] + " is not a algorithm type");
        System.exit(0);
    }
    return paths;
	}

   /**
   * Run Johnson algorithm.
   *
   * @param args command-line arguements decide which
   *				algorithm to use and priority queue type
	 *				Proper usage: java Graph <algorithm number> <priority queue type>
	 *				l for linear array priority queue
   *				b for binary heap priority queue
   *				f for fibonacci heap priority queue
   * @param testGraph the graph in adjacency list form
   */
	public static ShortestPath[] runJohnson(
		String args[], Node [] testGraph, boolean euclid ) {

    switch( args[1].charAt(0) ) {
    	case 'l':
    	    return APSP.linearJohnson( testGraph, euclid );/////////////////////////////////////////////////////////////////////////
    	case 'b':
    	    return APSP.binaryJohnson( testGraph, euclid );
    	case 'f':
    	    return APSP.fibJohnson( testGraph, euclid );
    	default:
    	    System.out.println( "error, not a know priority queue type");
    	    System.exit(0);
    	}
    return null;

	}


	 /**
	 * Run Dijkstra algorithm.
	 *
	 * @param choice l for linear array priority queue
   *				b for binary heap priority queue
   *				f for fibonacci heap priority queue
   *
   * @param testGraph the graph in adjacency list form
	 * @param source the source node
	 */
    public static ShortestPath choosePQ(
    	char choice, Node [] graph, int source ) {

    	switch( choice ) {
    	case 'l':
    	    return APSP.linearDijkstra( graph, source );
    	case 'b':
    	    return APSP.binaryDijkstra( graph, source );
    	case 'f':
    	    return APSP.fibDijkstra( graph, source );
    	default:
    	    System.out.println( "error, not a know priority queue type");
    	    System.exit(0);
    	}
    	return null;
    }

 	/**
 	* The first graph used for testing.
 	*
 	* @return the graph in adjacency list form
 	*/
  public static Node [] genTestGraph() {
    Node testGraph [] = new Node[6];

		for( int i = 1; i < 6; i++ )
			testGraph[i] = new Node( i );

		testGraph[1].setNext( new Node( 2 ), 3 );
		Node cursor = testGraph[1].next;
		cursor.setNext( new Node( 4 ), 4 );
		cursor = cursor.next;
		cursor.setNext( new Node( 5 ), 8 );

		testGraph[2].setNext( new Node( 3 ), 1 );


		testGraph[3].setNext( new Node( 4 ), 1 );

		testGraph[4].setNext( new Node( 5 ), 2 );

    testGraph[5].setNext( new Node( 4 ), 6 );

    return testGraph;
  }


	/**
 	* Graph used for testing Johnson algorithm.
 	*
 	* @return the graph in adjacency list form
 	*/
    public static Node [] genJohnsonTestGraph() {
        Node testGraph [] = new Node[6];
        for( int i = 1; i < 6; i++ )
            testGraph[i] = new Node( i );

        testGraph[1].setNext( new Node( 2 ), 3 );
        testGraph[1].next.setNext( new Node( 3 ), 8 );
        testGraph[1].next.next.setNext( new Node( 5 ), -4 );

        testGraph[2].setNext( new Node( 4 ), 1 );
        testGraph[2].next.setNext( new Node( 5 ), 7 );

        testGraph[3].setNext( new Node( 2 ), 4 );

        testGraph[4].setNext( new Node( 3 ), -5 );
        testGraph[4].next.setNext( new Node( 1 ), 2 );

        testGraph[5].setNext( new Node( 4 ), 6 );

        return testGraph;
    }

	/**
 	* Generate a random graph.
 	*
 	* @param n the number of nodes
 	* @param p the probability of a edge connection
 	* @param w the minimum edge weight
 	* @param W the maximum edge weight
 	*
 	* @return the graph in adjacency list form
 	*/
  public static Node [] genRandGraph(int n, double p, double w, double W ) {
  	Random rand = new Random( SEED );
  	Node randGraph [] = new Node[n + 1];
  	Node current;
  	if( INT_ONLY ) {
	  	for( int i = 1; i <= n; i++ ) {
	  	    current = randGraph[i] = new Node( i );
	  	    for( int j = 1; j < n; j++ ) {
	  	        if( i + j == n + 1 ) j++;  //adjustment because 0 index not used
	  	        if( p > rand.nextDouble() ) {
	  	            current.setNext( new Node( ( i + j ) % ( n + 1 ) ),
	  	                            (double)((int) (
	  	                            ( W - w ) * rand.nextDouble() + w  ) ) );
	  	            current = current.next;
	  	        }
	  	    }
	  	}
  }
  else {
  	for( int i = 1; i <= n; i++ ) {
	  	    current = randGraph[i] = new Node( i );
	  	    for( int j = 1; j < n; j++ ) {
	  	        if( i + j == n + 1 ) j++;  //adjustment because 0 index not used
	  	        if( p > rand.nextDouble() ) {
	  	            current.setNext( new Node( ( i + j ) % ( n + 1 ) ),
	  	                            ( W - w ) * rand.nextDouble() + w  ) ;
	  	            current = current.next;
	  	        }
	  	    }
	  	}
  }
  	return randGraph;
  }

  /**
 	* Generate a random graph.
 	*
 	* @param n the number of nodes
 	* @param p the probability of a edge connection
 	* @param w the minimum edge weight
 	* @param W the maximum edge weight
 	*
 	* @return the graph in adjacency list form
 	*/
  public static Node [] genRandGraphMaxOrMinOnly(int n, double p, double w, double W ) {
  	Random rand = new Random( SEED );
  	Node randGraph [] = new Node[n + 1];
  	Node current;

  	for( int i = 1; i <= n; i++ ) {
	  	    current = randGraph[i] = new Node( i );
	  	    for( int j = 1; j < n; j++ ) {
	  	        if( i + j == n + 1 ) j++;  //adjustment because 0 index not used
	  	        if( p > rand.nextDouble() ) {
	  	            current.setNext( new Node( ( i + j ) % ( n + 1 ) ),
	  	                            rand.nextDouble()< .5 ? w : W  ) ;
	  	            current = current.next;
	  	        }
	  	    }
	  	}
  	return randGraph;
  }

	/**
 	* Generate a random Euclidian graph.
 	*
 	* @param n gives the dimension for an n x n graph
 	* @param p the probability of a edge connection
 	*
 	* @return the graph in adjacency list form
 	*/
	public static EuclidNode [] genRandEuclid( int n, double p, int numSource ) {
		int row, col, x, y;
		Random rand = new Random( SEED );
		int size = (int)Math.pow( n, 2 );
		System.out.println("n = " + n + " size = " + size);
		EuclidNode randEuclid [] = new EuclidNode[size + 1];
		Node current;
		for( int i = 1; i <= size; i++) {
			row = ( ( i - 1 ) / n ) + 1;
			col = ( ( i - 1 ) % n ) + 1;
			current = randEuclid[i] = new EuclidNode( i, row, col);
			for( int j = -1; j < 2; j++ ) {
				x = row + j;
				for( int k = -1; k < 2; k++ ) {
					y = col + k;
					if( x > 0 && y > 0 && x <= n && y <= n &&
						!( x == row && y == col ) ) {
						if( p > rand.nextDouble() ) {
							current.setNext( new EuclidNode( ( ( x - 1 ) * n ) + y,
								x, y ), 1 );
							current = current.next;
						}
					}
				}
			}
		}

		//set relevant nodes for the many pairs problem
		boolean done = false;
		int totalChosen = 0;
		while( !done ) {

			for( int i = 1; i <= size && !done; i++) {
				if( ( ((double)numSource) / size ) > rand.nextDouble() ) {

					if(	!randEuclid[i].relevantNode ) {
							randEuclid[i].relevantNode = true;
							totalChosen++;
							//System.out.println(totalChosen + " " + i);
							if ( totalChosen == numSource )
								done = true;
					}
				}
			}
		}
		return randEuclid;
	}

	/**
 	* Convert an adjacency list graph to adjacency matrix form.
 	*
 	* @param graph the graph in adjacency list form
 	*
 	* @return the graph in adjacency matrix form
 	*/
	public static double[][] convertToMatrix( Node graph[] ) {
		double[][] graphAM = new double[graph.length][graph.length];
		for( int i = 1; i < graph.length; i++ )
			for( int j = 1; j < graph.length; j++ )
				if( i == j )
					graphAM[i][j] = 0;
				else
					graphAM[i][j] = Double.MAX_VALUE;
		Node current;
		for( int i = 1; i < graph.length; i++ ) {
			if( graph[i] != null ) {
				current = graph[i];
				while( current.next != null ) {
					graphAM[graph[i].id][current.next.id] = current.edgeWeight;
					current = current.next;
				}
			}
		}
		return graphAM;
	}

	/**
 	* Print an adjacency matrix graph.
 	*
 	* @param graphAM the grpah to print out
 	*/
	public static void printMatrix( double[][] graphAM ) {
		for( int i = 1; i < graphAM.length; i++ ) {
			System.out.print("( ");
			for( int j = 1;j < graphAM.length; j++ )
				if( graphAM[i][j] == Double.MAX_VALUE )
					System.out.print( "INF " );
				else
					System.out.print( graphAM[i][j] + " " );
			System.out.print(") ");
			System.out.println();
		}
	}

	/**
 	* Print an adjacency list graph.
 	*
 	* @param graph the grpah to print out
 	*/
	public static void printGraph( Node graph[] ) {
		System.out.println( "Number of nodes: " + ( graph.length - 1 ) );
		for( int i = 0; i < graph.length; i++ ) {
			if( graph[i] != null ) {
				Node current = graph[i];
				if( EUCLID )
					System.out.print( current.id + "(" + ((EuclidNode)current).x +
						", " + ((EuclidNode)current).y + ")" );
				else
					System.out.print( current.id );
				while( current.next != null ) {
					System.out.print( " -" + current.edgeWeight + "-> " );
					current = current.next;
					if( EUCLID )
						System.out.print( current.id + "(" + ((EuclidNode)current).x +
							", " + ((EuclidNode)current).y + ")" );
					else
						System.out.print( current.id );
				}
				System.out.println();
			}
		}
	}
}
