/*
 * function [Isub] = addpaths(V, Vi, R, tol, layout, blockR)
 *
 * Choose columns of Vi to extend V and update the Cholesky factor R.
 * tol is the drop tolerance; layout is 'full' or 'packed', depending
 * whether the block packed storage format is used; and blockR is
 * the size of the blocking factor.
 */

#include <mex.h>

/*
 * SUBROUTINE TRTRS( UPLO, TRANS, DIAG, N, NRHS, A, LDA,  B,
 *                   LDB, INFO )
 *
 * LAPACK routine -- perform a triangular solve.
 */
void dtrtrs_(char* uplo, char* trans, char* diag,
             int* n, int* nrhs,
             double* a, int* lda,
             double* b, int* ldb,
             int* info);

/*
 * SUBROUTINE GEMM ( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA,
 *                   B, LDB, BETA, C, LDC )
 *
 * BLAS routine -- perform a matrix-matrix multiplication.
 */
void dgemm_(char* transA, char* transB, 
            int* m, int* n, int* k, double* alpha,
            double* A, int* ldA,
            double* B, int* ldB, double* beta, 
            double* C, int* ldC);

/*
 * SUBROUTINE GEQP3( M, N, A, LDA, JPVT, TAU,  WORK,  LWORK, INFO )
 * 
 * LAPACK routine -- QR with column pivoting.  Level 3 BLAS.
 */
void dgeqp3_(int* m, int* n, double* a, int* lda, int* jpvt, double* tau,
             double* work, int* lwork, int* info);

/*
 * SUBROUTINE POTRF( UPLO, N, A, LDA, INFO )
 *
 * LAPACK routine -- Cholesky factorization
 */
void dpotrf_(char* uplo, int* n, double* A, int* ldA, int* info);


/* Compute overlap for two path vectors
 *  p1 -- l1 edge numbers in sorted order
 *  p2 -- l2 edge numbers in sorted order
 */
static int path_overlap(int* p1, int l1, 
                        int* p2, int l2)
{
    int i1      = 0;
    int i2      = 0;
    int overlap = 0;

    while (i1 < l1 && i2 < l2) {
        if (p1[i1] < p2[i2]) {
            ++i1;
        } else if (p1[i1] > p2[i2]) {
            ++i2;
        } else {
            ++overlap;
            ++i1;
            ++i2;
        }
    }

    return overlap;
}


/* Compute overlap contributions:
 *   R(1:k,     k+1:k+m) = V' *Vi
 *   R(k+1:k+k, k+1:k+m) = Vi'*Vi
 */
static void compute_overlaps(double* R12, double* R22, 
                             int k, int* irV,  int* jcV,
                             int m, int* irVi, int* jcVi)
{
    int i, j;
    
    for (j = 0; j < m; ++j) {
        for (i = 0; i < k; ++i) {
            R12[j*k+i] = path_overlap(irV  + jcV [i], jcV [i+1] - jcV [i],
                                      irVi + jcVi[j], jcVi[j+1] - jcVi[j]);
        }
        for (i = 0; i < m; ++i) {
            R22[j*m+i] = path_overlap(irVi + jcVi[i], jcVi[i+1] - jcVi[i],
                                      irVi + jcVi[j], jcVi[j+1] - jcVi[j]);
        }
    }
}


/*
 * Blocked backsolve with R^T
 */
static void block_trs(double* R, int k, int kb, double* b, int m)
{
    int n, J, info;
    double one     =  1;
    double neg_one = -1;

    for (J = 0; J <= (k-1)/kb; ++J) {

        int Jkb    = J*kb;
        int Jkbp   = Jkb + kb;
        int blockn = (Jkbp < k) ? kb : (k - Jkb);

        dgemm_("T", "N", &blockn, &m, &Jkb,
               &neg_one,
               R    + (Jkb*Jkbp)/2,  &Jkbp,
               b,                    &k,
               &one,
               b    + Jkb,           &k);
        
        dtrtrs_("U", "T", "N", &blockn, &m,
                R    + Jkb + (Jkb*Jkbp)/2,  &Jkbp,
                b    + Jkb,                 &k,
                &info);
    }
}


/* Choose which vectors we should keep via QR with column pivots
 */
static int add_paths(int* jpvt, double* R, int ldR,
                     int k, int* irV,  int* jcV,
                     int m, int* irVi, int* jcVi,
                     double tol, int kb)
{

    int i, j, J, nchosen, lwork, info;
    double twork;
    double* work;
    double* scratch;
    double* R12t;
    double* R22t;
    double* a;
    double neg_one = -1;
    double one     =  1;

    /* 
     * Workspace query and allocations:
     *   k * m     -- scratch for R12
     *   m * m     -- scratch for R22
     *   m * m     -- scratch for R22 factorization
     *   lwork + m -- scratch for pivoted QR
     */
    lwork = -1;
    dgeqp3_(&m, &m, NULL, &m, NULL, NULL, &twork, &lwork, &info);
    lwork = (int) twork;

    scratch = mxMalloc((lwork + m + (k+m+m)*m) * sizeof(double));
    R12t    = scratch;
    R22t    = R12t + k*m;
    a       = R22t + m*m;
    work    = a    + m*m;

    /* Extend R with columns for V'*Vi and Vi'*Vi
     */
    compute_overlaps(R12t, R22t,
                     k, irV,  jcV,
                     m, irVi, jcVi);

    /* Do the first step of Cholesky updating:
     *   R12 := R11^-T R12
     *   R22 := R22 - R12'*R12
     */
    if (k != 0) {

        if (kb)
            block_trs(R, k, kb, R12t, m);
        else
            dtrtrs_("U", "T", "N", &k, &m, R, &ldR, R12t, &k, &info);

        dgemm_("T", "N", &m, &m, &k, 
               &neg_one,
               R12t, &k,
               R12t, &k, 
               &one,
               R22t, &m);

    }

    /* Pivoted QR factorization on copy of R22 */
    memset(jpvt, 0, m  *sizeof(int));
    memcpy(a, R22t, m*m*sizeof(double));
    dgeqp3_(&m, &m, a, &m, jpvt, work, work + m, &lwork, &info);

    /* Mark selected columns */
    nchosen = 0;
    for (i = 0; i < m; ++i) {
        if (a[i*(m+1)] > tol || a[i*(m+1)] < -tol) {
            jpvt[nchosen++] = jpvt[i]-1;
        }
    }

    /* Subset chosen rows and columns of R22 into a */
    for (i = 0; i < nchosen; ++i)
        for (j = 0; j < nchosen; ++j)
            a[j*m + i] = R22t[jpvt[j]*m + jpvt[i]];

    /* Run Cholesky on the trailing nchosen-by-nchosen block */
    dpotrf_("U", &nchosen, a, &m, &info);

    /* Copy R12 and R22 blocks into place */
    if (kb) {

        /* Packed version */
        for (j = 0; j < nchosen; ++j) {

            int J    = (k+j)/kb;
            int Jkb  = J*kb;
            int Jkbp = Jkb + kb;
            
            for (i = 0; i < k; ++i)
                R[i   + ((2*(k+j) - Jkb)*Jkbp) / 2] = R12t[jpvt[j]*k + i];
            for (i = 0; i <= j; ++i)
                R[k+i + ((2*(k+j) - Jkb)*Jkbp) / 2] = a[j*m + i];
        }

    } else {

        /* Unpacked version */
        for (j = 0; j < nchosen; ++j) {
            for (i = 0; i < k; ++i)
                R[i   + (k+j)*ldR] = R12t[jpvt[j]*k + i];
            for (i = 0; i <= j; ++i)
                R[k+i + (k+j)*ldR] = a[j*m + i];
        }

    }       

    mxFree(scratch);
    return nchosen;
}


void mexaddpathsd(int nlhs, mxArray** plhs,
                  int nrhs, const mxArray** prhs)
{
    int     i, k, m, nchosen, ldR;
    int     is_full;
    int     blockR;
    int*    jpvt;
    int*    irV;
    int*    jcV;
    int*    irVi;
    int*    jcVi;
    double*  R;
    double   tol;
    char    stmp[15];

    if (nrhs != 6)
        mexErrMsgTxt("Wrong number of arguments");

    if (!mxIsSparse(prhs[0]) && mxGetN(prhs[0]) != 0)
        mexErrMsgTxt("U must be sparse");
    if (!mxIsSparse(prhs[1]))
        mexErrMsgTxt("V must be sparse");
    if (!mxIsNumeric(prhs[2]))
        mexErrMsgTxt("R must be dense");

    k    = mxGetN (prhs[0]);
    irV  = mxGetIr(prhs[0]);
    jcV  = mxGetJc(prhs[0]);
    
    m    = mxGetN (prhs[1]);
    irVi = mxGetIr(prhs[1]);
    jcVi = mxGetJc(prhs[1]);

    R    = (double*) mxGetPr(prhs[2]);
    tol  = (double) *mxGetPr(prhs[3]);

    mxGetString(prhs[4], stmp, sizeof(stmp));
    is_full = ( strcmp(stmp, "full") == 0 );

    ldR    = mxGetM(prhs[2]);
    blockR = (int) *mxGetPr(prhs[5]);
   
    jpvt = (int*) mxMalloc(m * sizeof(int));

    nchosen = is_full ? 
        add_paths(jpvt, R, ldR, k, irV,  jcV, m, irVi, jcVi, tol, 0) :
        add_paths(jpvt, R, ldR, k, irV,  jcV, m, irVi, jcVi, tol, blockR);

    plhs[0] = mxCreateDoubleMatrix(nchosen,1, mxREAL);
    for (i = 0; i < nchosen; ++i) {
        mxGetPr(plhs[0])[i] = jpvt[i]+1;
    }
    mxFree(jpvt);
}
