package cluster;

import java.util.*;
import java.io.*;

/*************************************************************************************
* <CODE>Cluster</CODE> is used for representing clusters of multiple hosts with one host
* as the center. This class extends <CODE>HashSet</CODE>.
*
* @author Chris Karlof
*   <A HREF="mailto:ckarlof@eecs.berkeley.edu"> (ckarlof@eecs.berkeley.edu) </A>
**************************************************************************************/
 
public class Cluster implements Serializable, WNMMSConstants
{
    // center of cluster
    protected Host center;
    protected HashSet cluster;
    protected double avgRTTFromCenter;
    protected double maxRTTFromCenter;
    protected double avgRTTAcrossCluster;
    protected double maxRTTAcrossCluster;
    protected double avgNDistanceFromCenter;
    protected HashMap avgRTTMap;
    protected HashMap triangulationMap;

    // this isnt set until error calculation
    private HostMap hostmap = null;

    /**
    * Constructs a <CODE>Cluster</CODE> with the given center.
    *
    * @param center
    *   the <CODE>Host</CODE> at the center of this <CODE>Cluster</CODE>
    **/

    public Cluster(Host center)
    {
        cluster = new HashSet();
        cluster.add(center);

        this.center = center;
	avgRTTFromCenter = maxRTTFromCenter = avgRTTAcrossCluster = maxRTTAcrossCluster = -1.0;
	avgNDistanceFromCenter = -1.0;
	avgRTTMap = new HashMap();
	triangulationMap = new HashMap();
    }

    public Cluster()
    {
	cluster = new HashSet();
	center = null;
	avgRTTFromCenter = maxRTTFromCenter = avgRTTAcrossCluster = maxRTTAcrossCluster = -1.0;
	avgNDistanceFromCenter = -1.0;
	avgRTTMap = new HashMap();
	triangulationMap = new HashMap();
    }

    /**
    * Returns a string representation of this <CODE>Cluster</CODE>.
    *
    * @return
    *   a string representing all the hosts in this <CODE>Cluster</CODE>
    **/
    public String toString()
    {
        StringBuffer result = new StringBuffer();
        Iterator iterator = cluster.iterator();

        result.append(center + ": ");

        while(iterator.hasNext())
        {
            result.append(((Host) iterator.next()) + " ");
        }

	/*
	//if(avgRTTFromCenter > 0)
	result.append("\nAvgRTTFromCenter: " + avgRTTFromCenter);
	//if(maxRTTFromCenter > 0)
	result.append("\nMaxRTTFromCenter: " + maxRTTFromCenter);
	//if(avgRTTAcrossCluster > 0)
	result.append("\nAvgRTTAcrossCluster: " + avgRTTAcrossCluster);
	//if(maxRTTAcrossCluster > 0)
	result.append("\nMaxRTTAcrossCluster: " + maxRTTAcrossCluster);
	//if(avgNDistanceFromCenter > 0)
	result.append("\nAvgNDistanceFromCenter: " + avgNDistanceFromCenter);
	*/
        
	return result.toString();
    }

    /**
    * Returns the host at the center of this <CODE>Cluster</CODE>.
    *
    * @return
    *   the center of this <CODE>Cluster</CODE>
    **/
    public Host getCenter()
    {
        return center;
    }

    public void setCenter(Host center)
    {
	this.center = center;
    }

    public void setAvgRTTFromCenter(double avg)
    {
	avgRTTFromCenter = avg;
    }

    public double getAvgRTTFromCenter()
    {
	return avgRTTFromCenter;
    }

    public double getAvgRTTAcrossCluster()
    {
	return avgRTTAcrossCluster;
    }

    public void setAvgNDistanceFromCenter(double avg)
    {
	avgNDistanceFromCenter = avg;
    }

    public double getAvgNDistanceFromCenter()
    {
	return avgNDistanceFromCenter;
    }

    public void minimizeAvgNDistanceToFindCenter(double[][] eucMatrix)
    {
	double best = -1.0;
	Host candidate = null;

	Iterator iterator1 = iterator();
	
	while(iterator1.hasNext())
	{
	    Host seed = (Host) iterator1.next();
	    Iterator iterator2 = iterator();
	    double total = 0.0;
	    int count = 0;
	    while(iterator2.hasNext())
	    {
		Host other = (Host) iterator2.next();
		if (seed.getNum() == other.getNum())
		    continue;
		else
		{
		    count++;
		    total += eucMatrix[seed.getNum()][other.getNum()];
		}
	    }
	    
	    if(candidate == null)
	    {
		best = total/count;
		candidate = seed;
	    }
	    else if(total/count < best)
	    {
		best = total/count;
		candidate = seed;
	    }
	}

	setAvgNDistanceFromCenter(best);
	setCenter(candidate);
	// calcRTTInfo(rttMatrix);
    }

    public void minimizeGeoDistanceToFindCenter(double[][] euclidian_matrix,HostMap hostMap,RTTMatrix rttMatrix)
    {
	double best = -1.0;
	Host candidate = null;

	Iterator iterator1 = iterator();
	
	while(iterator1.hasNext())
	{
	    Host seed = (Host) iterator1.next();
	    Iterator iterator2 = iterator();
	    double total = 0.0;
	    int count = 0;
	    while(iterator2.hasNext())
	    {
		Host other = (Host) iterator2.next();
		double distance = euclidian_matrix[hostMap.hash(seed)][hostMap.hash(other)];
		if(distance < 0.001)
		    continue;
		else
		{
		    count++;
		    total += distance;
		}
	    }
	    
	    if(candidate == null)
	    {
		best = total/count;
		candidate = seed;
	    }
	    else if(total/count < best)
	    {
		best = total/count;
		candidate = seed;
	    }
	}

	setCenter(candidate);
	calcAvgNDistance(rttMatrix);
	calcRTTInfo(rttMatrix);
    }

    public void minimizeAvgVectorSimilarityToFindCenter(double[][] vector_sim_matrix,HostMap hostMap,RTTMatrix rttMatrix)
    {
	double best = -1.0;
	Host candidate = null;

	Iterator iterator1 = iterator();
	
	while(iterator1.hasNext())
	{
	    Host seed = (Host) iterator1.next();
	    Iterator iterator2 = iterator();
	    double total = 0.0;
	    int count = 0;
	    while(iterator2.hasNext())
	    {
		Host other = (Host) iterator2.next();
		double distance = vector_sim_matrix[hostMap.hash(seed)][hostMap.hash(other)];
		if(distance < 0.001)
		    continue;
		else
		{
		    count++;
		    total += distance;
		}
	    }
	    
	    if(candidate == null)
	    {
		best = total/count;
		candidate = seed;
	    }
	    else if(total/count < best)
	    {
		best = total/count;
		candidate = seed;
	    }
	}

	setCenter(candidate);
	calcAvgNDistance(rttMatrix);
	calcRTTInfo(rttMatrix);
    }

    public void minimizeAvgRTTToFindCenter(RTTMatrix rttMatrix)
    {
	double best = -1.0;
	Host candidate = null;

	Iterator iterator1 = iterator();
	
	while(iterator1.hasNext())
	{
	    Host seed = (Host) iterator1.next();
	    Iterator iterator2 = iterator();
	    double total = 0.0;
	    int count = 0;
	    while(iterator2.hasNext())
	    {
		Host other = (Host) iterator2.next();
		double rtt = rttMatrix.get(seed,other);
		if(rtt < 0.001)
		    continue;
		else
		{
		    count++;
		    total += rtt;
		}
	    }
	    
	    if(candidate == null)
	    {
		best = total/count;
		candidate = seed;
	    }
	    else if(total/count < best)
	    {
		best = total/count;
		candidate = seed;
	    }
	}

	setAvgRTTFromCenter(best);
	setCenter(candidate);
	calcAvgNDistance(rttMatrix);
    }

    public void calcInfo(RTTMatrix rttMatrix)
    {
	if(center != null)
	    calcAvgNDistance(rttMatrix);
	calcRTTInfo(rttMatrix);
    }

    public double findIntraClusterEstimate(String date, Host host1,
					   Host host2, int index, String method,
					   PrintWriter errorStream)
    { 
	double result = -1;
	// this line determines intra cluster estimation
	// we have 4 ways: historical avgRTT, dynamic RTT, subclustering, triangulation
	// switches to default if method not found
	// note that subclustering not an option here
	if(method.equals("historicalAvgRTT"))
	{
	    result = getAvgRTTFromCenter();
	}
	else if(method.equals("dynamicAvgRTT"))
	{
	    if(!avgRTTMap.containsKey(date))
			calcIntraAvgRTT(date,errorStream);
	    result = ((double[]) avgRTTMap.get(date))[index];
	}
	else if(method.equals("triangulation"))
	{
	    if(!triangulationMap.containsKey(date))
		calcIntraTriangulation(date,errorStream);
	    double[][] matrix = (double[][]) triangulationMap.get(date);
	    double leg1 = matrix[index][hostmap.hash(host1)];
	    double leg2 = matrix[index][hostmap.hash(host2)];
	    if(leg1 < 0 || leg2 < 0)
		result = -1;  // this indicates "loss" on one of the legs
	    else
		result = (leg1 + leg2)/2;
	}
	else
	{
	    System.out.println("warning switching to default");
	    result = findIntraClusterEstimate(date,host1,host2,index,DEFAULT_INTRA_ESTIMATION_METHOD,errorStream);
	}
		 
	return result;
    }

    private void calcIntraTriangulation(String comparisonDate, PrintWriter errorStream)
    {
	String baseFilename = ".." + SLASH + "data" + SLASH + "raw-data" +
	    SLASH + comparisonDate + SLASH + center.getName() + MIDDLE;
	double[][] result = new double[NUMBER_OF_READINGS][size()];
	hostmap = new HostMap(this);

	// initialize to negative distance. this will indicate loss on either leg
	for(int i=0;i<NUMBER_OF_READINGS;i++)
	    for(int j=0;j<size();j++)
		    result[i][j] = -1;

	Iterator iterator = iterator();
	while(iterator.hasNext())
	{
	    Host host = (Host) iterator.next();
	    try
	    {
		String filename = baseFilename + host.getName() + SLASH + comparisonDate;
		BufferedReader reader = new BufferedReader(
		       new InputStreamReader(new FileInputStream(new File(filename))));
		String input;
		int index = 0;
		while((input = reader.readLine()) != null && index < NUMBER_OF_READINGS)
		{
		    if(input.trim().equals(""))
		    {
			System.out.println("Error blank line in "+filename);
			index++;
			continue;
		    }
		    
		    String value = input.substring(input.indexOf(' ')+1);
		    if(value.equals("loss"))
			index++;
		    else
		    {
			result[index][hostmap.hash(host)] = (new Double(value)).doubleValue();
			index++;
		    }
		}	    
	    }

	    catch(IOException e)
	    {
		if(e instanceof FileNotFoundException)
		{
		    errorStream.println("Doing triangulation calc: "+center+" to "+host);
		    errorStream.println(e.getMessage());
		}
		else
		{
		    errorStream.println("File format exception.\n" + e.getMessage());
		    System.exit(0);
		}
	    }
	}

	triangulationMap.put(comparisonDate,result);
    }

    private void calcIntraAvgRTT(String comparisonDate, PrintWriter errorStream)
    {
        String baseFilename = ".." + SLASH + "data" + SLASH + "raw-data" +
	    	SLASH + comparisonDate + SLASH + center.getName() + MIDDLE;
		double[] result = new double[NUMBER_OF_READINGS];
		int[] count = new int[NUMBER_OF_READINGS];

		for(int i=0;i<NUMBER_OF_READINGS;i++)
        {
	    	count[i] = 0;
	    	result[i] = -0.0000000001;
		}

		Iterator iterator = cluster.iterator();
		while(iterator.hasNext()) {
	    	Host host = (Host) iterator.next();
	    	if(host.equals(center)) 	continue;
	    
	    	try {
				String filename = baseFilename + host.getName() + SLASH + comparisonDate;
				File file = new File(filename);
				BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(file)));
	    
				int index = 0;
				String input;
				while((input = reader.readLine()) != null && index < NUMBER_OF_READINGS)
				{
		    		if(input.trim().equals("")) {
						System.out.println("Error blank line in "+filename);
						index++;
						continue;
		    		}
		    
		    		String value = input.substring(input.indexOf(' ')+1);
		    		if(value.equals("loss")) {
						index++;
						continue;
		    		}

		    		result[index] += (new Double(value)).doubleValue();
		    		count[index]++;
		    		index++;
				}
				reader.close();
	    	}
	    	catch(IOException e) {
				if(e instanceof FileNotFoundException) {
		    		errorStream.println("Doing dynamicIntraRTT calc: "+center+" to "+host);
		    		errorStream.println(e.getMessage());
				} else {
		    		errorStream.println("File format exception.\n" + e.getMessage());
		    		System.exit(0);
				}
	    	}
		}
	
		for(int i=0;i<NUMBER_OF_READINGS;i++) {
	    	if(count[i] == 0) 	continue; 
			result[i] = result[i]/count[i];
		}
        avgRTTMap.put(comparisonDate,result);	       
    }

    private void calcAvgNDistance(RTTMatrix rttMatrix)
    {
	Iterator iterator = cluster.iterator();

	double total = 0.0;
	int count = 0;

	while(iterator.hasNext())
	{
	    Host site = (Host) iterator.next();
	    total += rttMatrix.squareEuclidianDistance(center,site).getDistance();
	    count++;
	}

	setAvgNDistanceFromCenter(total/count);
    }

    private void calcRTTInfo(RTTMatrix rttMatrix)
    {
	if(center != null)
	    calcCenterBasedRTTInfo(rttMatrix);
	calcClusterBasedRTTInfo(rttMatrix);
    }

    private void calcCenterBasedRTTInfo(RTTMatrix rttMatrix)
    {
	// first calculate radius information
        Iterator iterator = cluster.iterator();

	double total = 0.0;
	double max = 0.0;
	int count = 0;

        while(iterator.hasNext())
        {
            Host site = (Host) iterator.next();
	    double dist = rttMatrix.get(center,site);
	    if(dist > max)
		max = dist;
	    if(dist < 0.001)
		continue;
	    else
	    {
		total += dist;
		count++;
	    }
	}

	avgRTTFromCenter = total/count;
	maxRTTFromCenter = max;
    }

    private void calcClusterBasedRTTInfo(RTTMatrix rttMatrix)
    {
	double total = 0.0;
	double max = 0.0;
	int count = 0;
	Iterator iterator = cluster.iterator(); 
        
	while(iterator.hasNext())
        {
	    Host source = (Host) iterator.next();
	    Iterator iterator2 = cluster.iterator();
	    while(iterator2.hasNext())
	    {
		Host site = (Host) iterator2.next();
		double dist = rttMatrix.get(source,site);
		if(dist > max)
		    max = dist;
		if(dist < 0.001)
		    continue;
		else
		{
		    total += dist;
		    count++;
		}
	    }
	}

	avgRTTAcrossCluster = total/count;
	maxRTTAcrossCluster = max;
    }

	/** 
	 *	Determine whether src1 match with other cluster hosts in term of correlation to dest
	 *	It has to match with every existing hosts
	 **/
    public boolean matchDest(int src1, int dest, double[][][] corCoefPaths)
    {
		Iterator iterator = cluster.iterator(); 
        
		while(iterator.hasNext())
        {
	    	int src2 = ((Host)iterator.next()).getNum();
			if (corCoefPaths[src1][src2][dest] < 0.8) 	return false;
		}
		return true;
	}

	/** 
	 *	Determine whether dest1 match with other cluster hosts in term of correlation to 
	 *	src. It has to match with every existing hosts
	 **/
    public boolean matchSrc(int dest1, int src, double[][][] corCoefPaths)
    {
		Iterator iterator = cluster.iterator(); 
        
		while(iterator.hasNext())
        {
	    	int dest2 = ((Host)iterator.next()).getNum();
			if (corCoefPaths[src][dest1][dest2] < CorCoefPathClusteringThre) return false;
		}
		return true;
	}

    /**
    * Returns an iterator over the hosts in this <CODE>Cluster</CODE>.
    *
    * @return
    *   an iterator over the hosts in this <CODE>Cluster</CODE>
    **/
    public Iterator iterator()
    {
        return cluster.iterator();
    }

    /**
    * Determines if the given hosts is contained in this <CODE>Cluster</CODE>.
    * @param host
    *   a <CODE>Host</CODE> whose presence is to be tested
    * @return
    *   true if the <CODE>host</CODE> is contained in this <CODE>Cluster</CODE>,
    *     false otherwise
    **/
    public boolean contains(Host host)
    {
        return cluster.contains(host);
    }

    /**
    * Adds the given host to this <CODE>Cluster</CODE>.
    * @param host
    *   a <CODE>Host</CODE> to be added to this <CODE>Cluster</CODE>
    **/ 
    public void add(Host host)
    {
        cluster.add(host);
    }

    /**
    * Returns the number hosts in this <CODE>Cluster</CODE>.
    *
    * @return
    *   the number of hosts in  this <CODE>Cluster</CODE>
    **/
    public int size()
    {
        return cluster.size();
    }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
	out.writeObject(center);
	out.writeObject(cluster);
	out.writeDouble(avgRTTFromCenter);
	out.writeDouble(maxRTTFromCenter);
	out.writeDouble(avgRTTAcrossCluster);
	out.writeDouble(maxRTTAcrossCluster);
	out.writeDouble(avgNDistanceFromCenter);
	out.writeObject(avgRTTMap);
	out.writeObject(triangulationMap);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
	center = (Host) in.readObject();
	cluster = (HashSet) in.readObject();
	avgRTTFromCenter = in.readDouble();
	maxRTTFromCenter = in.readDouble();
	avgRTTAcrossCluster = in.readDouble();
	maxRTTAcrossCluster = in.readDouble();
	avgNDistanceFromCenter = in.readDouble();
	avgRTTMap = (HashMap) in.readObject();
	triangulationMap = (HashMap) in.readObject();
    }

}













