package cluster;

import java.io.*;
import java.util.*;

/*************************************************************************************
* The <CODE>RTTMatrix</CODE> class is used to represent round trip times between a set of hosts.
*
* @author Chris Karlof
*   <A HREF="mailto:ckarlof@eecs.berkeley.edu"> (ckarlof@eecs.berkeley.edu) </A>
**************************************************************************************/

public class GeoDistanceMatrix implements Serializable
{
    // holds latency measurements for all hosts
    private double[][] distanceMatrix;

    // number of hosts
    private int hosts;

    // description of values (gmean, amean)
    private String description;

    // datafile used
    private String datafile;

    // a mapping of integers to actual hosts
    private HostMap hostMap;

    /**
    * Creates a new <CODE>RTTMatrix</CODE> using the data in the <CODE>dataFile</CODE>.
    * @param dataFile
    *   the filename of the file containing the round-trip time data
    * @param numberOfHosts
    *   the number of hosts is the file <CODE>dataFile</CODE>
    * @param description
    *   the type of aggregation done on the raw data (gmean, amean, etc)
    **/
    public GeoDistanceMatrix(String datafile, HostMap hostMap, String description, int hosts)
    {
        this.hosts = hosts;
        this.distanceMatrix = new double[hosts][2];
        this.description = description;
	this.hostMap = hostMap;
	this.datafile = datafile;

        try
        {
            BufferedReader in = new BufferedReader(new InputStreamReader(
                new FileInputStream(new File(datafile))));

            for(int i=0;i<hosts;i++)
            {
                String line = in.readLine();
                StringTokenizer st = new StringTokenizer(line);

		st.nextToken();
		String latitude  = st.nextToken();
		String longitude = st.nextToken();
		try {
		    distanceMatrix[i][0] = (new Double(latitude)).doubleValue();
		    distanceMatrix[i][1] = (new Double(longitude)).doubleValue();
		} catch(ArrayIndexOutOfBoundsException e) { System.out.println("error in datafile: "+datafile); System.exit(0); }
      
	    }

            in.close();
        }

        catch(IOException e)
        {
            e.printStackTrace();
        }
    }

    public int getNumberOfHosts()
    {
	return hosts;
    }

    public HostMap getHostMap()
    {
	return hostMap;
    }

    /**
    * Gets the filename of the file used to create this <CODE>RTTMatrix</CODE>.
    * @return
    *   the filename containing the data
    **/
    public String getDatafile()
    {
        return datafile;
    }

    /**
    * Calculates the n-space distance between the hosts <CODE>h1</CODE> and <CODE>h2</CODE>.
    * @param h1
    *   a host
    * @param h2
    *   a host
    * @return
    *   a <CODE>Distance</CODE> object representing the n-space distance between the two hosts
    **/
    public Distance squareEuclidianDistance(Host h1, Host h2)
    {
        int i = hostMap.hash(h1);
        int j = hostMap.hash(h2);

	double sum = Math.pow(distanceMatrix[i][0]-distanceMatrix[j][0],2) +
	    Math.pow(distanceMatrix[i][1]-distanceMatrix[j][1],2);

        return new Distance(sum,2);
    }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
        out.writeObject(distanceMatrix);
	out.writeInt(hosts);
	out.writeObject(description);
	out.writeObject(datafile);
	out.writeObject(hostMap);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
        distanceMatrix = (double[][]) in.readObject();
	hosts = in.readInt();
	description = (String) in.readObject();
	datafile = (String) in.readObject();
	hostMap = (HostMap) in.readObject();
    }
}















