package cluster;

import java.io.*;
import java.util.*;

/*************************************************************************************
* The <CODE>Histogram</CODE> class is used to represent histograms of the total,
* inter-cluster, and intra-cluster relative errors 
* contained in an associated <CODE>RelativeErrorMatrix</CODE>. This class also calculates
* cumulative distribution functions of total, intra-cluster, and inter-cluster relative errors.
*
* @author Chris Karlof
*   <A HREF="mailto:ckarlof@eecs.berkeley.edu"> (ckarlof@eecs.berkeley.edu) </A>
**************************************************************************************/

public class Histogram implements Serializable
{
    /**
    * This field indicates the range of our histogram and CDF. Our histogram will be 
    * calculated on this many buckets, each of range <CODE>0.01</CODE>.
    **/
    public static final int NUMBER_OF_BUCKETS = 20;
    public static final int SCALE_FACTOR = 10;
    private int[] histogram;

    private int[] interHistogram;
    private int[] intraHistogram;

    private double[] cdf;

    private double[] interCDF;
    private double[] intraCDF;

    private int numberOfReadings;

    private int intraNumberOfReadings;
    private int interNumberOfReadings;

    
    //added to keep total so that we know what is the 80th and 90th percentile
    private int totalValid;
    private TreeSet ts; 
    

    /**
    * Creates a <CODE>Histogram</CODE> associated with the given
    * <CODE>RelativeErrorMatrix</CODE>.
    * @param m
    *   the <CODE>RelativeErrorMatrix</CODE> to construct the histogram for
    **/
//      public Histogram(RelativeErrorMatrix m, int dummy)
//      {
//  	histogram = new int[NUMBER_OF_BUCKETS + 2];
//  	interHistogram = new int[NUMBER_OF_BUCKETS + 2];
//  	intraHistogram = new int[NUMBER_OF_BUCKETS + 2];

//  	intraNumberOfReadings = 0;
//  	interNumberOfReadings = 0;

//  	int hosts = m.getNumberOfHosts();
//  	HostMap hostMap = m.getHostMap();

//  	for(int i=0;i<hosts;i++)
//  	{
//  	    for(int j=0;j<hosts;j++)
//  	    {
//  		int bucket = (int) (SCALE_FACTOR*(m.getRelativeError(hostMap.getHost(i),
//  								     hostMap.getHost(j))));
//  		if(bucket >= NUMBER_OF_BUCKETS)
//  		{
//  		    histogram[NUMBER_OF_BUCKETS]++;
//  		    if(m.isInterClusterError(i,j))
//  		    {
//  			interHistogram[NUMBER_OF_BUCKETS]++;
//  			interNumberOfReadings++;
//  		    }
//  		    else if(m.isIntraClusterError(i,j))
//  		    {
//  			intraHistogram[NUMBER_OF_BUCKETS]++;
//  			intraNumberOfReadings++;
//  		    }
//  		    else throw new IllegalStateException("Illegal error type.");
//  		}
//  		else if(bucket < 0)
//  		{
//  		    histogram[NUMBER_OF_BUCKETS+1]++;
//  		    if(m.isInterClusterError(i,j))
//  			interHistogram[NUMBER_OF_BUCKETS+1]++;
//  		    else if(m.isIntraClusterError(i,j))
//  			intraHistogram[NUMBER_OF_BUCKETS+1]++;
//  		    else throw new IllegalStateException("Illegal error type.");
//  		}
//  		else
//  		{
//  		    histogram[bucket]++;
//  		    if(m.isInterClusterError(i,j))
//  		    {
//  			interHistogram[bucket]++;
//  			interNumberOfReadings++;
//  		    }
//  		    else if(m.isIntraClusterError(i,j))
//  		    {
//  			intraHistogram[bucket]++;
//  			intraNumberOfReadings++;
//  		    }
//  		    else throw new IllegalStateException("Illegal error type.");
//  		}
//  	    }
//  	}
	
//  	numberOfReadings = hosts*hosts - histogram[NUMBER_OF_BUCKETS + 1];
//  	for(int i=0;i<histogram.length;i++)
//  	    System.out.println("bucket "+i+" : "+histogram[i]);
//  	findCDF(m.getNumberOfHosts());
//      }

    public Histogram(RelativeErrorMatrix m)
    {
	//we will use a treeset to add in doubles to find the 80th and 90th percentile
	ts = new TreeSet(new DoubleComparator());
	

	int hosts = m.getNumberOfHosts();
	HostMap hostMap = m.getHostMap();
	histogram = new int[NUMBER_OF_BUCKETS+1];
	cdf = new double[NUMBER_OF_BUCKETS+1];
	int count=0;
	int errors=0;
	for(int i=0;i<hosts;i++)
	    for(int j=0;j<hosts;j++)
	    {
		double re = m.getRelativeError(hostMap.getHost(i),hostMap.getHost(j));
		if(re < 0)
		{
		    errors++;
		    continue;
		}
		/*
		if(re < 0.01)
		    histogram[0]++;
		if(re <= 0.1)
		    histogram[1]++;
		if(re <= 0.2)
		    histogram[2]++;
		if(re <= 0.3)
		    histogram[3]++;
		if(re <= 0.4)
		    histogram[4]++;
		if(re <= 0.5)
		    histogram[5]++;
		if(re <= 0.6)
		    histogram[6]++;
		if(re <= 0.7)
		    histogram[7]++;
		if(re <= 0.8)
		    histogram[8]++;
		if(re <= 0.9)
		    histogram[9]++;
		if(re <= 1.0)
		    histogram[10]++;
 		if(re <= 1.1)
		    histogram[11]++;
		if(re <= 1.2)
		    histogram[12]++;
		if(re <= 1.3)
		    histogram[13]++;
		if(re <= 1.4)
		    histogram[14]++;
		if(re <= 1.5)
		    histogram[15]++;
		if(re <= 1.6)
		    histogram[16]++;
		if(re <= 1.7)
		    histogram[17]++;
		if(re <= 1.8)
		    histogram[18]++;
		if(re <= 1.9)
		    histogram[19]++;
		if(re <= 2.0)
		    histogram[20]++;
		*/
		
		if(re < 0.01)
		    histogram[0]++;
		if(re <= 0.05)
		    histogram[1]++;
		if(re <= 0.10)
		    histogram[2]++;
		if(re <= 0.15)
		    histogram[3]++;
		if(re <= 0.20)
		    histogram[4]++;
		if(re <= 0.25)
		    histogram[5]++;
		if(re <= 0.30)
		    histogram[6]++;
		if(re <= 0.35)
		    histogram[7]++;
		if(re <= 0.40)
		    histogram[8]++;
		if(re <= 0.45)
		    histogram[9]++;
		if(re <= 0.50)
		    histogram[10]++;
		if(re <= 0.55)
		    histogram[11]++;
		if(re <= 0.60)
		    histogram[12]++;
		if(re <= 0.65)
		    histogram[13]++;
		if(re <= 0.70)
		    histogram[14]++;
		if(re <= 0.75)
		    histogram[15]++;
		if(re <= 0.80)
		    histogram[16]++;
		if(re <= 0.85)
		    histogram[17]++;
		if(re <= 0.90)
		    histogram[18]++;
		if(re <= 0.95)
		    histogram[19]++;
		if(re <= 1.00)
		    histogram[20]++;
		ts.add(new Double(re));
		count++;
	    }
	for(int i=0;i<cdf.length;i++)
	    cdf[i] = ((double) histogram[i])/((double) count);
	totalValid = count;
//  	System.out.println("relative error 'errors': "+errors);
//  	System.out.println("number of readings: "+count);
    }

        public double getPercentile(int p) {
	Iterator it = ts.iterator();
	int count = 0;
	int obj =  (p * totalValid)/ 100;
	
	while (it.hasNext()) {
	   
	    Double d = (Double)it.next();
	    if (count == obj) 
		return d.doubleValue();
	    
	    
	    
	    count++;
	}
	return 0.0; //indication of error
    }
    
    

    /**
    * Returns a list of coordinates representing the CDF of all relative errors. This list 
    * is acceptable input to a plotting tool such as Gnuplot.
    * @return
    *   a list of coordinates representing the CDF of all relative errors
    **/

    public String getCDF()
    {
	StringBuffer result = new StringBuffer();
	
	for(int i=0;i<NUMBER_OF_BUCKETS+1;i++)
	    result.append(Format.format(i/20.0,2) + "\t" + Format.format(cdf[i],3) + "\n");
	//result.append(Format.format(i/10.0,2) + "\t" + Format.format(cdf[i],3) + "\n");

	return result.toString();
    }

    /**
    * Returns a list of coordinates representing the CDF of inter-cluster relative errors.
    * This list is acceptable input to a plotting tool such as Gnuplot.
    * @return
    *   a list of coordinates representing the CDF of inter-cluster relative errors
    **/
    public String getInterCDF()
    {
	StringBuffer result = new StringBuffer();
	
	for(int i=0;i<NUMBER_OF_BUCKETS;i++)
	    result.append(Format.format(i/100.0,2) + "\t" + Format.format(interCDF[i],3) + "\n");

	return result.toString();
    }

    /**
    * Returns a list of coordinates representing the CDF of intra-cluster relative errors.
    * This list is acceptable input to a plotting tool such as Gnuplot.
    * @return
    *   a list of coordinates representing the CDF of intra-cluster relative errors
    **/
    public String getIntraCDF()
    {
	StringBuffer result = new StringBuffer();
	
	for(int i=0;i<NUMBER_OF_BUCKETS;i++)
	    result.append(Format.format(i/100.0,2) + "\t" + Format.format(intraCDF[i],3) + "\n");

	return result.toString();
    }

    /**
    * Returns the proportion of all the relative errors that were greater than the larger bucket.
    * This will correspond to the tail of the CDF.
    * @return
    *   returns the overflow ratio of the CDF of all relative errors
    **/
    public double getOverflowRatio()
    {
	return cdf[NUMBER_OF_BUCKETS];
    }

    /**
    * Returns the number of relative errors that were incalculable due to loss at either the 
    * actual or monitoring site.
    * @return
    *   the number of relative errors that were incalculable due to loss at either the
    *   actual or monitoring site
    **/
    public int getLostReadings()
    {
	return histogram[NUMBER_OF_BUCKETS+1];
    }

    private void findCDF(int hosts)
    { 
	cdf = new double[NUMBER_OF_BUCKETS+1];
	intraCDF = new double[NUMBER_OF_BUCKETS+1];
	interCDF = new double[NUMBER_OF_BUCKETS+1];

	cdf[0] = histogram[0]/((double) numberOfReadings);
	interCDF[0] = interHistogram[0]/((double) interNumberOfReadings);
	intraCDF[0] = intraHistogram[0]/((double) intraNumberOfReadings);

	for(int i=1;i<NUMBER_OF_BUCKETS;i++)
	{
	    cdf[i] = cdf[i-1] + histogram[i]/((double) numberOfReadings);
	    interCDF[i] = interCDF[i-1] + interHistogram[i]/((double) interNumberOfReadings);
  	    intraCDF[i] = intraCDF[i-1] + intraHistogram[i]/((double) intraNumberOfReadings);
	}
  	
	cdf[NUMBER_OF_BUCKETS] = histogram[NUMBER_OF_BUCKETS]/((double) numberOfReadings);
	interCDF[NUMBER_OF_BUCKETS] =
              interHistogram[NUMBER_OF_BUCKETS]/((double) interNumberOfReadings);
	intraCDF[NUMBER_OF_BUCKETS] = 
              intraHistogram[NUMBER_OF_BUCKETS]/((double) intraNumberOfReadings);

    }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
	out.writeObject(histogram);
	out.writeObject(interHistogram);
	out.writeObject(intraHistogram);
	out.writeObject(cdf);
	out.writeObject(interCDF);
	out.writeObject(intraCDF);
	out.writeInt(numberOfReadings);
	out.writeInt(interNumberOfReadings);
	out.writeInt(intraNumberOfReadings);
	out.writeInt(totalValid);
	out.writeObject( ts);
	
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
	histogram = (int[]) in.readObject();
	interHistogram = (int[]) in.readObject();
	intraHistogram = (int[]) in.readObject();
	cdf = (double[]) in.readObject();
	interCDF = (double[]) in.readObject();
	intraCDF = (double[]) in.readObject();
	numberOfReadings = (int) in.readInt();
	interNumberOfReadings = (int) in.readInt();
	intraNumberOfReadings = (int) in.readInt();
	totalValid = (int)in.readInt();
	ts = (TreeSet)in.readObject();
    }
    
    class DoubleComparator implements Comparator, Serializable{
        public int compare(Object o1, Object o2) {
	    double d1 = ((Double)o1).doubleValue();
	    double d2 = ((Double)o2).doubleValue();
	    return ((d1<d2)? -1 : 1);
	}
	
	public boolean equals( Object o1, Object o2) {
	    return false;
	}
	
    }
    
} 










