package cluster;

import java.io.*;
import java.util.*;

/*************************************************************************************
* <CODE>HostMap</CODE> maintains a mapping between each <CODE>Host</CODE> and its
*   corresponding integer.
*
* @author Chris Karlof
*   <A HREF="mailto:ckarlof@eecs.berkeley.edu"> (ckarlof@eecs.berkeley.edu) </A>
**************************************************************************************/


public class HostMap implements Serializable
{
    private Host[] hashmap;

    /**
    * Initializes the mapping with a list of hosts in the <CODE>sitesFile</CODE>.
    * @param sitesFile
    *   the filename of the file containing the list of hosts
    * @param hostNumber
    *   the number of hosts is the file <CODE>sitesFile</CODE>
    **/
    public HostMap(String sitesFile, int hostNumber)
    {
        hashmap = new Host[hostNumber];

        try
        {
            BufferedReader in = new BufferedReader(new InputStreamReader(
                new FileInputStream(new File(sitesFile))));

            for(int i=0;i<hostNumber;i++)
                hashmap[i] = new Host(in.readLine(), i);

            in.close();
        }

        catch(IOException e)
        {
            e.printStackTrace();
	    System.exit(1);
        }
    }

    public HostMap(Cluster c)
    {
	hashmap = new Host[c.size()];
	int count = 0;
	Iterator iterator = c.iterator();
	while(iterator.hasNext())
	{
	    hashmap[count] = (Host) iterator.next();
	    count++;
	}
    } 

    public HostMap(Host[] hosts)
    {
	hashmap = hosts;
    }

    /**
    * Returns the <CODE>Host</CODE> represented by integer <CODE>i</CODE>.
    * @param i
    *   an integer less than the total number of hosts.
    * @return
    *   the host corresponding to the host <CODE>i</CODE>
    **/
    public Host getHost(int i)
    {
        return hashmap[i];
    }

    public int hash(Host host)
    {
	int index = 0;
	while(index < hashmap.length && !hashmap[index].equals(host))
	{
	    index++;
	}

	if(index == hashmap.length)
	{
	    System.out.println(this);
	    System.out.println("looking for: "+host);
	    throw new IllegalStateException("hashmap corrupted");
	}

        return index;
    }
	
    /**
    * Returns the total number of hosts in the mapping.
    *
    * @return
    *   the number of hosts in the mapping
    **/
    public int getNumberOfHosts()
    {
        return hashmap.length;
    }

    public String toString()
    {
	StringBuffer result = new StringBuffer("HOSTMAP:\n");

	for(int i=0;i<hashmap.length;i++)
	{
	    result.append(hashmap[i]+"\n");
	}

	return result.toString();
    }

    public Host[] loadMonitors(String filename, int number)
    {
	Host[] result = new Host[number];

        try
        {
            BufferedReader in = new BufferedReader(new InputStreamReader(
                new FileInputStream(new File(filename))));
	    
            for(int i=0;i<number;i++)
	    {
		Host temp = new Host(in.readLine());
                result[i] = getHost(hash(temp));
	    }

            in.close();
        }

        catch(IOException e)
        {
            e.printStackTrace();
	    System.exit(0);
        }
	
	return result;

    }


    private void writeObject(ObjectOutputStream out) throws IOException
    {
	out.writeObject(hashmap);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
	hashmap = (Host[]) in.readObject();
    }
}









