package cluster;

import java.io.*;
import java.util.*;

public class LimitNumberClusteringAdjust extends Clustering implements Serializable
{

    private Vector[] vClusters;
    private Cluster[] clusters;
  
    private RTTMatrix performance_matrix;

    private double[][] euclidian_matrix;

    private MonitorRank[] monitorRankings;
 
    private int clusterLimit;
    private int hosts;
    
    private MinimalCover minimalCover;
    private RelativeErrorMatrix relativeErrorMatrix;
    private HostMap hostMap;

    public LimitNumberClusteringAdjust(RTTMatrix performance_matrix,
						 int clusterLimit, 
						 String internalClusteringMethod,
						 double internalClusteringArg,
						 int internalClusteringLimit)
    {
	this.clusterLimit = clusterLimit;
	this.hosts = performance_matrix.getNumberOfHosts();
	this.performance_matrix = performance_matrix;
	this.hostMap = performance_matrix.getHostMap();
	this.monitorRankings = new MonitorRank[hosts];

	euclidian_matrix = new double[hosts][hosts];
	euclidianCalc(hosts);
	kCluster_limitNum(hosts);

	clusters = new Cluster[clusterLimit];

	for(int i=0;i<clusterLimit;i++)
	{
	    clusters[i] = new Cluster();
	    Vector c = vClusters[i];
	    Iterator iterator = c.iterator();
	    while(iterator.hasNext())
	    {
		Host h = hostMap.getHost(((Integer) iterator.next()).intValue());
		clusters[i].add(h);
	    }
	}
	findCenters();	
	minimalCover = new MinimalCover(clusters,performance_matrix,internalClusteringMethod,
					     internalClusteringArg,internalClusteringLimit);

	//System.out.println(minimalCover);

	for(int i=0;i<hosts;i++)
	    monitorRankings[i] = new MonitorRank(euclidian_matrix,minimalCover,hostMap,hostMap.getHost(i));

	// i found that because of the approximation algorithm, some sites were improperly placed
        // after we form the clusters, we can adjust them by calculating monitor rankings
	minimalCover.adjustMonitors(monitorRankings);

	//System.out.println(minimalCover);
    }
 
    public MinimalCover getMinimalCover()
    {
	return minimalCover;
    }

    public MonitorRank[] getMonitorRankings()
    {
	return monitorRankings;
    }

    public RelativeErrorMatrix getRelativeErrorMatrix(String srcDate, String compareDate, String intraEstimationMethod, boolean naive)
    {
	return new RelativeErrorMatrix(minimalCover,srcDate, compareDate,intraEstimationMethod,naive);
    }
 
    public RelativeErrorMatrixSW getRelativeErrorMatrixSW(String srcDate, String compareDate, String intraEstimationMethod, boolean naive)
    {
	return new RelativeErrorMatrixSW(minimalCover,srcDate, compareDate,intraEstimationMethod,naive);
    }

    public String toString()
    {
	StringBuffer result = new StringBuffer();
	for(int i=0;i<clusters.length;i++)
	    result.append(clusters[i].toString()+"\n");

	return result.toString();
    }
    
    private void findCenters()
    {

	for(int i=0;i<clusters.length;i++)
	{
	    clusters[i].minimizeAvgNDistanceToFindCenter(euclidian_matrix);
	}
    }

    //hosts: is num of hosts;
    private void euclidianCalc(int hosts)
    {

	for(int i=0;i<hosts;i++) {
	    for(int j=0;j<hosts;j++) {
		if(i == j)
		    continue;		
		Distance d = performance_matrix.squareEuclidianDistance(hostMap.getHost(i),
									hostMap.getHost(j));

		euclidian_matrix[i][j] = d.getDistance();
	    }
	}
    }
		
    private void kCluster_limitNum(int hosts) {
	// init the B sets
	Vector[] B = new Vector[clusterLimit];
	for (int i = 0; i < clusterLimit; i++) {
	    B[i] = new Vector();
	}
	vClusters = B;
	
	// init B1 with all the hosts
	for (int i = 0; i < hosts; i++) {
	    B[0].insertElementAt(new Integer(i), i);
	}

	// assume Bi[0] will always be the head of the set
	// begin algorithm ; for l->1 to k-1 do
	for (int i = 0; i < (clusterLimit - 1); i++) {
	    Vector Bi = B[i];
	    double h = -1; //highest distance
	    int viCluster_limitNum = 0;
	    int vi = 1;           // init to something actual host number
	    int viPosition = 1;   // init to something position in the cluster
	    // h <- max(W(headj, vi))
	    
	    //int headi = ((Integer)Bi.elementAt(0)).intValue();//old code
	    //for (int j = 1; j < Bi.size(); j++) {//old code
	    
	    for (int j = 0; j <= i; j++) {
			Vector Bj = B[j];
			int headj = ((Integer)Bj.elementAt(0)).intValue();
		
			for(int k=1; k<Bj.size(); k++){
		    	int viTemp = ((Integer)Bj.elementAt(k)).intValue();
		    	if (euclidian_matrix[headj][viTemp] > h) {
					h = euclidian_matrix[headj][viTemp];
					vi = viTemp;
					viPosition = k;
					viCluster_limitNum=j;
		    	}
			}
	    }
	    
	    if (h == -1) {
		break;
	    }
	    // move vi to B(l+1) ; make it the head
	    B[viCluster_limitNum].removeElementAt(viPosition);
	    B[(i + 1)].addElement(new Integer(vi));

	    
	    // for each vt...
	    for (int j = 0; j <= i; j++) {
		    Vector Bj = B[j];
		    int headj = ((Integer)Bj.elementAt(0)).intValue();

            Vector Bj_clone = (Vector)Bj.clone();
		    
            for (Enumeration e = Bj_clone.elements(); e.hasMoreElements();) {
                int vt = ((Integer)e.nextElement() ).intValue();
	    
		        if (euclidian_matrix[vt][headj] >= euclidian_matrix[vt][vi]) {
			        // move vt from Bj to B(l + 1)
			        Bj.remove(new Integer(vt));//newer version
			        B[(i + 1)].addElement(new Integer(vt));
    		    }
	    	}
	    }
	}
    }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
	out.writeObject(hostMap);
	out.writeObject(clusters);
	out.writeObject(minimalCover);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
	hostMap = (HostMap) in.readObject();
	clusters = (Cluster[]) in.readObject();
	minimalCover = (MinimalCover) in.readObject();
    }
}
