package cluster;

import java.io.*;
import java.util.*;

public class LimitNumberMinSumDistClustering extends Clustering implements Serializable
{

    private Vector[] vClusters;
    private Cluster[] clusters;
    private LinkedList[] clusterMembersOrdered;
  
    private RTTMatrix performance_matrix;

    private double[][] euclidian_matrix;
 
    private int clusterLimit;
    private int hosts;
    
    private MinimalCover minimalCover;
    private RelativeErrorMatrix relativeErrorMatrix;
    private HostMap hostMap;

    public LimitNumberMinSumDistClustering(RTTMatrix performance_matrix, int clusterLimit, 
				 String internalClusteringMethod,
				 double internalClusteringArg, int internalClusteringLimit)
    {
	this.clusterLimit = clusterLimit;
	this.hosts = performance_matrix.getNumberOfHosts();
	this.performance_matrix = performance_matrix;
	this.hostMap = performance_matrix.getHostMap();

	euclidian_matrix = new double[hosts][hosts];
	euclidianCalc(hosts);
	kCluster_limitNum(hosts);

	clusters = new Cluster[clusterLimit];

	for(int i=0;i<clusterLimit;i++)
	{
	    clusters[i] = new Cluster();
            //clusterMembersOrdered[i] = new LinkedList();
	    Vector c = vClusters[i];
	    Iterator iterator = c.iterator();
	    while(iterator.hasNext())
	    {
		Host h = hostMap.getHost(((Integer) iterator.next()).intValue());
		clusters[i].add(h);
                //clusterMembersOrdered[i].add(h);
	    }
	}
	findCenters();	
	minimalCover = new MinimalCover(clusters,performance_matrix,internalClusteringMethod,
					     internalClusteringArg,internalClusteringLimit);
    }

    public MinimalCover getMinimalCover()
    {
	return minimalCover;
    }

    public RelativeErrorMatrix getRelativeErrorMatrix(String srcDate, String compareDate, String intraEstimationMethod, boolean naive)
    {
	return new RelativeErrorMatrix(minimalCover,srcDate, compareDate,intraEstimationMethod,naive);
    }
 
	public RelativeErrorMatrixSW getRelativeErrorMatrixSW(String srcDate, String compareDate, String intraEstimationMethod, boolean naive) 
	{ 
	return new RelativeErrorMatrixSW(minimalCover,srcDate, compareDate,intraEstimationMethod,naive); 
	}

    public String toString()
    {
	StringBuffer result = new StringBuffer();
	for(int i=0;i<clusters.length;i++)
	    result.append(clusters[i].toString()+"\n\n");

	return result.toString();
    }
    
    public String toStringDebug()
    {
        StringBuffer result = new StringBuffer();
        Host [] tempArray;
        result.append("// Clusters and cluster members output in the order they are chosen by the algorithm.\n// Cluster heads are denoted in brackets.\n\n");
	for(int i=0;i<clusterMembersOrdered.length;i++) {
            tempArray = (Host []) clusterMembersOrdered[i].toArray(new Host[0]);
            for (int j=0; j<tempArray.length; j++) {
                if (j==0) result.append("[" + tempArray[j].toString() + "]\n");
                else result.append(tempArray[j].toString() + "\n");
            }
            result.append("\n");
        }
            
	return result.toString();
    }
    
    private void findCenters()
    {
	for(int i=0;i<clusters.length;i++)
	    clusters[i].minimizeAvgNDistanceToFindCenter(euclidian_matrix);
    }
	
    //hosts: is num of hosts;
    private void euclidianCalc(int hosts)
    {

	for(int i=0;i<hosts;i++) {
	    for(int j=0;j<hosts;j++) {
		if(i == j)
		    continue;		
		Distance d = performance_matrix.squareEuclidianDistance(hostMap.getHost(i),
									hostMap.getHost(j));
		euclidian_matrix[i][j] = d.getDistance();
	    }
	}
    }

    private double computeCost(int hostNum, int hosts)
    {
	if (hostNum<0 || hostNum>=hosts) {
	  System.out.println("Error in computeCost: invalid hostNum");
	  System.exit(1);
	}

	double cost = 0;

      for (int i=0; i<hosts; i++) {
	  if (i==hostNum) continue;

	  cost+=euclidian_matrix[hostNum][i];
      }

	return cost;
    }

    //************************************
    // minSumDist clustering algorithm 
    // by: Johnny Lam (johnnylam@acm.org)
    // date: 4/22/02
    // version 1.1: 5/6/02
    // version 1.2: 5/7/02
    //************************************
    private void kCluster_limitNum(int hosts) {
	// init the B sets
	Vector[] B = new Vector[clusterLimit];
        clusterMembersOrdered = new LinkedList[clusterLimit];
        
	for (int i = 0; i < clusterLimit; i++) {
	    B[i] = new Vector();
            clusterMembersOrdered[i] = new LinkedList();
	}
	vClusters = B;
	
	// init B1 with all the hosts
	for (int i = 0; i < hosts; i++) {
	    B[0].insertElementAt(new Integer(i), i);
            clusterMembersOrdered[0].add(hostMap.getHost(i));
	}

        //********** Johnny added the following code to put a minimum limit of number of nodes for clustering
        int [] numberOfClosestNodesToSite = new int[hosts];
        for (int f=0; f<hosts; f++) {
            numberOfClosestNodesToSite[f] = 0;
        }
        for (int g=0; g<hosts; g++) {
            double minCost;
            int minCostHost;
            for (int h=0; h<hosts; h++) {
                minCost = Double.MAX_VALUE;
                minCostHost = -1;
                for (int i=0; i<hosts; i++) {
                    if (i==h) continue;
                    if (euclidian_matrix[h][i]<=minCost) {
                        minCost = euclidian_matrix[h][i];
                        minCostHost = i;
                    }
                }
                if (minCostHost!=-1) numberOfClosestNodesToSite[minCostHost]++;
            }
        }
        
        int minNumNodesCoeff = 3;
        int minNumNodes = hosts / (minNumNodesCoeff * clusterLimit);
        
        System.out.println("\n\nMinSumDist Algorithm - Step 0: Determine number of closest nodes\n\n");
        System.out.println("  minNumNodes = " + hosts + " / (" + minNumNodesCoeff + " * " + clusterLimit + ") = " + minNumNodes + "\n\n");
        for (int e=0; e<hosts; e++) {
            System.out.println("  host " + e + ": " + numberOfClosestNodesToSite[e] + " node(s)\n");
        }
        System.out.println("\n\n");
        
	//**********
        // Johnny added the following code to determine the head of the first cluster
        
	// initialize the head of the first cluster, before starting algorithm.
        // Johnny decided to keep this out of the algorithm loop for improved clarity.
        Vector B0 = B[0];
	double h0 = -1;
	double tempCost = -1;
	int vi0 = 1;
	int vi0Position = 1;
        int head0 = ((Integer)B0.elementAt(0)).intValue();
	System.out.println("\n\nMinSumDist Algorithm - Step 1: Determine cluster heads\n");
        // find the node in the cluster whose sum of distances *to* all other nodes is minimum.
	for(int i=0; i<B0.size(); i++){
	  int viTemp = ((Integer)B0.elementAt(i)).intValue();
	  //if (i==0 || (tempCost = computeCost(viTemp, hosts)) < h0) {
	  if ((tempCost = computeCost(viTemp, hosts)) < h0 || i == 0) {
	    h0 = tempCost;
	    vi0 = viTemp;
	    vi0Position = i;
	  }
          System.out.println("    cluster 0 node " + i + " (site " + viTemp + ", " + hostMap.getHost(viTemp) + ") 's computed cost = " + tempCost);
	}
        if (h0 == -1) {
	  return;
        }
        System.out.println("  HEAD CHOSEN (cluster 0): cluster 0 node " + vi0Position + " (site " + vi0 + ", " + hostMap.getHost(vi0) + ") with computed cost = " + h0);
	// swap the head and the chosen monitor so that the monitor resides at index 0 for purposes of the algorithm
        B0.setElementAt(new Integer(head0), vi0Position);
        B0.setElementAt(new Integer(vi0), 0);
        // repeat this for our debugging image
	clusterMembersOrdered[0].set(vi0Position, hostMap.getHost(head0));
        clusterMembersOrdered[0].set(0, hostMap.getHost(vi0));
        
        // end Johnny's added code
	//**********
	     
        // now we can assume Bi[0] will always be the head of the set
        
	// begin minSumDist clustering algorithm (step [1])
        
        // provide (k-1) iterations for k clusters (k = clusterLimit)
	for (int i = 0; i < (clusterLimit - 1); i++) {
	    Vector Bi = B[i];
	    double h = -1; //lowest total cost
	    int viCluster_limitNum = 0;
	    int vi = 1;           // init to something, actual host number
	    int viPosition = 1;   // init to something, position in the cluster
            
            // for each processed cluster so far (initially only the first (index 0) cluster is processed)
	    for (int j = 0; j <= i; j++) {
		Vector Bj = B[j];
                // for each node in the cluster
                for(int k=1; k<Bj.size(); k++){
		    int viTemp = ((Integer)Bj.elementAt(k)).intValue();
                    // assuming that this node is a monitor, compute the cost for all nodes.
                    double cost = 0;
                    // go through all of the other nodes (regardless of which cluster they belong to)
                    for (int l=0; l<hosts; l++) {
                        // do not process the same-node pair. Only consider non-head nodes.
                        if (viTemp==l) continue;
                        // determine the cost of this non-head node
                        // initialize the cost of this non-head node to the distance *to* the "propsosed" monitor
                        double costTemp = euclidian_matrix[l][viTemp];
                        double costCompare;
                        // for all processed clusters (i.e. for all existing monitors)
                        for (int m=0; m<=i; m++) {
                            // we wish to find the distance from this non-head node *to* the nearest monitor only
                            if ((costCompare = euclidian_matrix[l][((Integer)B[m].elementAt(0)).intValue()]) < costTemp) {
                                costTemp = costCompare;
                            }
                        }
                        // add this distance to the total cost for this "proposed" monitor
                        cost+=costTemp;
                    }
                    System.out.println("    cluster " + j + " node " + k + " (site " + viTemp + ", " + hostMap.getHost(viTemp) + ") 's computed cost = " + cost);

                    // does this "proposed" monitor have lower cost (i.e. is better) than all "proposed" monitors we have considered so far?
                    if ((j==0 && k==1) || cost < h) {
                        // does this "proposed" monitor have at least the minimum number of nodes?
                        if (numberOfClosestNodesToSite[viTemp]>=minNumNodes) {
                            // if so, we isolate this node and prepare to move it to a new cluster in step [2] below.
                            h = cost;
                            vi = viTemp;
                            viPosition = k;
                            viCluster_limitNum=j;
                        }
		    }
		}
	    }
            
            System.out.println("  HEAD CHOSEN (cluster " + (i+1) + "): cluster " + viCluster_limitNum + " node " + viPosition + " (site " + vi + ", " + hostMap.getHost(vi) + ") 's computed cost = " + h);

	    // move vi to B(l+1) ; make it the head prior to step [2]
	    B[viCluster_limitNum].removeElementAt(viPosition);
	    B[(i + 1)].addElement(new Integer(vi));
            // repeat this for our debugging image
            clusterMembersOrdered[viCluster_limitNum].remove(viPosition);
            clusterMembersOrdered[(i+1)].add(hostMap.getHost(vi));
        }  
        
        System.out.println("\n\nMinSumDist Algorithm - Step 2: Determine cluster membership of non-head nodes\n");
        //System.out.println("Begin step 2 of minSumDist algorithm\n");
        // assume all non-head nodes are still in cluster 1
        // for each vt... (Step [2] stays the same, below)
        Vector B0_clone = (Vector)B0.clone();
        //System.out.println("Size of cluster 0 = " + B0.size());
        for (int n=1; n < B0_clone.size(); n++) {   // for each non-head node
            //System.out.println("Looping... n = " + n);
            int vn = ((Integer)B0_clone.elementAt(n)).intValue();
            double headCostMin = -1;
            int headCompare = -1;
            int targetCluster = -1;
            int targetHost = -1;    // for debugging purposes only

            for (int p=0; p < clusterLimit; p++) {  // compare against each cluster's head
                Vector Bp = B[p];
                headCompare = ((Integer)Bp.elementAt(0)).intValue();

                System.out.println("    Cluster 0 node " + n + " (site " + vn + ", " + hostMap.getHost(vn) + ") -> Cluster " + p + " node 0 (site " + headCompare + ", " + hostMap.getHost(headCompare) + ") = " + euclidian_matrix[vn][headCompare]);
                
                if (p==0 || euclidian_matrix[vn][headCompare] <= headCostMin) {
                    // move vn from Bj to B(i + 1)
                    headCostMin = euclidian_matrix[vn][headCompare];
                    targetCluster = p;
                    targetHost = headCompare;
                }
            }

            System.out.println("  CLUSTER CHOSEN: Cluster 0 node " + n + " (site " + vn + ", " + hostMap.getHost(vn) + ") belongs in cluster " + targetCluster + " with (site " + targetHost + ", " + hostMap.getHost(targetHost) + ") as head and cor_dist() = " + headCostMin);
            if (targetCluster==0) continue; // no need to move
            
            B0.remove(new Integer(vn));//newer version
            B[targetCluster].addElement(new Integer(vn));
            // repeat this for our debugging image
            clusterMembersOrdered[0].remove(hostMap.getHost(vn));
            clusterMembersOrdered[targetCluster].add(hostMap.getHost(vn));
        }
        System.out.println("\n\n\n");
    }


	/* original version
    private void kCluster_limitNum(int hosts) {
	// init the B sets
	Vector[] B = new Vector[clusterLimit];
	for (int i = 0; i < clusterLimit; i++) {
	    B[i] = new Vector();
	}
	vClusters = B;
	
	// init B1 with all the hosts
	for (int i = 0; i < hosts; i++) {
	    B[0].insertElementAt(new Integer(i), i);
	}

	// assume Bi[0] will always be the head of the set
	// begin algorithm ; for l->1 to k-1 do
	for (int i = 0; i < (clusterLimit - 1); i++) {
	    Vector Bi = B[i];
	    double h = -1; //highest distance
	    int viCluster_limitNum = 0;
	    int vi = 1;           // init to something actual host number
	    int viPosition = 1;   // init to something position in the cluster
	    // h <- max(W(headj, vi))
	    
	    //int headi = ((Integer)Bi.elementAt(0)).intValue();//old code
	    //for (int j = 1; j < Bi.size(); j++) {//old code
	    
	    for (int j = 0; j <= i; j++) {
		Vector Bj = B[j];
		int headj = ((Integer)Bj.elementAt(0)).intValue();
		
		for(int k=1; k<Bj.size(); k++){
		    int viTemp = ((Integer)Bj.elementAt(k)).intValue();
		    if (euclidian_matrix[headj][viTemp] > h) {
			h = euclidian_matrix[headj][viTemp];
			vi = viTemp;
			viPosition = k;
			viCluster_limitNum=j;
		    }
		}
	    }
	    
	    if (h == -1) {
		break;
	    }
	    // move vi to B(l+1) ; make it the head
	    B[viCluster_limitNum].removeElementAt(viPosition);
	    B[(i + 1)].addElement(new Integer(vi));

	    
	    // for each vt...
	    for (int j = 0; j <= i; j++) {
		    Vector Bj = B[j];
		    int headj = ((Integer)Bj.elementAt(0)).intValue();

            Vector Bj_clone = (Vector)Bj.clone();
		    
            for (Enumeration e = Bj_clone.elements(); e.hasMoreElements();) {
                int vt = ((Integer)e.nextElement() ).intValue();
	    
		        if (euclidian_matrix[vt][headj] >= euclidian_matrix[vt][vi]) {
			        // move vt from Bj to B(l + 1)
			        Bj.remove(new Integer(vt));//newer version
			        B[(i + 1)].addElement(new Integer(vt));
    		    }
	    	}
	    }
	}
    }
	*/

    private void writeObject(ObjectOutputStream out) throws IOException
    {
	out.writeObject(hostMap);
	out.writeObject(clusters);
	out.writeObject(minimalCover);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
	hostMap = (HostMap) in.readObject();
	clusters = (Cluster[]) in.readObject();
	minimalCover = (MinimalCover) in.readObject();
    }
}


/*	for(int i=0;i<hosts;i++) {
	    for(int j=i+1;j<hosts;j++) {
		double sum = 0;
		//int comparable_dimension = 0;//what is this??xxx
		for(int k=0;k<hosts;k++) {
			  This part is only for WNMMS clustering, which u do not count 
				the distance in the two dimensions determined by two comparator themselves.
		    if(k==i || 
		       k==j || 
		       performance_matrix[i][k] == performance_matrix[j][k]) {
			continue;
		    }
			
		    sum = sum + java.lang.Math.pow((performance_matrix.get(i,k) - 
						    performance_matrix.get(j,k)),2);
		    //comparable_dimension++;//what is this??xxx
		}
		
		euclidean_matrix[i][j]=euclidean_matrix[j][i]
		    = java.lang.Math.pow(sum, .5);
		what is this??xxx
		if(sum < (comparable_dimension * tolerance * tolerance)) {
		    clusters[i].addElement(new Integer(j));
		}
		
	    }
	}
    }
*/










