package cluster;

import java.io.*;
import java.util.*;

/*************************************************************************************
* <CODE>LimitRadiusClustering</CODE> objects are used to represent a solution to
* our clustering problem in which we limit the maximum size of the radius of the clusters.
* The radius is limited by the <CODE>tolerance</CODE> passed to the constructor.
*
* @author Chris Karlof
*   <A HREF="mailto:ckarlof@eecs.berkeley.edu"> (ckarlof@eecs.berkeley.edu) </A>
**************************************************************************************/
public class LimitRadiusProximityClustering extends Clustering implements Serializable
{

  private RTTMatrix performance_matrix;
  private Cluster[] clusters; // each cluster
  private MinimalCover minimalCover;
  private RelativeErrorMatrix relativeErrorMatrix = null;
  private HostMap hostMap;

  private boolean[] covered; // whether the minimal set covers the host
  private int hosts; // number of hosts
  private int forme = 0; // test variable
  private double tolerance; // tolerance variable
 
    /**
     * Creates a new <CODE>LimitRadiusClustering</CODE> object given a <CODE>RTTMatrix</CODE>
     * and an integer <CODE>tolerance</CODE>.
     * @param performance_matrix
     *   a <CODE>RTTMatrix</CODE> of aggregated RTT information
     * @param tolerance
     *   an integer represented a limit on the desired radius size
     **/
    public LimitRadiusProximityClustering(RTTMatrix performance_matrix, double intendedNumClusters,
					String internalClusteringMethod, 
					double internalClusteringArg, int internalClusterLimit)
  {
    this.performance_matrix = performance_matrix;
    this.hosts = performance_matrix.getNumberOfHosts();
    this.tolerance = tolerance;
    this.hostMap = performance_matrix.getHostMap();
    
    tolerance = 20;
    int iteration_no = 0;
    obtainClusters( internalClusteringMethod, 
			       internalClusteringArg,internalClusterLimit);
      ///
      
      while (Math.abs(minimalCover.getNumberOfClusters()- intendedNumClusters) > 2 && iteration_no < 20) {
	  System.out.println("Number of Clusters = " + minimalCover.getNumberOfClusters());
	    if (minimalCover.getNumberOfClusters() > intendedNumClusters) {
		tolerance =  tolerance + 10;
	    }
	    else {
		tolerance  = tolerance - 3 ;
	    
	    }
	    System.out.println("Try tolerance = " + tolerance);
	    
	    obtainClusters(internalClusteringMethod, 
			        internalClusteringArg, internalClusterLimit);
	    iteration_no++;
      }
  }

    public void obtainClusters(String internalClusteringMethod, 
			       double internalClusteringArg, int internalClusterLimit) {
	
    
    covered = new boolean[hosts];
	for(int i=0;i<hosts;i++) {
      covered[i] = false;
    }
	
    clusters = new Cluster[hosts];

    for(int a = 0;a < hosts; a++) {
      clusters[a] = new Cluster(hostMap.getHost(a));
    }

    findClusters();
    minimalCover = new MinimalCover(greedyMinCover(),clusters,performance_matrix,
				    internalClusteringMethod,internalClusteringArg,
				    internalClusterLimit,"lrprox");
    }
  /**
  * Returns the number of hosts used in the clustering.
  * @return
  *   the number of hosts
  **/
  public int getNumberOfHosts()
  {
      return hosts;
  }

  /**
  * Returns a <CODE>Cluster</CODE> array representing a clusters centered at each host
  * whose radius is limited by the given tolerance.
  * @return
  *   a <CODE>Cluster</CODE> array
  **/
  public Cluster[] getClusters()
  {
      return clusters;
  }

  /**
  * Returns a string representation of the clusters centered at each host.
  * @return
  *   a string resulting in calling <CODE>toString()</CODE> on each of the
  *   clusters centered at each host (i.e. the cluster array)
  **/
  public String toString()
  {
      StringBuffer result = new StringBuffer();

      for(int i=0;i<clusters.length;i++)
	  result.append(clusters[i].toString() + "\n");

      return result.toString();
  }

  /**
  * Returns a partition of the hosts into clusters.
  * @return
  *   a <CODE>MinimalCover</CODE> object representing a partition of the hosts into
  *   clusters, each centered at one of the hosts
  **/
  public MinimalCover getMinimalCover()
  {
      return minimalCover;
  }

  /**
  * Returns the matrix of relative errors between using monitor sites to estimate RTT between
  * two hosts and the actual RTT between two hosts. This calculation is done for the given
  * <CODE>date</CODE>. The format of the date is: 101.6.26 -> June 26, 2001
  * @return
  *   a <CODE>RelativeErrorMatrix</CODE> object
  **/
  public RelativeErrorMatrix getRelativeErrorMatrix(String date, String intraEstimationMethod, boolean naive)
  {
      return new RelativeErrorMatrix(minimalCover,date,intraEstimationMethod,naive);
  }

  // execute greedy algorithm to find the minimal number of clusters
  // to cover all hosts
  private HashSet greedyMinCover() {
    HashSet picked = new HashSet();
    HashSet result = new HashSet();

    for(int j=0;j<hosts;j++) {
      int victim = victimize(picked);

      if(victim == -1)
        continue;

      picked.add(hostMap.getHost(victim));
      Cluster t = clusters[victim];

      Iterator iterator = t.iterator();
      while(iterator.hasNext())
      {
		  covered[hostMap.hash((Host) iterator.next())] = true;
	  }

      result.add(hostMap.getHost(victim));
    }

    return result;
  }


  // Find the next host which has the most uncovered hosts
  private int victimize(HashSet pick) {
    int max = 0;
    int picked = -1;
    for(int i=0;i<hosts;i++) {
      if(pick.contains(hostMap.getHost(i)))
        continue;
      Cluster test = clusters[i];
      int count = 0;

      Iterator iterator = test.iterator();
      while(iterator.hasNext())
      {
		  Host intey = (Host) iterator.next();
		  if(!covered[hostMap.hash(intey)])
		  	count++;
      }

      // if the host isn't already covered, or wasn't already
      // accounted for, increase the count because a host
      // can cover itself
      if(!covered[i] && !test.contains(hostMap.getHost(i)))
        count++;

      if(count > max) {
        picked = i;
        max = count;
      }
    }
    forme = forme + max;

    return picked;
  }

  private void findClusters() {
    for(int i=0;i<hosts;i++) {
      for(int j=0;j<hosts;j++) {
        if(i == j) // if i == j don't include it
          continue;
	
	double value = performance_matrix.get(hostMap.getHost(i),hostMap.getHost(j));
	if(value < 0.01)
	    continue;

        if(value < tolerance) {
          clusters[i].add(hostMap.getHost(j));
        }
      }
    }
  }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
	out.writeObject(hostMap);
	out.writeObject(clusters);
	out.writeObject(minimalCover);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
	hostMap = (HostMap) in.readObject();
	clusters = (Cluster[]) in.readObject();
	minimalCover = (MinimalCover) in.readObject();
    }
}

















