package cluster;

import cluster.stability.*;
import java.util.*;
import java.io.*;

/*************************************************************************************
* The <CODE>MinimalCover</CODE> class is used to represent a partitioning of the hosts.
* Each member of the partition has one host designated as the monitor for that cluster.
*
* @author Chris Karlof
*   <A HREF="mailto:ckarlof@eecs.berkeley.edu"> (ckarlof@eecs.berkeley.edu) </A>
**************************************************************************************/

public class MinimalCover implements Serializable
{
    // maps sites to monitors
    private HashMap monitorMap;
    
    // maps sites to backup monitors
    private HashMap bkmonitorMap;

    // cover set of monitors
    private HashSet coverSet;

    // map of monitors to clusters representing a partition of the hosts
    private HashMap partition;
    
    // matrix with round trip times
    private RTTMatrix rttMatrix;

    // mapping of integers to hosts
    protected HostMap hostMap;

    // number of hosts
    private int numberOfHosts;
    
    /**
    * Creates a new <CODE>MinimalCover</CODE> using a cover set of clusters. This cover set
    * is not guaranteed to be a partition. Hosts that are in muliple clusters of the cover set
    * are placed in the cluster whose corresponding center is closest (in n-dimensional space).
    * @param coverSet
    *   a <CODE>HashSet</CODE> of centers(monitors) of clusters that cover all the hosts
    * @param clusters
    *   a <CODE>Cluster</CODE> array indexed by host number
    * @param rttMatrix
    *   the <CODE>RTTMatrix</CODE> used to compute the clusters
    **/
    public MinimalCover(HashSet coverSet, Cluster[] clusters, RTTMatrix rttMatrix,
			String method, double arg, int sizeLimit)
    {
        this.monitorMap = new HashMap();
        this.bkmonitorMap = new HashMap();
        this.coverSet = coverSet;
        this.rttMatrix = rttMatrix;
		this.hostMap = rttMatrix.getHostMap();
		numberOfHosts = rttMatrix.getNumberOfHosts();

        Iterator iterator = coverSet.iterator();

        while(iterator.hasNext())
        {
            Host center = (Host) iterator.next();
            Iterator clusterIterator = clusters[hostMap.hash(center)].iterator();

            while(clusterIterator.hasNext())
            {
                Host host = (Host) clusterIterator.next();

                if(!(monitorMap.containsKey(host)))
                    monitorMap.put(host,center);
                else
                {
                    Host previousCenter = (Host) monitorMap.get(host);
                    Distance d_pc = rttMatrix.squareEuclidianDistance(host,previousCenter);
                    Distance d_nc = rttMatrix.squareEuclidianDistance(host,center);

                    if(d_nc.getDistance() < d_pc.getDistance())
                        monitorMap.put(host,center);
                }
            }
        }
        findPartition();
		calcInfo();
		if(!method.equals("none"))
	    	findInternalClusters(method,arg,sizeLimit);
    }

    public MinimalCover(HashSet coverSet, Cluster[] clusters, GeoDistanceMatrix geoMatrix,
			String method, double arg, int sizeLimit, RTTMatrix rttMatrix)
    {
        this.monitorMap = new HashMap();
        this.bkmonitorMap = new HashMap();
        this.coverSet = coverSet;
        this.rttMatrix = rttMatrix;
	this.hostMap = rttMatrix.getHostMap();
	numberOfHosts = rttMatrix.getNumberOfHosts();

        Iterator iterator = coverSet.iterator();

        while(iterator.hasNext())
        {
            Host center = (Host) iterator.next();
            Iterator clusterIterator = clusters[hostMap.hash(center)].iterator();

            while(clusterIterator.hasNext())
            {
                Host host = (Host) clusterIterator.next();

                if(!(monitorMap.containsKey(host)))
                    monitorMap.put(host,center);
                else
                {
                    Host previousCenter = (Host) monitorMap.get(host);
                    Distance d_pc = geoMatrix.squareEuclidianDistance(host,previousCenter);
                    Distance d_nc = geoMatrix.squareEuclidianDistance(host,center);

                    if(d_nc.getDistance() < d_pc.getDistance())
                        monitorMap.put(host,center);
                }
            }
        }
        findPartition();
	calcInfo();
	if(!method.equals("none"))
	    findInternalClusters(method,arg,sizeLimit);
    }

    // limited radius vector similarity
    public MinimalCover(HashSet coverSet, Cluster[] clusters,
			String method, double arg, int sizeLimit, RTTMatrix rttMatrix)
    {
        this.monitorMap = new HashMap();
        this.bkmonitorMap = new HashMap();
        this.coverSet = coverSet;
        this.rttMatrix = rttMatrix;
	this.hostMap = rttMatrix.getHostMap();
	numberOfHosts = rttMatrix.getNumberOfHosts();

        Iterator iterator = coverSet.iterator();

        while(iterator.hasNext())
        {
            Host center = (Host) iterator.next();
            Iterator clusterIterator = clusters[hostMap.hash(center)].iterator();

            while(clusterIterator.hasNext())
            {
                Host host = (Host) clusterIterator.next();

                if(!(monitorMap.containsKey(host)))
                    monitorMap.put(host,center);
                else
                {
                    Host previousCenter = (Host) monitorMap.get(host);
                    Distance d_pc = rttMatrix.vectorSimilarity(host,previousCenter);
                    Distance d_nc = rttMatrix.vectorSimilarity(host,center);

                    if(d_nc.getDistance() < d_pc.getDistance())
                        monitorMap.put(host,center);
                }
            }
        }
        findPartition();
	calcInfo();
	if(!method.equals("none"))
	    findInternalClusters(method,arg,sizeLimit);
    }

    // minimal cover constructor for LimitNumberClustering technique
    public MinimalCover(Cluster[] clusters, RTTMatrix rttMatrix, String method, double arg,
			int sizeLimit)
    {
	this.monitorMap = new HashMap();
    this.bkmonitorMap = new HashMap();
	this.coverSet = new HashSet();
	this.rttMatrix = rttMatrix;	
	this.hostMap = rttMatrix.getHostMap();
	numberOfHosts = rttMatrix.getNumberOfHosts();

	for(int i=0;i<clusters.length;i++)
	{
	    Host monitor = clusters[i].getCenter();
	    coverSet.add(monitor);
	    Iterator iterator = clusters[i].iterator();
	    while(iterator.hasNext())
	    {
		Host current = (Host) iterator.next();
		monitorMap.put(current,monitor);
	    }
	}
	
	findPartition();
	calcInfo();
	if(!method.equals("none"))
	    findInternalClusters(method,arg,sizeLimit);
    }

    // technique for stability clustering with limit radius clustering
    public MinimalCover(HashSet coverSet, Cluster[] clusters, FrequencyVector[] frequency,
		        HostMap hostMap, String method, double arg, int sizeLimit)
    {
        this.monitorMap = new HashMap();
        this.coverSet = coverSet;
        this.rttMatrix = null;
	this.hostMap = hostMap;
	numberOfHosts = frequency.length;

        Iterator iterator = coverSet.iterator();

        while(iterator.hasNext())
        {
            Host center = (Host) iterator.next();
            Iterator clusterIterator = clusters[hostMap.hash(center)].iterator();

            while(clusterIterator.hasNext())
            {
                Host host = (Host) clusterIterator.next();

                if(!(monitorMap.containsKey(host)))
                    monitorMap.put(host,center);
                else
                {
                    Host previousCenter = (Host) monitorMap.get(host);
                    double f_pc = frequency[hostMap.hash(previousCenter)].getFrequency(host);
                    double f_nc = frequency[hostMap.hash(center)].getFrequency(host);

                    Distance d_pc = rttMatrix.squareEuclidianDistance(host,previousCenter);
                    Distance d_nc = rttMatrix.squareEuclidianDistance(host,center);

                    if(f_nc > f_pc)
                        monitorMap.put(host,center);
		    else if(Math.abs(f_nc-f_pc) < 0.0001 &&
			    d_nc.getDistance() < d_pc.getDistance())
			monitorMap.put(host,center);
                }
            }
        }
        findPartition();
	if(!method.equals("none"))
	    findInternalClusters(method,arg,sizeLimit);
    }

    // limit radius proximity
    public MinimalCover(HashSet coverSet, Cluster[] clusters, RTTMatrix rttMatrix,
			String method, double arg, int sizeLimit, String dummy)
    {
        this.monitorMap = new HashMap();
        this.bkmonitorMap = new HashMap();
        this.coverSet = coverSet;
        this.rttMatrix = rttMatrix;
	this.hostMap = rttMatrix.getHostMap();
	numberOfHosts = rttMatrix.getNumberOfHosts();

        Iterator iterator = coverSet.iterator();

        while(iterator.hasNext())
        {
            Host center = (Host) iterator.next();
            Iterator clusterIterator = clusters[hostMap.hash(center)].iterator();

            while(clusterIterator.hasNext())
            {
                Host host = (Host) clusterIterator.next();

                if(!(monitorMap.containsKey(host)))
                    monitorMap.put(host,center);
                else
                {
                    Host previousCenter = (Host) monitorMap.get(host);
                    double d_pc = rttMatrix.get(host,previousCenter);
                    double d_nc = rttMatrix.get(host,center);

                    if(d_nc < d_pc)
                        monitorMap.put(host,center);
                }
            }
        }
        findPartition();
	calcInfo();
	if(!method.equals("none"))
	    findInternalClusters(method,arg,sizeLimit);
    }

    public RTTMatrix getRTTMatrix()
    {
	return rttMatrix;
    }

    public int getNumberOfHosts()
    {
	return numberOfHosts;
    }

    public HostMap getHostMap()
    {
	return hostMap;
    }

    public void adjustMonitors(MonitorRank[] monitorRankings)
    {
	System.out.println("Adjusting monitors....");
	for(int i=0;i<monitorRankings.length;i++) {
	    if(!(monitorRankings[i].getRank(1).equals(getMonitor(hostMap.getHost(i)))))
	    {
		monitorMap.put(hostMap.getHost(i),monitorRankings[i].getRank(1));
		System.out.println(hostMap.getHost(i)+" changed to monitor "+monitorRankings[i].getRank(1));
	    }
		bkmonitorMap.put(hostMap.getHost(i),monitorRankings[i].getRank(2));
	}
	findPartition();
    }

    /**************************************************************************
    * Returns a string representing the minimal cover.
    * @return
    *   a string
    ***************************************************************************/
    public String toString()
    {
        Iterator iterator = monitorMap.keySet().iterator();
        StringBuffer result = new StringBuffer();

        while(iterator.hasNext())
        {
            Host host = (Host) iterator.next();
            result.append(host.toString() + ": in " + monitorMap.get(host).toString() + "\n");
        }

        result.append("\n");

        iterator = coverSet.iterator();
        while(iterator.hasNext())
        {
            Cluster c = (Cluster) partition.get(iterator.next());
            result.append("size=" + c.size() + " " + c.toString() + "\n");
        }

        result.append("\n");

        return result.toString();
    }

    /**************************************************************************
    * Returns the monitor for a given site.
    * @param site
    *   a host
    * @return
    *   the corresponding monitor for the <CODE>site</CODE>
    ***************************************************************************/
    public Host getMonitor(Host site)
    {
        return (Host) monitorMap.get(site);
    }

    public Host getBKMonitor(Host site)
    {
        return (Host) bkmonitorMap.get(site);
    }

    /**************************************************************************
    * Returns the set of monitors.
    *
    * @return
    *   a <CODE>HashSet</CODE> of the monitors
    ***************************************************************************/
    public HashSet getMonitors()
    {
        return coverSet;
    }

    public Cluster getCluster(Host center)
    {
		return (Cluster) partition.get(center);
    }

    private void findInternalClusters(String method, double arg, int sizeLimit)
    {
	Iterator iterator = partition.keySet().iterator();
	
	while(iterator.hasNext())
	{
	    Host monitor = (Host) iterator.next();
	    Cluster cluster = (Cluster) partition.get(monitor);
	    if(cluster.size() > sizeLimit && sizeLimit > 0)
	    {
		RecCluster newCluster = new RecCluster(cluster,rttMatrix,method,arg,sizeLimit);
		partition.put(monitor,newCluster);
	    }
	}
    }

    // this method calculates the avg RTT and NDistance info for each Cluster
    // in the partition
    private void calcInfo()
    {
	Iterator iterator = partition.keySet().iterator();

	while(iterator.hasNext())
	{
	    Cluster cluster = (Cluster) partition.get((Host) iterator.next());
	    cluster.calcInfo(rttMatrix);
	}
    }

    private void findPartition()
    {
        this.partition = new HashMap();

        Iterator iterator = coverSet.iterator();

        while(iterator.hasNext())
        {
            Host center = (Host) iterator.next();
            partition.put(center,new Cluster(center));
        }

        iterator = monitorMap.keySet().iterator();

        while(iterator.hasNext())
        {
            Host next = (Host) iterator.next();
            Host center = (Host) monitorMap.get(next);
            Cluster image = (Cluster) partition.get(center);
            image.add(next);
            partition.put(center,image);
        }
    }

    public int getNumberOfClusters()
    {
	return coverSet.size();
    }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
        out.writeObject(monitorMap);
        out.writeObject(coverSet);
        out.writeObject(partition);
        out.writeObject(rttMatrix);
	out.writeObject(hostMap);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
        monitorMap = (HashMap) in.readObject();
        coverSet = (HashSet) in.readObject();
        partition = (HashMap) in.readObject();
        rttMatrix = (RTTMatrix) in.readObject();
	hostMap = (HostMap) in.readObject();
    }

}













