package cluster;

import java.io.*;
import java.util.*;

/*************************************************************************************
* The <CODE>RTTMatrix</CODE> class is used to represent round trip times between a set of hosts.
*
* @author Chris Karlof
*   <A HREF="mailto:ckarlof@eecs.berkeley.edu"> (ckarlof@eecs.berkeley.edu) </A>
**************************************************************************************/

public class RTTMatrix implements Serializable
{
    // holds latency measurements for all hosts
    private double[][] rttMatrix;

    // number of hosts
    private int hosts;

    // description of values (gmean, amean)
    private String description;

    // datafile used
    private String datafile;

    // a mapping of integers to actual hosts
    private HostMap hostMap;

    /**
    * Creates a new <CODE>RTTMatrix</CODE> using the data in the <CODE>dataFile</CODE>.
    * @param dataFile
    *   the filename of the file containing the round-trip time data
    * @param numberOfHosts
    *   the number of hosts is the file <CODE>dataFile</CODE>
    * @param description
    *   the type of aggregation done on the raw data (gmean, amean, etc)
    **/
    public RTTMatrix(String datafile, HostMap hostMap, String description, int hosts)
    {
        this.hosts = hosts;
        this.rttMatrix = new double[hosts][hosts];
        this.description = description;
	this.hostMap = hostMap;
	this.datafile = datafile;

        try
        {
            BufferedReader in = new BufferedReader(new InputStreamReader(
                new FileInputStream(new File(datafile))));

            for(int i=0;i<hosts;i++)
            {
                String line = in.readLine();
                StringTokenizer st = new StringTokenizer(line);

                int j=0;
                while(st.hasMoreTokens())
                {
                    String next = st.nextToken();
                    try {
		    rttMatrix[i][j] = (new Double(next)).doubleValue();
                    } catch(ArrayIndexOutOfBoundsException e) { System.out.println(i+" "+j); }
		    j++;
                }
            }

            in.close();
        }

        catch(IOException e)
        {
            e.printStackTrace();
        }
    }

    public RTTMatrix(RTTMatrix oldRTTMatrix, Host[] monitorMask)
    {
	this.datafile = "projection";
	this.hosts = monitorMask.length;
	this.description = oldRTTMatrix.description;
	this.rttMatrix = new double[hosts][hosts];
	this.hostMap = new HostMap(monitorMask);
	for(int i=0;i<hosts;i++)
	    for(int j=0;j<hosts;j++)
		this.rttMatrix[i][j] = oldRTTMatrix.get(hostMap.getHost(i),hostMap.getHost(j));
    }

    public RTTMatrix(RTTMatrix oldRTTMatrix, RecCluster cluster)
    {
	this.datafile = "projection";
	this.hosts = cluster.size();
	this.description = oldRTTMatrix.description;
	this.rttMatrix = new double[hosts][hosts];
	this.hostMap = cluster.getHostMap();
	for(int i=0;i<hosts;i++)
	    for(int j=0;j<hosts;j++)
		this.rttMatrix[i][j] = oldRTTMatrix.get(hostMap.getHost(i),hostMap.getHost(j));
    }

    public int getNumberOfHosts()
    {
	return hosts;
    }

    public HostMap getHostMap()
    {
	return hostMap;
    }

    /**
    * Gets the RTT information between the hosts <CODE>i</CODE> and <CODE>j</CODE>.
    * @param i
    *   a host
    * @param j
    *   a host
    * @return
    *   the RTT information between the two hosts
    **/
    public double get(Host h1, Host h2)
    {
        return rttMatrix[hostMap.hash(h1)][hostMap.hash(h2)];
    }

    /**
    * Gets the filename of the file used to create this <CODE>RTTMatrix</CODE>.
    * @return
    *   the filename containing the data
    **/
    public String getDatafile()
    {
        return datafile;
    }

    /**
    * Calculates the n-space distance between the hosts <CODE>h1</CODE> and <CODE>h2</CODE>.
    * @param h1
    *   a host
    * @param h2
    *   a host
    * @return
    *   a <CODE>Distance</CODE> object representing the n-space distance between the two hosts
    **/
    public Distance squareEuclidianDistance(Host h1, Host h2)
    {
        int i = hostMap.hash(h1);
        int j = hostMap.hash(h2);

        double sum = 0;
        int comparable_dimension = 0;

        for(int k=0;k<hosts;k++)
        {
            if(k==i || k==j) 
  			// if(k==i || k==j || rttMatrix[i][k] == 0 || rttMatrix[j][k] == 0)
            {
                continue;
            }
            sum = sum + java.lang.Math.pow((rttMatrix[i][k]-rttMatrix[j][k]),2);
            comparable_dimension++;
        }

        return new Distance(sum,comparable_dimension);
    }

    public Distance squareEuclidianDistanceNormalized(Host h1, Host h2)
    {
        int i = hostMap.hash(h1);
        int j = hostMap.hash(h2);

        double sum = 0;
        int comparable_dimension = 0;

        for(int k=0;k<hosts;k++)
        {
            if(k==i || k==j) 
  			// if(k==i || k==j || rttMatrix[i][k] == 0 || rttMatrix[j][k] == 0)
            {
                continue;
            }
            sum = sum + java.lang.Math.pow((rttMatrix[i][k]-rttMatrix[j][k]),2);
            comparable_dimension++;
        }

        return new Distance(sum/comparable_dimension,comparable_dimension);
    }

    public Distance[] squareEuclidianDistance(Host h1, Host[] monitors)
    {
	Distance[] result = new Distance[monitors.length];
      
        for(int j=0;j<monitors.length;j++)
        {
	    Host m_j = monitors[j];
	    double sum = 0;
	    int comparable_dimension = 0;

	    for(int k=0;k<monitors.length;k++)
	    {
		Host m_k = monitors[k];
		if(k==j || rttMatrix[hostMap.hash(m_j)][hostMap.hash(m_k)] == 0
		        || rttMatrix[hostMap.hash(h1)][hostMap.hash(m_k)] == 0)
		{
		    continue;
		}
		   sum = sum + java.lang.Math.pow((rttMatrix[hostMap.hash(h1)][hostMap.hash(m_k)]-
                                                   rttMatrix[hostMap.hash(m_j)][hostMap.hash(m_k)]),2);
		   comparable_dimension++;
	    }
	    result[j] = new Distance(sum,comparable_dimension);
	}

        return result;
    }


    public Distance vectorSimilarity(Host h1, Host h2)
    {
        int i = hostMap.hash(h1);
        int j = hostMap.hash(h2);

        double numerator = 0;
        int comparable_dimension = 0;
	double h1_squared_sum = 0;
	double h2_squared_sum = 0;

        for(int k=0;k<hosts;k++)
        {
            if(k==i || k==j ||
                     rttMatrix[i][k] == 0 || rttMatrix[j][k] == 0)
            {
                continue;
            }
            numerator += rttMatrix[i][k]*rttMatrix[j][k];
	    h1_squared_sum += rttMatrix[i][k]*rttMatrix[i][k];
	    h2_squared_sum += rttMatrix[j][k]*rttMatrix[j][k];
            comparable_dimension++;
        }

	double denominator = Math.sqrt(h1_squared_sum*h2_squared_sum);

        return new Distance(1.0-(numerator/denominator),comparable_dimension);
    }	

    private void writeObject(ObjectOutputStream out) throws IOException
    {
        out.writeObject(rttMatrix);
	out.writeInt(hosts);
	out.writeObject(description);
	out.writeObject(datafile);
	out.writeObject(hostMap);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
        rttMatrix = (double[][]) in.readObject();
	hosts = in.readInt();
	description = (String) in.readObject();
	datafile = (String) in.readObject();
	hostMap = (HostMap) in.readObject();
    }
}















