package cluster;

import java.io.*;
import java.util.*;

public class RecCluster extends Cluster implements Serializable
{
    private HostMap hostmap;
    private Clustering clustering;
    private HashMap subclusterMap;
   
    public RecCluster(Cluster c, RTTMatrix oldRTTMatrix, String clusteringMethod,
		      double clusteringArg, int sizeLimit)
    {
	super();
	super.center = c.center;

	// we want to add every host that is in the parent except
	// for the center. we don't want to recursively cluster 
	// the center. we will add the center in at the end
	// for completeness
	Iterator iterator = c.iterator();
	while(iterator.hasNext())
	{
	    Host h = (Host) iterator.next();
	    if(!h.equals(super.center))
		super.cluster.add(h);
	}

	super.avgRTTFromCenter = c.avgRTTFromCenter;
	super.maxRTTFromCenter = c.maxRTTFromCenter;
	super.avgRTTAcrossCluster = c.avgRTTAcrossCluster;
	super.maxRTTAcrossCluster = c.maxRTTAcrossCluster;
	super.avgNDistanceFromCenter = c.avgNDistanceFromCenter;
	super.avgRTTMap = c.avgRTTMap; // this needs to be changed? i dont think so.
	super.triangulationMap = c.triangulationMap;
	this.hostmap = new HostMap(this);
	this.subclusterMap = new HashMap();

	RTTMatrix rttMatrix = new RTTMatrix(oldRTTMatrix,this);
 
	// break... need to recluster with new rttmatrix
	if(clusteringMethod.equals("radius"))
	    clustering = new LimitRadiusClustering(rttMatrix,clusteringArg,"none",0.0,-1);
	else if(clusteringMethod.equals("number"))
	    clustering = new LimitNumberClustering(rttMatrix,(int) (size()/sizeLimit+1),
						   "none",0.0,-1);
	else
	    throw new IllegalArgumentException("Bad clustering method.");

	// this should be done after the recursive clustering
	super.cluster.add(super.center);

    }

    public HostMap getHostMap()
    {
	return hostmap;
    }

    public String toString()
    {
	return super.toString() + "\n" + clustering.toString();
    }

    // handles finding intra cluster error for 
    // those clusters with subclustering
    public double findIntraClusterEstimate(String date, Host host1,
					   Host host2, int index, String method,
					   PrintWriter errorStream)
    {
	double result = -1;

	if(method.equals("subclustering"))
	{
	    if(!subclusterMap.containsKey(date))
	       calcIntraSubcluster(date);
	    
	    result = ((double[][][]) subclusterMap.get(date))
		[index][hostmap.hash(host1)][hostmap.hash(host2)];
	}
	else 
	{
	    System.out.println("warning: switching to super class in intra error.");
	    result = super.findIntraClusterEstimate(date,host1,host2,index,method,errorStream);
	}

	return result;
    }

    private void calcIntraSubcluster(String comparisonDate)
    {
	MinimalCover minimalCover = clustering.getMinimalCover();
	String baseFilename = ".." + SLASH + "data" + SLASH + "raw-data" +
	    SLASH + comparisonDate + SLASH;
	double[][][] result = new double[NUMBER_OF_READINGS][size()-1][size()-1];

	for(int i=0;i<NUMBER_OF_READINGS;i++)
	    for(int j=0;j<size()-1;j++)
		for(int k=0;k<size()-1;k++)
		    result[i][j][k] = -1;

	Iterator iterator1 = iterator();
	try
	{
	    while(iterator1.hasNext())
	    {
		Host host1 = (Host) iterator1.next();
		if(host1.equals(super.center))
		    continue;
		
		Iterator iterator2 = iterator();
		
		while(iterator2.hasNext())
		{
		    Host host2 = (Host) iterator2.next();
		    if(host2.equals(super.center))
			continue;

		    Host monitor1 = minimalCover.getMonitor(host1);
		    Host monitor2 = minimalCover.getMonitor(host2);
		    
		    // first case: both sites are in same subcluster
		    // use avgRTTAcrossCluster
		    if(monitor1.equals(monitor2))
		    {
			double value = minimalCover.getCluster(monitor1).getAvgRTTAcrossCluster();
			for(int i=0;i<NUMBER_OF_READINGS;i++)
			    result[i][hostmap.hash(host1)][hostmap.hash(host2)] = value;
		    }
		    else  // handles each site in different subcluster
		    {
			String filename = baseFilename + monitor1.getName() + MIDDLE + 
			    monitor2.getName() + SLASH + comparisonDate;
			BufferedReader reader = new BufferedReader(
						  new InputStreamReader(
                                                    new FileInputStream(new File(filename))));
			String input;
			int index = 0;
			while((input = reader.readLine()) != null && index < NUMBER_OF_READINGS)
			{
			    if(input.trim().equals(""))
			    {
				System.out.println("Error blank line in "+filename);
				index++;
				continue;
			    }
		    
			    String value = input.substring(input.indexOf(' ')+1);
			    if(value.equals("loss"))
				index++;
			    else
			    {
				result[index][hostmap.hash(host1)][hostmap.hash(host2)] =
				    (new Double(value)).doubleValue();
				index++;
			    }	    
			}
		    }
		}
	    }
	}

	catch(IOException e)
	{
	    System.out.println("File format exception.\n" + e.getMessage());
	    System.exit(0);
	}
	subclusterMap.put(comparisonDate,result);

    }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
	out.writeObject(hostmap);
	out.writeObject(clustering);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
	hostmap = (HostMap) in.readObject();
	clustering = (Clustering) in.readObject();
    }
}











