package cluster;

import java.io.*;

/*************************************************************************************
* The <CODE>RelativeError</CODE> class is used to represent the relative error between
* an estimated RTT and the acutal measurement. It keeps track of the relative error as well
* as auxilary infomation such as the number of readings used to calculate the relative error,
* the number of times either or both of the host or monitor recording a loss.
*
* @author Chris Karlof
*   <A HREF="mailto:ckarlof@eecs.berkeley.edu"> (ckarlof@eecs.berkeley.edu) </A>
**************************************************************************************/

public class RelativeError implements Serializable
{
    private double relativeError;
    private int bothLossCount;
    private int estimateLossOnlyCount;
    private int actualLossOnlyCount;
    private int numberOfReadings;
    private boolean isInterClusterError;

	private int peakHourLoss;	// record the max # of loss in any hour.
	private int peakHourNo;

	private int localHit, localMiss, wanHit, wanMiss;

    /**
    * Creates a new <CODE>RelativeError</CODE> object. All information is initialized to 0.
    **/
    public RelativeError(String type)
    {
	if(type.equals("inter"))
	    isInterClusterError = true;
	else if(type.equals("intra"))
	    isInterClusterError = false;
	else throw new IllegalArgumentException("Error type must be 'inter' or 'intra'");
 
        relativeError = 0.0;
        numberOfReadings = bothLossCount = estimateLossOnlyCount = actualLossOnlyCount = 0;
		peakHourLoss = localHit = localMiss = wanHit = wanMiss = 0;
    }

	public void incLocalHit() {
		localHit++;
	}

	public void incLocalMiss() {
		localMiss++;
	}

	public void incWanHit() {
		wanHit++;
	}

	public void incWanMiss() {
		wanMiss++;
	}

	public int getLocalHit() {
		return localHit;
	}

	public int getLocalMiss() {
		return localMiss;
	}

	public int getWanHit() {
		return wanHit;
	}

	public int getWanMiss() {
		return wanMiss;
	}

	public int getPeakHourLoss() {
		return peakHourLoss;
	}

	public int getPeakHourNo() {
		return peakHourNo;
	}

	public void setPeakHourLoss(int value, int hourNo) {
		peakHourLoss = value;
		peakHourNo = hourNo;
	}

    /**
    * Sets the number of readings used to calculate the relative error.
    * @param i
    *   number of readings
    **/
    public void setNumberOfReadings(int i)
    {
        numberOfReadings = i;
    }

    /**
    * Gets the number of readings used to calculate the relative error.
    * @return
    *   the number of readings used to calculate the relative error
    **/
    public int getNumberOfReadings()
    {
        return numberOfReadings;
    }

    /**
    * Sets the relative error.
    * @param d
    *   the relative error
    **/
    public void setRelativeError(double d)
    {
        relativeError = d;
    }

    /**
    * Gets the relative error.
    * @return
    *   the relative error represented by this <CODE>RelativeError</CODE>
    **/
    public double getRelativeError()
    {
        return relativeError;
    }

    public int getLostReadings()
    {
	return estimateLossOnlyCount + actualLossOnlyCount;
    }

    /**
    * Increments the count of both the monitor and the host recording a loss.
    **/
    public void incBothLossCount()
    {
        bothLossCount++;
    }

    /**
    * Increments the count of only the monitor recording a loss.
    **/
    public void incEstimateLossOnlyCount()
    {
        estimateLossOnlyCount++;
    }

    /**
    * Increments the count of only the host recording a loss.
    **/
    public void incActualLossOnlyCount()
    {
        actualLossOnlyCount++;
    }

    /**
    * Increments the count of all losses
    **/
    public void incAllLossCount()
    {
        actualLossOnlyCount++;
        estimateLossOnlyCount++;
        bothLossCount++;
    }

    /**
    * Gets the number of times both the monitor and the real host recorded a loss.
    * @return
    *   loss count of both monitor and host
    **/
    public int getBothLossCount()
    {
        return bothLossCount;
    }

    /**
    * Gets the number of times only the monitor recorded a loss, but not the host.
    * @return
    *   loss count of monitor only
    **/
    public int getEstimateLossOnlyCount()
    {
        return estimateLossOnlyCount;
    }

    /**
    * Gets the number of times only the host recorded a loss, but not the monitor.
    * @return
    *   loss count of host only
    **/
    public int getActualLossOnlyCount()
    {
        return actualLossOnlyCount;
    }

    /*************************************************************************
    * Determines whether this <CODE>RelativeError</CODE> represents a inter-cluster
    * error.
    * @return
    *   true if this <CODE>RelativeError</CODE> represents an inter-cluster
    *   error, otherwise false
    **************************************************************************/
    public boolean isInterClusterError()
    {
	return isInterClusterError;
    }

    /*************************************************************************
    * Determines whether this <CODE>RelativeError</CODE> represents a intra-cluster
    * error.
    * @return
    *   true if this <CODE>RelativeError</CODE> represents an intra-cluster
    *   error, otherwise false
    **************************************************************************/
    public boolean isIntraClusterError()
    {
	return !isInterClusterError;
    }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
        out.writeDouble(relativeError);
        out.writeInt(bothLossCount);
        out.writeInt(estimateLossOnlyCount);
        out.writeInt(actualLossOnlyCount);
        out.writeInt(numberOfReadings);
	out.writeBoolean(isInterClusterError);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
        relativeError = in.readDouble();
        bothLossCount = in.readInt();
        estimateLossOnlyCount = in.readInt();
        actualLossOnlyCount = in.readInt();
        numberOfReadings = in.readInt();
	isInterClusterError = in.readBoolean();
    }


}
