package cluster;

import java.io.*;
import java.util.*;

/*************************************************************************************
* The <CODE>RelativeErrorMatrix</CODE> class is used to represent the matrix of
* relative errors between monitors and hosts on a specific date.
*
* 11/25/02 modify it to evaluate the ability of congestion/failure detection. 
**************************************************************************************/

public class RelativeErrorMatrix implements Serializable, WNMMSConstants
{

    protected RelativeError[][] relativeError;
    private MinimalCover hostCovering;
    protected String comparisonDate;
    protected String baseFilename;
    protected Histogram histogram;
    protected String intraEstimationMethod;
    private HostMap hostMap;
    private RTTMatrix rttMatrix;

    private double[][] estimates;
    private double[][] actuals;
	private double[][] gmeanM;
	private double[][] gstdM;

	private boolean[][] monAbnormal;

	// private double[][] meanM;
	// private double[][] stdM;

    public RelativeErrorMatrix()
    {
    }

    /**
    * Creates a new <CODE>RelativeErrorMatrix</CODE> using the clustering represented by
    * <CODE>hostCovering</CODE> on the <CODE>comparisonDate</CODE>. The format of the
    * date is: 101.6.26 -> June 26, 2001
    * @param hostCovering
    *   a <CODE>MinimalCover</CODE> representing the partition of the hosts
    * @param comparisonDate
    *   the date which the relative errors are to be calculated
    **/
    public RelativeErrorMatrix(MinimalCover hostCovering, String srcDate, String comparisonDate, String intraEstimationMethod, boolean naive)
    {
        int hosts = hostCovering.getNumberOfHosts();
        this.relativeError = new RelativeError[hosts][hosts];
        this.baseFilename = ".." + SLASH + "data" + SLASH + "raw-data" + SLASH + comparisonDate + SLASH;
        this.comparisonDate = comparisonDate;
	this.intraEstimationMethod = intraEstimationMethod;
        this.hostCovering = hostCovering;
	this.hostMap = hostCovering.getHostMap();
	this.rttMatrix = hostCovering.getRTTMatrix();
	
	estimates = new double[hosts][hosts];
	actuals = new double[hosts][hosts];
	// meanM = new double[hosts][hosts];
	// stdM = new double[hosts][hosts];
	gmeanM = new double[hosts][hosts];
	gstdM = new double[hosts][hosts];

	// read in the normal latency file (g-mean & g-std) to help decide lat anomaly
	try
	{
	File gmeanFile = new File("../data/gmean-data/gmean-"+comparisonDate+".dat"); 
	BufferedReader latReader = new BufferedReader(new InputStreamReader(new FileInputStream(gmeanFile))); 

		for(int i=0;i<hosts;i++) { 
			String line = latReader.readLine(); 
			StringTokenizer st = new StringTokenizer(line); 
			int j=0; 
			while(st.hasMoreTokens()) { 
				String next = st.nextToken(); 
				gmeanM[i][j] = (new Double(next)).doubleValue(); 
				j++; 
			} 
		}
		latReader.close(); 
	} catch(IOException e) { 
		e.printStackTrace(); 
	}

	try
	{
	File gstdFile = new File("../data/gstd-data/gstd-"+comparisonDate+".dat"); 
	BufferedReader latReader = new BufferedReader(new InputStreamReader(new FileInputStream(gstdFile))); 

		for(int i=0;i<hosts;i++) { 
			String line = latReader.readLine(); 
			StringTokenizer st = new StringTokenizer(line); 
			int j=0; 
			while(st.hasMoreTokens()) { 
				String next = st.nextToken(); 
				gstdM[i][j] = (new Double(next)).doubleValue(); 
				j++; 
			} 
		}
		latReader.close(); 
	} catch(IOException e) { 
		e.printStackTrace(); 
	}

	try {
	    PrintWriter errorStream = new PrintWriter(new FileOutputStream(comparisonDate + ".error"));
	    this.doRelativeErrorCalc(errorStream,naive);
	    errorStream.flush();
	    errorStream.close();
	}

	catch(FileNotFoundException e)
	{
	    System.out.println(e.getMessage());
	    System.exit(0);
	}
	this.histogram = new Histogram(this);
    }
    
    public RelativeErrorMatrix(MinimalCover hostCovering, String comparisonDate, String intraEstimationMethod, boolean naive)
    {
        int hosts = hostCovering.getNumberOfHosts();
        this.relativeError = new RelativeError[hosts][hosts];
        this.baseFilename = ".." + SLASH + "data" + SLASH + "raw-data" + SLASH + comparisonDate + SLASH;
        this.comparisonDate = comparisonDate;
	this.intraEstimationMethod = intraEstimationMethod;
        this.hostCovering = hostCovering;
	this.hostMap = hostCovering.getHostMap();
	this.rttMatrix = hostCovering.getRTTMatrix();
	
	estimates = new double[hosts][hosts];
	actuals = new double[hosts][hosts];
	// meanM = new double[hosts][hosts];
	// stdM = new double[hosts][hosts];
	gmeanM = new double[hosts][hosts];
	gstdM = new double[hosts][hosts];

	try {
	    PrintWriter errorStream = new PrintWriter(new FileOutputStream(comparisonDate + ".error"));
	    this.doRelativeErrorCalc(errorStream,naive);
	    errorStream.flush();
	    errorStream.close();
	}

	catch(FileNotFoundException e)
	{
	    System.out.println(e.getMessage());
	    System.exit(0);
	}
	this.histogram = new Histogram(this);
    }
    
    private void doRelativeErrorCalc(PrintWriter errorStream, boolean naive)
    {
		int hosts = hostCovering.getNumberOfHosts();
		int actualLoss, estimateLoss, bothLoss, totalLocalHit, totalLocalMiss, totalWanHit, totalWanMiss;
		actualLoss = estimateLoss = bothLoss = totalLocalHit = totalLocalMiss = totalWanHit = totalWanMiss = 0;

		// find the anomaly cases of monitors, the space may be a bit waste ...
		monAbnormal = new boolean[hosts][NUMBER_OF_READINGS];
		monAnomalyTest(errorStream);

		// get the relative error for each (host_i, host_j) pairs
        for(int i=0;i<hosts;i++)
        {
            Host h_i = hostMap.getHost(i);

            System.out.print("\n" + h_i + "\t" + hostCovering.getMonitor(h_i)+ ": ");
            for(int j=0;j<hosts;j++)
            {

                if (i == j)	{
					// the parameter "inter" here doesn't matter, just initize the value to be 0
                    relativeError[i][j] = new RelativeError("inter");
					System.out.print("0\t");
				} else
				{                
                    Host h_j = hostMap.getHost(j);
                    try {
                        relativeError[i][j] = findRelativeError(h_i,h_j,errorStream,naive);
						actualLoss += relativeError[i][j].getActualLossOnlyCount();
						estimateLoss += relativeError[i][j].getEstimateLossOnlyCount();
						bothLoss += relativeError[i][j].getBothLossCount();

						totalLocalHit += relativeError[i][j].getLocalHit();
						totalLocalMiss += relativeError[i][j].getLocalMiss();
						totalWanHit += relativeError[i][j].getWanHit();
						totalWanMiss += relativeError[i][j].getWanMiss();

                		System.out.print(relativeError[i][j].getActualLossOnlyCount()+"/" + relativeError[i][j].getEstimateLossOnlyCount() + "/" + relativeError[i][j].getBothLossCount() + "\t");

                    } catch(FileNotFoundException e)  {
                        System.out.println(e.getMessage());
					}
                }
            }
        }
        System.out.println("\n"+ actualLoss + "\t" + estimateLoss + "\t" + bothLoss + "\t"
			+ (double)bothLoss/actualLoss * 100.0 + "\t" + (1.0- (double)bothLoss/estimateLoss)*100.0);
        System.out.println(totalLocalHit + "\t" + totalLocalMiss + "\t" + totalWanHit + "\t" + totalWanMiss);
    }

    /**
    * Returns the comparison date used to calculate the relative errors.
    * @return
    *   the comparison date
    **/
    public String getComparisonDate()
    {
        return comparisonDate;
    }

    public MinimalCover getMinimalCover()
    {
	return hostCovering;
    }

    /**
    * Returns a string representing the relative error matrix.
    * @return
    *   a string
    **/
    public String toString()
    {
        int hosts = relativeError.length;
        StringBuffer result = new StringBuffer();

        for(int i=0;i<hosts;i++)
        {
            for(int j=0;j<hosts;j++)
            {
                result.append(relativeError[i][j].getActualLossOnlyCount()+"/" + relativeError[i][j].getEstimateLossOnlyCount() + "/" + relativeError[i][j].getBothLossCount() + "\t");
             //   result.append(hostMap.getHost(i)+" to "+hostMap.getHost(j)+": "+Format.format(relativeError[i][j].getRelativeError(),3)+"\n");
            }

            result.append("\n");
        }

        return result.toString();
    }

    /**************************************************************************
    * Returns the relative error between the estimated and actual RTT times 
    * between two hosts.
    * @param i
    *   an integer representing a host
    * @param j
    *   an integer representing a host
    * @return
    *   a double representing the relative error between the estimated and actual
    *   RTT between the two hosts
    ***************************************************************************/
    public double getRelativeError(Host i, Host j)
    {
	return relativeError[hostMap.hash(i)][hostMap.hash(j)].getRelativeError();
    }

    /**
    * Determines whether the relative error calculated for hosts i and j is an 
    * inter-cluster error, that is, hosts i and j are in the same cluster.
    * @param i
    *   an integer representing a host
    * @param j
    *   an integer representing a host
    * @return
    *   true if the error is an inter-cluster error, otherwise false
    **/
    public boolean isInterClusterError(int i, int j)
    {
	return relativeError[i][j].isInterClusterError();
    }

    /**
    * Determines whether the relative error calculated for hosts i and j is an 
    * intra-cluster error, that is, hosts i and j are in different clusters.
    * @param i
    *   an integer representing a host
    * @param j
    *   an integer representing a host
    * @return
    *   true if the error is an intra-cluster error, otherwise false
    **/
    public boolean isIntraClusterError(int i, int j)
    {
	return relativeError[i][j].isIntraClusterError();
    }

    /**
    * Returns the <CODE>Histogram</CODE> object associated with this relative error 
    * matrix.
    * @return
    *   the associated <CODE>Histogram</CODE> object
    **/
    public Histogram getHistogram()
    {
	return histogram;
    }

    public String getEstimatesVsActuals()
    {
	StringBuffer result = new StringBuffer();
	int hosts = hostCovering.getHostMap().getNumberOfHosts();

	result.append("EstimatesVsActuals:\n");

	double ac25=0;
	double ac50=0;
	double ac75=0;
	double ac100=0;

	double rc25=0;
	double rc50=0;
	double rc75=0;
	double rc100=0;

	for(int i=0;i<hosts;i++)
	    for(int j=0;j<hosts;j++)
	    {
		double are = Math.abs(estimates[i][j]-actuals[i][j])/min(estimates[i][j],actuals[i][j]);
		double rre = relativeError[i][j].getRelativeError();
		result.append("aggregate: "+hostMap.getHost(i)+" -> "+hostMap.getHost(j)+": est="+Format.format(estimates[i][j],1)+" act="+Format.format(actuals[i][j],1)+
			      " are: "+Format.format(are,2)+"\n");
		result.append("other relative error: "+rre+"\n");
		if(are<.25)
		    ac25++;
		if(are<.50)
		    ac50++;
		if(are<.75)
		    ac75++;
		if(are<1)
		    ac100++;

		if(rre<.25)
		    rc25++;
		if(rre<.50)
		    rc50++;
		if(rre<.75)
		    rc75++;
		if(rre<1)
		    rc100++;
	    }
	int t = 106*106;
	result.append("ac25= "+ac25/t+" ac50= "+ac50/t+" ac75= "+ac75/t+" ac100= "+ac100/t+"\n");
	result.append("rc25= "+rc25/t+" rc50= "+rc50/t+" rc75= "+rc75/t+" rc100= "+rc100/t+"\n");
	return result.toString();
    }

    private double min(double i, double j)
    {
	if(i>j)
	    return j;
	else
	    return i;
    }

    public int getNumberOfHosts()
    {
	return relativeError.length;
    }

    public int getLostReadings()
    {
	int total = 0;

	for(int i=0;i<relativeError.length;i++)
	    for(int j=0;j<relativeError.length;j++)
		total += relativeError[i][j].getLostReadings();

	return total;
    }

    public double getLostReadingRatio()
    {
	double den = 0;
	double num = 0;

	for(int i=0;i<relativeError.length;i++)
	    for(int j=0;j<relativeError.length;j++)
	    {
		num += relativeError[i][j].getLostReadings();
		den += relativeError[i][j].getNumberOfReadings();
	    }

	return num/(num+den);
    }

    public double getInterClusterErrorRatio()
    {
	double total = 0;

	for(int i=0;i<relativeError.length;i++)
	    for(int j=0;j<relativeError.length;j++)
	    {
		if(relativeError[i][j].isInterClusterError())
		    total++;
	    }

	return total/(relativeError.length*relativeError.length);
    }
	

    public HostMap getHostMap()
    {
	return hostMap;
    }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
        out.writeObject(relativeError);
        out.writeObject(hostCovering);
        out.writeObject(comparisonDate);
        out.writeObject(baseFilename);
	out.writeObject(histogram);
	out.writeObject(intraEstimationMethod);
	out.writeObject(hostMap);
	out.writeObject(rttMatrix);
	out.writeObject(actuals);
	out.writeObject(estimates);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
        relativeError = (RelativeError[][]) in.readObject();
        hostCovering = (MinimalCover) in.readObject();
        comparisonDate = (String) in.readObject();
        baseFilename = (String) in.readObject();
	histogram = (Histogram) in.readObject();
	intraEstimationMethod = (String) in.readObject();
	hostMap = (HostMap) in.readObject();
	rttMatrix = (RTTMatrix) in.readObject();
	actuals = (double[][]) in.readObject();
	estimates = (double[][]) in.readObject();
    }

	/** Given the file name, read in the NUM_OF_READING entries, and return the latency array.
	 *	Record -1 for empty entry, LOSS constant for "loss", and real values for others.
	 **/
	private int[] readLatency(String fileName, PrintWriter errorStream) {

		int[] lat = new int[NUMBER_OF_READINGS];
        String latLine, value;
		int i=0, loss_num=0;

		try {
        	File latFile = new File(fileName);
        	BufferedReader latReader = new BufferedReader(new InputStreamReader(new FileInputStream(latFile)));

 			while ((latLine = latReader.readLine()) != null && i < NUMBER_OF_READINGS) {

				if (latLine.trim().equals("")) { 
					errorStream.println("Error blank line in "+fileName);
					lat[i] = -1;
		   		} else {
            		value = latLine.substring(latLine.indexOf(' ')+1);
					if (value.equals("loss")) {
						lat[i] = LOSS;
						loss_num++;
					}
					else 	lat[i] = (new Integer(value)).intValue();
				}
				// System.out.print(lat[i] + " ");
				i++;
			}
			if (loss_num == NUMBER_OF_READINGS)  {	// server shutdown, shouldn't be counted!
				for (i=0; i<NUMBER_OF_READINGS; i++)
					lat[i] = -1;
			}
			latReader.close();
		} catch(IOException e) { 
			errorStream.println("File format exception.\n" + e.getMessage()); 
			System.exit(0);     
		}
		return lat;
	}

	/** Simple helper function for anomaly test, assuming normal gmean value (not LOSS) */
	private boolean anomalyTest(int i, int[] lat, Host host1, Host host2) {
		int max = 0;
		for (int j=0; j<SLIDINGWINDOWSIZE; j++)  {
			// if (host1.getName().equals("amp-okstate") ) 
			// 	System.out.print("lat["+(i-j)+"] = " + lat[i-j] + " ");
			// if (lat[i-j] <= gmeanM[host1.getNum()][host2.getNum()]*gstdM[host1.getNum()][host2.getNum()]) {
			if (lat[i-j] <= gmeanM[host1.getNum()][host2.getNum()]*gstdM[host1.getNum()][host2.getNum()]*gstdM[host1.getNum()][host2.getNum()]) {
			// if (lat[i-j] <= meanM[host1.getNum()][host2.getNum()] + 2*stdM[host1.getNum()][host2.getNum()]) {
				return false;
			} else
				if (lat[i-j] > max)		max = lat[i-j];
		}
		// if (host1.getName().equals("amp-okstate") )
		// 	System.out.println("mean: " + meanM[host1.getNum()][host2.getNum()] + " std: " + stdM[host1.getNum()][host2.getNum()]);
		if ( max >=  gmeanM[host1.getNum()][host2.getNum()] * ANOMALY_RATIO_THRE) 
			return true;
		else return false;
	}

	/** Test all the monitors, whether they r abnormal in each min slot of that day. 
	 *
	 **/
	private void monAnomalyTest(PrintWriter errorStream)	{

	String latFilename;
	int lat[];

		HashSet monitors = hostCovering.getMonitors();
		Iterator mon_iter = monitors.iterator(); 
		while(mon_iter.hasNext()) {
			Host mon = (Host) mon_iter.next();
			int mon_num = mon.getNum();

			for (int i = SLIDINGWINDOWSIZE-1; i <NUMBER_OF_READINGS; i++)  {
				Cluster c = hostCovering.getCluster(mon);
				
				// check each member in its cluster 
				Iterator iterator = c.iterator();
				int count = 0;
				int abnormal_count = 0;
				while (iterator.hasNext()) { 
					Host site = (Host)iterator.next();
					if (gmeanM[mon_num][site.getNum()] != LOSS) {
						count++;
						latFilename = baseFilename + mon.getName() + MIDDLE + site.getName() + SLASH + comparisonDate; 
						lat = readLatency(latFilename, errorStream);
						if (anomalyTest(i, lat, mon, site))	abnormal_count++;
					}
				}

				// check other monitors
				Iterator mon_iter2 = monitors.iterator(); 
				while(mon_iter2.hasNext()) { 
					Host mon2 = (Host) mon_iter2.next(); 
					int mon2_num = mon2.getNum();
					if ( (mon2_num != mon_num) && (gmeanM[mon_num][mon2_num] != LOSS) ) {
						count++;
						latFilename = baseFilename + mon.getName() + MIDDLE + mon2.getName() + SLASH + comparisonDate; 
						lat = readLatency(latFilename, errorStream);
						if (anomalyTest(i, lat, mon, mon2))	abnormal_count++;
					}
				}

				if ((double)abnormal_count/count > MON_ANOMALY_THRE)	
					monAbnormal[mon_num][i] = true;
				else 	monAbnormal[mon_num][i] = false;
			}
		}
	}

    /** Note the change from chris' version.
     * A slidingwindow average is taken using <CODE>SlidingWindow </CODE> class
	 *	Assumption: host1 != host2
     **/
    private RelativeError findRelativeError(Host host1, Host host2, PrintWriter errorStream, boolean naive) throws FileNotFoundException
    {
        RelativeError re = null;

        Host monitor1 = hostCovering.getMonitor(host1);
        Host monitor2 = hostCovering.getMonitor(host2);

		// check if gmean data is available for anomaly test
		//	Otherwise, return a random RelativeError with 0 loss counts
		if (gmeanM[host1.getNum()][host2.getNum()] == LOSS)
	    	return (new RelativeError("inter"));

		if ( (!host1.equals(monitor1)) && (gmeanM[monitor1.getNum()][host1.getNum()] == LOSS) )
	    	return (new RelativeError("inter"));
		if ( (!monitor1.equals(monitor2)) && (gmeanM[monitor1.getNum()][monitor2.getNum()] == LOSS) )
	    	return (new RelativeError("inter"));
		if ( (!host2.equals(monitor2)) && (gmeanM[monitor2.getNum()][host2.getNum()] == LOSS) )
	    	return (new RelativeError("inter"));

		// read in the original lat file anyway, used by all options.
       	String latFilename = baseFilename + host1.getName() + MIDDLE + host2.getName() + SLASH + comparisonDate;
		int lat[] = readLatency(latFilename, errorStream);
		/* for (int i=SLIDINGWINDOWSIZE-1; i <NUMBER_OF_READINGS; i++)  
			System.out.print(lat[i] + " ");
		*/

        // handles the case the two monitors are in the same cluster
        if	(monitor1.equals(monitor2))
        {
	    	re = new RelativeError("intra");
			
		   	// if host1 is a monitor, then the distance is exactly measured
			//		read in a->b file and report the # of loss/congestion captured
		   	if  (host1.equals(monitor1)) 	{
				for (int i=SLIDINGWINDOWSIZE-1; i <NUMBER_OF_READINGS; i++)  {
					if (lat[i] > 0)	{	// has some meaningful data
						boolean abnormal = anomalyTest(i, lat, host1, host2);
						if (abnormal)  {
							re.incAllLossCount();
							re.incLocalHit();
						}
					}
				}	// end of for()
			}

			// if host2 is a monitor, we use the distance of (host2 host1), 
			//		read in a->b and b->a and use b->a to estimate the latency 
			//		and loss/congestion captured
			else if (host2.equals(monitor1))    {
        		String monFilename = baseFilename + host2.getName() + MIDDLE + host1.getName() + SLASH + comparisonDate;
				int[] mon = readLatency(monFilename, errorStream);

				for (int i=SLIDINGWINDOWSIZE-1; i <NUMBER_OF_READINGS; i++)  {
					if (lat[i] > 0)	{	// has some meaningful data

						boolean abnormal = anomalyTest(i, lat, host1, host2);
						if (abnormal)  re.incActualLossOnlyCount();

						boolean monAbnor = anomalyTest(i, mon, host1, host2);
						if (monAbnor)  {
							re.incEstimateLossOnlyCount();
							if (abnormal) {
								re.incBothLossCount();
								re.incLocalHit();
							} else 	re.incLocalMiss();
						}
					}
				}	// end of for()
			}
			// Else read in a->b, m->a, and m->b, use 1/2(m->a + m->b) to estimate 
			//		the latency, and evaluate the loss/congestion captured.
			else {
        		String monFName1 = baseFilename + monitor1.getName() + MIDDLE + host1.getName() + SLASH + comparisonDate;
				int[] mon1 = readLatency(monFName1, errorStream);
        		String monFName2 = baseFilename + monitor1.getName() + MIDDLE + host2.getName() + SLASH + comparisonDate;
				int[] mon2 = readLatency(monFName2, errorStream);

				for (int i=SLIDINGWINDOWSIZE-1; i <NUMBER_OF_READINGS; i++)  {
					if (lat[i] > 0)	{	// has some meaningful data

						boolean abnormal = anomalyTest(i, lat, host1, host2);
						if (abnormal)  re.incActualLossOnlyCount();

						if (!monAbnormal[monitor1.getNum()][i]) {
							boolean monAbnormal1 = anomalyTest(i, mon1, monitor1, host1);
							boolean monAbnormal2 = anomalyTest(i, mon2, monitor1, host2);
							if (monAbnormal1 || monAbnormal2)  {
								re.incEstimateLossOnlyCount();
								if (abnormal) {
									re.incBothLossCount();
									re.incLocalHit();
								} else	 re.incLocalMiss();
							}
						}
					}
				}	// end of for()
			}
			/* special case for naive
		    		if(naive)
						estimate = rttMatrix.get(host1,host2)+"";
			*/
			/* Average for relative error computing
	    	actuals[hostMap.hash(host1)][hostMap.hash(host2)] = actual_total/numberOfReadings;
	    	estimates[hostMap.hash(host1)][hostMap.hash(host2)] = estimate_total/numberOfReadings;
	    
	    if(numberOfReadings == 0)
			relativeError.setRelativeError(-1.0);
	    else
			relativeError.setRelativeError(total/numberOfReadings);
	    
	    relativeError.setNumberOfReadings(numberOfReadings);
		*/
	    
        }
        else    // else the two hosts are in different clusters
	    {
			re = new RelativeError("inter");

		   	// if host1 and host2 r both monitors, then the distance is exactly measured
			//		read in a->b file and report the # of loss/congestion captured
		   	if  ( (host1.equals(monitor1)) && (host2.equals(monitor2)) ) 	{
				for (int i=SLIDINGWINDOWSIZE-1; i <NUMBER_OF_READINGS; i++)  {
					if (lat[i] > 0)	{	// has some meaningful data
						boolean abnormal = anomalyTest(i, lat, host1, host2);
						if (abnormal)  {
							re.incAllLossCount();
							re.incWanHit();
						} 
					}
				}	// end of for()
			}

			// if only host1 is a monitor, we use the distance of (host1 monitor2), 
			//		read in host1->monitor2 and monitor2->host2, to estimate the latency 
			//		and loss/congestion captured
			else if (host1.equals(monitor1))    {
        		String monFName1 = baseFilename + host1.getName() + MIDDLE + monitor2.getName() + SLASH + comparisonDate;
				int[] mon1 = readLatency(monFName1, errorStream);

        		String monFName2 = baseFilename + monitor2.getName() + MIDDLE + host2.getName() + SLASH + comparisonDate;
				int[] mon2 = readLatency(monFName2, errorStream);

				for (int i=SLIDINGWINDOWSIZE-1; i <NUMBER_OF_READINGS; i++)  {
					if (lat[i] > 0)	{	// has some meaningful data
						boolean abnormal = anomalyTest(i, lat, host1, host2);
						if (abnormal)  re.incActualLossOnlyCount();

						if (!monAbnormal[monitor2.getNum()][i]) {
							boolean monAbnormal1 = anomalyTest(i, mon1, monitor1, monitor2);
							boolean monAbnormal2 = anomalyTest(i, mon2, monitor2, host2);
							if (monAbnormal1 || monAbnormal2)  {
								re.incEstimateLossOnlyCount();
								if (abnormal) {
									re.incBothLossCount();
									if (monAbnormal1) re.incWanHit();
									if (monAbnormal2) re.incLocalHit();
								} else {
									if (monAbnormal1) re.incWanMiss(); 
									if (monAbnormal2) re.incLocalMiss();
								}
							}
						}
					}
				}	// end of for()
			}

			// if only host2 is a monitor, we use the distance of (monitor1, host2), 
			//		read in monitor1->host1 and monitor1->monitor2, to estimate the latency 
			//		and loss/congestion captured
			else if (host2.equals(monitor2))    {
        		String monFName1 = baseFilename + monitor1.getName() + MIDDLE + host1.getName() + SLASH + comparisonDate;
				int[] mon1 = readLatency(monFName1, errorStream);

        		String monFName2 = baseFilename + monitor1.getName() + MIDDLE + host2.getName() + SLASH + comparisonDate;
				int[] mon2 = readLatency(monFName2, errorStream);

				for (int i=SLIDINGWINDOWSIZE-1; i <NUMBER_OF_READINGS; i++)  {
					if (lat[i] > 0)	{	// has some meaningful data
						boolean abnormal = anomalyTest(i, lat, host1, host2);
						if (abnormal)  re.incActualLossOnlyCount();

						if (!monAbnormal[monitor1.getNum()][i]) {
							boolean monAbnormal1 = anomalyTest(i, mon1, monitor1, host1);
							boolean monAbnormal2 = anomalyTest(i, mon2, monitor1, monitor2);
							if (monAbnormal1 || monAbnormal2)  {
								re.incEstimateLossOnlyCount();
								if (abnormal) {
									re.incBothLossCount();
									if (monAbnormal1) re.incLocalHit();
									if (monAbnormal2) re.incWanHit();
								} else {
									if (monAbnormal1) re.incLocalMiss(); 
									if (monAbnormal2) re.incWanMiss();
								}
							}
						}
					}
				}	// end of for()
			}

			// Else use dist of (monitor1, monitor2). read in monitor1->host1, 
			//	monitor1->monitor2, and monitor2->host2, and evaluate the loss/congestion captured.
			else {
        		String monFName1 = baseFilename + monitor1.getName() + MIDDLE + host1.getName() + SLASH + comparisonDate;
				int[] mon1 = readLatency(monFName1, errorStream);
        		String monFName2 = baseFilename + monitor1.getName() + MIDDLE + monitor2.getName() + SLASH + comparisonDate;
				int[] mon2 = readLatency(monFName2, errorStream);
        		String monFName3 = baseFilename + monitor2.getName() + MIDDLE + host2.getName() + SLASH + comparisonDate;
				int[] mon3 = readLatency(monFName3, errorStream);

				for (int i=SLIDINGWINDOWSIZE-1; i <NUMBER_OF_READINGS; i++)  {
					if (lat[i] > 0)	{	// has some meaningful data

						boolean abnormal = anomalyTest(i, lat, host1, host2);
						if (abnormal)  re.incActualLossOnlyCount();

						boolean monAbnormal1 = anomalyTest(i, mon1, monitor1, host1);
						boolean monAbnormal2 = anomalyTest(i, mon2, monitor1, monitor2);
						boolean monAbnormal3 = anomalyTest(i, mon3, monitor2, host2);

						if (monAbnormal[monitor1.getNum()][i]) {
							if (!monAbnormal[monitor2.getNum()][i]) { // only m1 fails
								if (monAbnormal3) {
									re.incEstimateLossOnlyCount();
									if (abnormal) {
										re.incBothLossCount();
										re.incLocalHit();
									}	else re.incLocalMiss();
								}
							} else {	// both fail, do nothing.  
							}
						} else if (monAbnormal[monitor2.getNum()][i])  {	// only m2 fails
							if (monAbnormal1) {
								re.incEstimateLossOnlyCount();
								if (abnormal) {
									re.incBothLossCount();
									re.incLocalHit();
								}	else re.incLocalMiss();
							}
						} else {	// no monitor fails
							if (monAbnormal1 || monAbnormal2 || monAbnormal3)  {
								// System.out.println(host1 + " " + monitor1 + " " + monitor2 + " " + host2 + " at slot " + i +": " + monAbnormal1 + " " + monAbnormal2 + " " + monAbnormal3);
								re.incEstimateLossOnlyCount();
								if (abnormal) {
									re.incBothLossCount();
									if (monAbnormal1) re.incLocalHit();
									if (monAbnormal2) re.incWanHit();
									if (monAbnormal3) re.incLocalHit();
								} else {
									if (monAbnormal1) re.incLocalMiss(); 
									if (monAbnormal2) re.incWanMiss();
									if (monAbnormal3) re.incLocalMiss(); 
								}
							}
						}
					}
				}	// end of for()
			}
		}
		/*
		try {
		    //the slidingWinow constructor will perform all calculations and throw exceptions if error encounterd
		    
		    SlidingWindow slidingWindow = new SlidingWindow(host1, host2,monitorFilename, hostFilename, relativeError, hostReader, monitorReader,errorStream, naive, rttMatrix);
		    
		    actuals[hostMap.hash(host1)][hostMap.hash(host2)] = slidingWindow.getActual();
		    estimates[hostMap.hash(host1)][hostMap.hash(host2)] = slidingWindow.getEstimate();
	    
		    if (slidingWindow.getNumberOfReadings() == 0) 
			relativeError.setRelativeError(-1.0);
		    else 
			relativeError.setRelativeError(slidingWindow.getTotal()/slidingWindow.getNumberOfReadings());
	    
		    relativeError.setNumberOfReadings(slidingWindow.getNumberOfReadings());
		}
		*/

        // System.out.println(re.getActualLossOnlyCount()+"/" + re.getEstimateLossOnlyCount() + "/" + re.getBothLossCount());
	   	return re;
    }

    /** 
     * Note the change in the metric for calculation
     **/
	
    
    private double relativeErrorCalc(String estimate, String actual, RelativeError relativeError,
				     PrintWriter errorStream)
    {
	if(estimate.equals("zero"))
	    return 0.0;
	
        if(estimate.equals("loss"))
        {
            if(actual.equals("loss"))
		{
		    relativeError.incBothLossCount();
		    return 0.0;
            }
            else
            {
                relativeError.incEstimateLossOnlyCount();
                return -100.0;
            }
        }
        else
        {
            if(actual.equals("loss"))
            {
                relativeError.incActualLossOnlyCount();
                return -100.0;
            }
             else
             {
                try
                {
                    double est = (new Double(estimate)).doubleValue();
                    double act = (new Double(actual)).doubleValue();
                    // if actual is 0, make it 1
                    if(act<1) act = 1.0;
		    ///
		    /*
                    return (Math.abs(est-act))/((act<est) ? act : est);
		    */
		    return (Math.abs(log2(est) - log2(act)));
		    
		}
		
                catch(NumberFormatException e)
                {
                    errorStream.println("Number format exception. \n" + e.getMessage());
                    errorStream.println("Est='"+estimate+"'");
		    errorStream.println("Act='"+actual+"'");
		    throw new IllegalArgumentException();
                }
            }
        }
    }

    private double relativeErrorCalcNaive(String estimate, String actual, RelativeError relativeError,
					  PrintWriter errorStream)
    {
	if(estimate.equals("zero"))
	    return 0.0;

        if(estimate.equals("loss"))
        {
            if(actual.equals("loss"))
            {
                relativeError.incBothLossCount();
                return 0.0;
            }
            else
            {
                relativeError.incEstimateLossOnlyCount();
                return -100.0;
            }
        }
        else
        {
            if(actual.equals("loss"))
            {
                relativeError.incActualLossOnlyCount();
                return -100.0;
            }
             else
             {
                try
                {
                    double est = (new Double(estimate)).doubleValue();
                    double act = (new Double(actual)).doubleValue();
                    // if actual is 0, make it 1
                    if(act<1) act = 1.0;
		    ///
		    /*
                    return (Math.abs(est-act))/((act<est) ? act : est);
		    */
		    return (Math.abs(log2(est) - log2(act)));
		    
                }

                catch(NumberFormatException e)
                {
                    errorStream.println("Number format exception. \n" + e.getMessage());
                    errorStream.println("Est='"+estimate+"'");
		    errorStream.println("Act='"+actual+"'");
		    throw new IllegalArgumentException();
                }
            }
        }
    }
    public static double log2 (double x) 
    {
	return (Math.log(x) / Math.log(2));
    }
    
    
}















