package cluster;

import java.io.*;
import java.util.*;

/*************************************************************************************
* The <CODE>RelativeErrorMatrixSW</CODE> class is used to represent the matrix of
* relative errors between monitors and hosts on a specific date.
*
* 11/25/02 modify it to evaluate the ability of congestion/failure detection. 
* 1. Read in the gmean, gstd, and parsed slinding window data file to define the loss
*	matrix for any <src, dest> pair on each 1431 (1440-10+1) slots.
**************************************************************************************/

public class RelativeErrorMatrixSW implements Serializable, WNMMSConstants
{

    protected RelativeError[][] relativeError;
    private MinimalCover hostCovering;
    protected String comparisonDate;
    protected String baseFilename;
    protected Histogram histogram;
    protected String intraEstimationMethod;
    private HostMap hostMap;
    private RTTMatrix rttMatrix;

    private double[][] estimates;
    private double[][] actuals;
	private double[][] gmeanM;
	private double[][] gstdM;
	// private double[][] meanM;
	// private double[][] stdM;
	// variable for keeping all the loss info for all host pair at each time slot 
	private int[][][] loss;
	// variables for monitor self-diagnosis, checking if the monitor itself has LM c/f.
	private boolean[][] monAbnormal;

    public RelativeErrorMatrixSW()
    {
    }

    /**
    * Creates a new <CODE>RelativeErrorMatrixSW</CODE> using the clustering represented by
    * <CODE>hostCovering</CODE> on the <CODE>comparisonDate</CODE>. The format of the
    * date is: 101.6.26 -> June 26, 2001
    * @param hostCovering
    *   a <CODE>MinimalCover</CODE> representing the partition of the hosts
    * @param comparisonDate
    *   the date which the relative errors are to be calculated
    **/
    public RelativeErrorMatrixSW(MinimalCover hostCovering, String srcDate, String comparisonDate, String intraEstimationMethod, boolean naive)
    {
        int hosts = hostCovering.getNumberOfHosts();
        this.relativeError = new RelativeError[hosts][hosts];
        this.baseFilename = ".." + SLASH + "data" + SLASH + "sliding-max" + SLASH + comparisonDate + SLASH;
        this.comparisonDate = comparisonDate;
	this.intraEstimationMethod = intraEstimationMethod;
        this.hostCovering = hostCovering;
	this.hostMap = hostCovering.getHostMap();
	this.rttMatrix = hostCovering.getRTTMatrix();
	
	estimates = new double[hosts][hosts];
	actuals = new double[hosts][hosts];
	// meanM = new double[hosts][hosts];
	// stdM = new double[hosts][hosts];
	gmeanM = new double[hosts][hosts];
	gstdM = new double[hosts][hosts];
	// aux variable for improving the computation speed
	loss= new int[hosts][hosts][NUMBER_OF_READINGS-SLIDINGWINDOWSIZE+1];	
	// find the anomaly cases of monitors, the space may be a bit waste ...
	monAbnormal = new boolean[hosts][NUMBER_OF_READINGS-SLIDINGWINDOWSIZE+1];

	// read in the normal latency file (g-mean & g-std) to help decide lat anomaly
	try
	{
	File gmeanFile = new File("../data/gmean-data/gmean-"+ comparisonDate +".dat"); 
	BufferedReader latReader = new BufferedReader(new InputStreamReader(new FileInputStream(gmeanFile))); 

		for(int i=0;i<hosts;i++) { 
			String line = latReader.readLine(); 
			StringTokenizer st = new StringTokenizer(line); 
			int j=0; 
			while(st.hasMoreTokens()) { 
				String next = st.nextToken(); 
				gmeanM[i][j] = (new Double(next)).doubleValue(); 
				j++; 
			} 
		}
		latReader.close(); 
	} catch(IOException e) { 
		e.printStackTrace(); 
	}

	try
	{
	File gstdFile = new File("../data/gstd-data/gstd-"+comparisonDate+".dat"); 
	BufferedReader latReader = new BufferedReader(new InputStreamReader(new FileInputStream(gstdFile))); 

		for(int i=0;i<hosts;i++) { 
			String line = latReader.readLine(); 
			StringTokenizer st = new StringTokenizer(line); 
			int j=0; 
			while(st.hasMoreTokens()) { 
				String next = st.nextToken(); 
				gstdM[i][j] = (new Double(next)).doubleValue(); 
				j++; 
			} 
		}
		latReader.close(); 
	} catch(IOException e) { 
		e.printStackTrace(); 
	}

	try {
	    PrintWriter errorStream = new PrintWriter(new FileOutputStream(comparisonDate + ".error"));
		// initialize the loss information for all host pairs + all 1431 windows to simplify error calc.
		this.computeLoss();
		this.monAnomalyTest(errorStream);
	    this.doRelativeErrorCalc(errorStream,naive);
	    errorStream.flush();
	    errorStream.close();
	}

	catch(FileNotFoundException e)
	{
	    System.out.println(e.getMessage());
	    System.exit(0);
	}
	// this.histogram = new Histogram(this);
    }
    
    public RelativeErrorMatrixSW(MinimalCover hostCovering, String comparisonDate, String intraEstimationMethod, boolean naive)
    {
        int hosts = hostCovering.getNumberOfHosts();
        this.relativeError = new RelativeError[hosts][hosts];
        this.baseFilename = ".." + SLASH + "data" + SLASH + "raw-data" + SLASH + comparisonDate + SLASH;
        this.comparisonDate = comparisonDate;
	this.intraEstimationMethod = intraEstimationMethod;
        this.hostCovering = hostCovering;
	this.hostMap = hostCovering.getHostMap();
	this.rttMatrix = hostCovering.getRTTMatrix();
	
	estimates = new double[hosts][hosts];
	actuals = new double[hosts][hosts];
	// meanM = new double[hosts][hosts];
	// stdM = new double[hosts][hosts];
	gmeanM = new double[hosts][hosts];
	gstdM = new double[hosts][hosts];

	try {
	    PrintWriter errorStream = new PrintWriter(new FileOutputStream(comparisonDate + ".error"));
	    this.doRelativeErrorCalc(errorStream,naive);
	    errorStream.flush();
	    errorStream.close();
	}

	catch(FileNotFoundException e)
	{
	    System.out.println(e.getMessage());
	    System.exit(0);
	}
	// this.histogram = new Histogram(this);
    }
    
	/** Test all the monitors, whether they r abnormal in each time window slot. 
	 **/
	private void monAnomalyTest(PrintWriter errorStream)	{

		HashSet monitors = hostCovering.getMonitors();
		Iterator mon_iter = monitors.iterator(); 
		while(mon_iter.hasNext()) {
			Host mon = (Host) mon_iter.next();
			int mon_num = mon.getNum();

			for (int t = 0; t < (NUMBER_OF_READINGS-SLIDINGWINDOWSIZE+1); t++)  {
				Cluster c = hostCovering.getCluster(mon);
				
				// check each member in its cluster 
				Iterator iterator = c.iterator();
				int count = 0;
				int abnormal_count = 0;
				while (iterator.hasNext()) { 
					Host site = (Host)iterator.next();
					if (gmeanM[mon_num][site.getNum()] != LOSS) {
						count++;
						if (loss[mon_num][site.getNum()][t]==1) abnormal_count++;
					}
				}

				// check other monitors
				Iterator mon_iter2 = monitors.iterator(); 
				while(mon_iter2.hasNext()) { 
					Host mon2 = (Host) mon_iter2.next(); 
					int mon2_num = mon2.getNum();
					if ( (mon2_num != mon_num) && (gmeanM[mon_num][mon2_num] != LOSS) ) {
						count++;
						if (loss[mon_num][mon2_num][t]==1) abnormal_count++;
					}
				}

				if ((double)abnormal_count/count > MON_ANOMALY_THRE)	
					monAbnormal[mon_num][t] = true;
				else 	monAbnormal[mon_num][t] = false;
			}
		}
	}

	private void computeLoss() {

		String latFilename;
		int lat[];
		int count, lossCount;
        int hosts = hostCovering.getNumberOfHosts();

		// init the loss[][] matrix
		for (int i=0; i<hosts; i++) 
			for (int j=0; j<hosts; j++)
				for (int t = 0; t < (NUMBER_OF_READINGS-SLIDINGWINDOWSIZE+1); t++) 
					loss[i][j][t] = -1;

		for(int i=0;i<hosts;i++) {
			Host h_i = hostMap.getHost(i);

			for(int j=0;j<hosts;j++) 	// vertically set up the values
				if ( (i != j) && (gmeanM[i][j] != LOSS) ) 	{	// only count the meaningful sites
					Host h_j = hostMap.getHost(j);

					latFilename = baseFilename + h_i.getName() + MIDDLE + h_j.getName() + SLASH + comparisonDate; 
					readSWinLat(latFilename, i, j); 
				}
		}
	}


	/** Given the slinding-data file name, read in the (NUM_OF_READING-SLIDINGWINDOWSIZE+1) entries, 
	 *  and return two arrays, min and max.
	 *	Record -1 for empty entry, LOSS constant for "loss", and real values for others.
	 **/
	private void readSWinLat(String fileName, int src, int dest) {

        String latLine, min_value, max_value;
		int i=0, loss_num=0, min, max;

		try {
        	File latFile = new File(fileName);
        	BufferedReader latReader = new BufferedReader(new InputStreamReader(new FileInputStream(latFile)));
			// each file has strictly (NUMBER_OF_READINGS-SLIDINGWINDOWSIZE+1) entries.
			int size = NUMBER_OF_READINGS-SLIDINGWINDOWSIZE+1;
			for (i=0; i < size; i++ )  {

 				latLine = latReader.readLine();
				if (latLine.trim().equals("")) { 
					System.err.println("Error blank line in "+fileName);
					System.exit(-1);
		   		} else {	// 3 column format, "time min max"
            		latLine = latLine.substring(latLine.indexOf(' ')+1);	// skip time
					min_value = latLine.substring(0, latLine.indexOf(' '));
					max_value = latLine.substring(latLine.indexOf(' ')+1);
					if (min_value.equals("loss")) {
						loss[src][dest][i] = 1;
						loss_num++;
					}	else {
						min = (new Integer(min_value)).intValue();
						if (max_value.equals("loss") ) 	max = LOSS;
						else 	max = (new Integer(max_value)).intValue();
						if ( (min > gmeanM[src][dest]*gstdM[src][dest]) && 
							(max > gmeanM[src][dest]* ANOMALY_RATIO_THRE) ) {
							loss[src][dest][i] = 1;
							loss_num++;
						} else loss[src][dest][i] = 0;
					}
				}
			}
			if (loss_num == size)  {	// server shutdown, shouldn't be counted!
				for (i=0; i<size; i++)
					loss[src][dest][i] = -1;
			}
			if (loss_num <= LOSS_LENGTH_MIN)  {	// remove those spoty congestion/failures
				for (i=0; i<size; i++)
					loss[src][dest][i] = 0;
			}
			latReader.close();
			/* System.out.println("\nhost " + src + " to host " + dest);
			for (i=0; i<size; i++) 
					System.out.print(loss[src][dest][i] + " ");
			*/
		} catch(IOException e) { 
			System.err.println("File format exception.\n" + e.getMessage()); 
			System.exit(0);     
		}
	}

    private void doRelativeErrorCalc(PrintWriter errorStream, boolean naive)
    {
		int hosts = hostCovering.getNumberOfHosts();
		int actualLoss, estimateLoss, bothLoss;
		actualLoss = estimateLoss = bothLoss = 0;

        for(int i=0;i<hosts;i++)
        {
            Host h_i = hostMap.getHost(i);

			// for loss count matrix
            System.out.print("\n" + h_i + "\t" + hostCovering.getMonitor(h_i)+ ": ");
            for(int j=0;j<hosts;j++)
            {

                if (i == j)	{
					// the parameter "inter" here doesn't matter, just initize the value to be 0
                    relativeError[i][j] = new RelativeError("inter");
                	System.out.print("0\t");
                } else 
				{                
                    Host h_j = hostMap.getHost(j);
                    try {
                        relativeError[i][j] = findRelativeError(h_i,h_j,errorStream,naive);
						actualLoss += relativeError[i][j].getActualLossOnlyCount();
						estimateLoss += relativeError[i][j].getEstimateLossOnlyCount();
						bothLoss += relativeError[i][j].getBothLossCount();

                		System.out.print(relativeError[i][j].getActualLossOnlyCount()+"/" + relativeError[i][j].getEstimateLossOnlyCount() + "/" + relativeError[i][j].getBothLossCount() + "\t");

                    } catch(FileNotFoundException e)  {
                        System.out.println(e.getMessage());
					}
                }
            }
        }
        System.out.println(actualLoss + "\t" + estimateLoss + "\t" + bothLoss + "\t"
			+ (double)bothLoss/actualLoss * 100.0 + "\t" 
			+ (1.0- (double)bothLoss/estimateLoss)*100.0);
    }

    /**
    * Returns the comparison date used to calculate the relative errors.
    * @return
    *   the comparison date
    **/
    public String getComparisonDate()
    {
        return comparisonDate;
    }

    public MinimalCover getMinimalCover()
    {
	return hostCovering;
    }

    /**
    * Returns a string representing the relative error matrix.
    * @return
    *   a string
    **/
    public String toString()
    {
        int hosts = relativeError.length;
        StringBuffer result = new StringBuffer();

        for(int i=0;i<hosts;i++)
        {
            for(int j=0;j<hosts;j++)
            {
                result.append(relativeError[i][j].getActualLossOnlyCount()+"/" + relativeError[i][j].getEstimateLossOnlyCount() + "/" + relativeError[i][j].getBothLossCount() + "\t");
             //   result.append(hostMap.getHost(i)+" to "+hostMap.getHost(j)+": "+Format.format(relativeError[i][j].getRelativeError(),3)+"\n");
            }

            result.append("\n");
        }

        return result.toString();
    }

    /**************************************************************************
    * Returns the relative error between the estimated and actual RTT times 
    * between two hosts.
    * @param i
    *   an integer representing a host
    * @param j
    *   an integer representing a host
    * @return
    *   a double representing the relative error between the estimated and actual
    *   RTT between the two hosts
    ***************************************************************************/
    public double getRelativeError(Host i, Host j)
    {
	return relativeError[hostMap.hash(i)][hostMap.hash(j)].getRelativeError();
    }

    /**
    * Determines whether the relative error calculated for hosts i and j is an 
    * inter-cluster error, that is, hosts i and j are in the same cluster.
    * @param i
    *   an integer representing a host
    * @param j
    *   an integer representing a host
    * @return
    *   true if the error is an inter-cluster error, otherwise false
    **/
    public boolean isInterClusterError(int i, int j)
    {
	return relativeError[i][j].isInterClusterError();
    }

    /**
    * Determines whether the relative error calculated for hosts i and j is an 
    * intra-cluster error, that is, hosts i and j are in different clusters.
    * @param i
    *   an integer representing a host
    * @param j
    *   an integer representing a host
    * @return
    *   true if the error is an intra-cluster error, otherwise false
    **/
    public boolean isIntraClusterError(int i, int j)
    {
	return relativeError[i][j].isIntraClusterError();
    }

    /**
    * Returns the <CODE>Histogram</CODE> object associated with this relative error 
    * matrix.
    * @return
    *   the associated <CODE>Histogram</CODE> object
    **/
    public Histogram getHistogram()
    {
	return histogram;
    }

    public String getEstimatesVsActuals()
    {
	StringBuffer result = new StringBuffer();
	int hosts = hostCovering.getHostMap().getNumberOfHosts();

	result.append("EstimatesVsActuals:\n");

	double ac25=0;
	double ac50=0;
	double ac75=0;
	double ac100=0;

	double rc25=0;
	double rc50=0;
	double rc75=0;
	double rc100=0;

	for(int i=0;i<hosts;i++)
	    for(int j=0;j<hosts;j++)
	    {
		double are = Math.abs(estimates[i][j]-actuals[i][j])/min(estimates[i][j],actuals[i][j]);
		double rre = relativeError[i][j].getRelativeError();
		result.append("aggregate: "+hostMap.getHost(i)+" -> "+hostMap.getHost(j)+": est="+Format.format(estimates[i][j],1)+" act="+Format.format(actuals[i][j],1)+
			      " are: "+Format.format(are,2)+"\n");
		result.append("other relative error: "+rre+"\n");
		if(are<.25)
		    ac25++;
		if(are<.50)
		    ac50++;
		if(are<.75)
		    ac75++;
		if(are<1)
		    ac100++;

		if(rre<.25)
		    rc25++;
		if(rre<.50)
		    rc50++;
		if(rre<.75)
		    rc75++;
		if(rre<1)
		    rc100++;
	    }
	int t = 106*106;
	result.append("ac25= "+ac25/t+" ac50= "+ac50/t+" ac75= "+ac75/t+" ac100= "+ac100/t+"\n");
	result.append("rc25= "+rc25/t+" rc50= "+rc50/t+" rc75= "+rc75/t+" rc100= "+rc100/t+"\n");
	return result.toString();
    }

    private double min(double i, double j)
    {
	if(i>j)
	    return j;
	else
	    return i;
    }

    public int getNumberOfHosts()
    {
	return relativeError.length;
    }

    public int getLostReadings()
    {
	int total = 0;

	for(int i=0;i<relativeError.length;i++)
	    for(int j=0;j<relativeError.length;j++)
		total += relativeError[i][j].getLostReadings();

	return total;
    }

    public double getLostReadingRatio()
    {
	double den = 0;
	double num = 0;

	for(int i=0;i<relativeError.length;i++)
	    for(int j=0;j<relativeError.length;j++)
	    {
		num += relativeError[i][j].getLostReadings();
		den += relativeError[i][j].getNumberOfReadings();
	    }

	return num/(num+den);
    }

    public double getInterClusterErrorRatio()
    {
	double total = 0;

	for(int i=0;i<relativeError.length;i++)
	    for(int j=0;j<relativeError.length;j++)
	    {
		if(relativeError[i][j].isInterClusterError())
		    total++;
	    }

	return total/(relativeError.length*relativeError.length);
    }
	

    public HostMap getHostMap()
    {
	return hostMap;
    }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
        out.writeObject(relativeError);
        out.writeObject(hostCovering);
        out.writeObject(comparisonDate);
        out.writeObject(baseFilename);
	out.writeObject(histogram);
	out.writeObject(intraEstimationMethod);
	out.writeObject(hostMap);
	out.writeObject(rttMatrix);
	out.writeObject(actuals);
	out.writeObject(estimates);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
        relativeError = (RelativeError[][]) in.readObject();
        hostCovering = (MinimalCover) in.readObject();
        comparisonDate = (String) in.readObject();
        baseFilename = (String) in.readObject();
	histogram = (Histogram) in.readObject();
	intraEstimationMethod = (String) in.readObject();
	hostMap = (HostMap) in.readObject();
	rttMatrix = (RTTMatrix) in.readObject();
	actuals = (double[][]) in.readObject();
	estimates = (double[][]) in.readObject();
    }

	/** Given the file name, read in the NUM_OF_READING entries, and return the latency array.
	 *	Record -1 for empty entry, LOSS constant for "loss", and real values for others.
	 **/
	private int[] readLatency(String fileName, PrintWriter errorStream) {

		int[] lat = new int[NUMBER_OF_READINGS];
        String latLine, value;
		int i=0, loss_num=0;

		try {
        	File latFile = new File(fileName);
        	BufferedReader latReader = new BufferedReader(new InputStreamReader(new FileInputStream(latFile)));

 			while ((latLine = latReader.readLine()) != null && i < NUMBER_OF_READINGS) {

				if (latLine.trim().equals("")) { 
					errorStream.println("Error blank line in "+fileName);
					lat[i] = -1;
		   		} else {
            		value = latLine.substring(latLine.indexOf(' ')+1);
					if (value.equals("loss")) {
						lat[i] = LOSS;
						loss_num++;
					}
					else 	lat[i] = (new Integer(value)).intValue();
				}
				// System.out.print(lat[i] + " ");
				i++;
			}
			if (loss_num == NUMBER_OF_READINGS)  {	// server shutdown, shouldn't be counted!
				for (i=0; i<NUMBER_OF_READINGS; i++)
					lat[i] = -1;
			}
			latReader.close();
		} catch(IOException e) { 
			errorStream.println("File format exception.\n" + e.getMessage()); 
			System.exit(0);     
		}
		return lat;
	}

	/** Simple helper function for anomaly test, assuming normal gmean value (not LOSS) */
	private boolean anomalyTest(int i, int[] lat, Host host1, Host host2) {
		boolean abnormal = true;
		for (int j=0; j<SLIDINGWINDOWSIZE; j++)  {
			// if (host1.getName().equals("amp-okstate") ) 
			// 	System.out.print("lat["+(i-j)+"] = " + lat[i-j] + " ");
			if (lat[i-j] <= gmeanM[host1.getNum()][host2.getNum()]*gstdM[host1.getNum()][host2.getNum()]*gstdM[host1.getNum()][host2.getNum()]) {
			// if (lat[i-j] <= meanM[host1.getNum()][host2.getNum()] + 2*stdM[host1.getNum()][host2.getNum()]) {
				abnormal = false;
				return abnormal;
			}
		}
		// if (host1.getName().equals("amp-okstate") )
		// 	System.out.println("mean: " + meanM[host1.getNum()][host2.getNum()] + " std: " + stdM[host1.getNum()][host2.getNum()]);
		return abnormal;		
	}

	/** Test whether the monitor mon has local anomalies at time i, by checking its connectivity to
	 *	local members, if the majority (say, 80%) has problems, we set the monitor to be abonrmal.
	 *	Note: this may not be an efficient implementaion, but just a simple one.
	 **/
	private boolean anomalyMonTest(int i, Host mon, PrintWriter errorStream)	{

		Cluster c = hostCovering.getCluster(mon);
		Iterator iterator = c.iterator();
		int count = 0;
		int abnormal_count = 0;
		while (iterator.hasNext()) { 
			Host site = (Host)iterator.next();
			if (gmeanM[mon.getNum()][site.getNum()] != LOSS) {
				count++;
				//	read in latency file
				String latFilename = baseFilename + mon.getName() + MIDDLE + site.getName() + SLASH + comparisonDate; 
				int lat[] = readLatency(latFilename, errorStream);
				if (anomalyTest(i, lat, mon, site))	abnormal_count++;
			}
		}
		if ((double)abnormal_count/count >= MON_ANOMALY_THRE)	return true;
		else 	return false;
	}

    /** Note the change from chris' version.
     * A slidingwindow average is taken using <CODE>SlidingWindow </CODE> class
	 *	Assumption: host1 != host2
     **/
    private RelativeError findRelativeError(Host host1, Host host2, PrintWriter errorStream, boolean naive) throws FileNotFoundException
    {
        RelativeError re = null;

        Host monitor1 = hostCovering.getMonitor(host1);
        Host monitor2 = hostCovering.getMonitor(host2);
		int h1_num = host1.getNum();
		int h2_num = host2.getNum();
		int m1_num = monitor1.getNum();
		int m2_num = monitor2.getNum();

		// check if gmean data is available for anomaly test
		//	Otherwise, return a random RelativeError with 0 loss counts
		if (gmeanM[h1_num][h2_num] == LOSS)
	    	return (new RelativeError("inter"));

		if ( (!host1.equals(monitor1)) && (gmeanM[m1_num][h1_num] == LOSS) )
	    	return (new RelativeError("inter"));
		if ( (!monitor1.equals(monitor2)) && (gmeanM[m1_num][m2_num] == LOSS) )
	    	return (new RelativeError("inter"));
		if ( (!host2.equals(monitor2)) && (gmeanM[m2_num][h2_num] == LOSS) )
	    	return (new RelativeError("inter"));

        // handles the case the two monitors are in the same cluster
        if	(monitor1.equals(monitor2))
        {
	    	re = new RelativeError("intra");
			
		   	// if host1 is a monitor, then the distance is exactly measured
			//		read in a->b file and report the # of loss/congestion captured
		   	if  (host1.equals(monitor1))
				for (int i=0; i < (NUMBER_OF_READINGS - SLIDINGWINDOWSIZE + 1); i++)  {
					if (loss[h1_num][h2_num][i]==1)  
						re.incAllLossCount();
				}

			// if host2 is a monitor, we use the distance of (host2 host1), 
			//		read in a->b and b->a and use b->a to estimate the latency 
			//		and loss/congestion captured
			else if (host2.equals(monitor1))    {
				for (int i=0; i < (NUMBER_OF_READINGS - SLIDINGWINDOWSIZE + 1); i++)  {
					if (loss[h1_num][h2_num][i]==1)  
						re.incActualLossOnlyCount();
					if (loss[h2_num][h1_num][i]==1)  {
						re.incEstimateLossOnlyCount();
						if (loss[h1_num][h2_num][i]==1) re.incBothLossCount();
					}
				}	// end of for()
			}
			// Else read in a->b, m->a, and m->b, use 1/2(m->a + m->b) to estimate 
			//		the latency, and evaluate the loss/congestion captured.
			else {
				for (int i=0; i < (NUMBER_OF_READINGS - SLIDINGWINDOWSIZE + 1); i++)  {
					if (loss[h1_num][h2_num][i]==1)  
						re.incActualLossOnlyCount();

					if (!monAbnormal[m1_num][i]) {
						if ( (loss[m1_num][h1_num][i]==1) || (loss[m1_num][h2_num][i]==1) )  {
							re.incEstimateLossOnlyCount();
							if (loss[h1_num][h2_num][i]==1) re.incBothLossCount();
						}
					}	
				}// end of for()
			}
        }
        else    // else the two hosts are in different clusters
	    {
			re = new RelativeError("inter");

		   	// if host1 and host2 r both monitors, then the distance is exactly measured
			//		read in a->b file and report the # of loss/congestion captured
		   	if  ( (host1.equals(monitor1)) && (host2.equals(monitor2)) ) 
				for (int i=0; i < (NUMBER_OF_READINGS - SLIDINGWINDOWSIZE + 1); i++)  {
					if (loss[h1_num][h2_num][i]==1)
						re.incAllLossCount();
				}	// end of for()

			// if only host1 is a monitor, we use the distance of (host1 monitor2), 
			//		read in host1->monitor2 and monitor2->host2, to estimate the latency 
			//		and loss/congestion captured
			else if (host1.equals(monitor1))    {
				for (int i=0; i < (NUMBER_OF_READINGS - SLIDINGWINDOWSIZE + 1); i++)  {
					if (loss[h1_num][h2_num][i]==1)  
						re.incActualLossOnlyCount();

					if (!monAbnormal[m2_num][i])
						if ( (loss[m1_num][m2_num][i]==1) || (loss[m2_num][h2_num][i]==1) )  {
							re.incEstimateLossOnlyCount();
							if (loss[h1_num][h2_num][i]==1) re.incBothLossCount();
						}
				}	// end of for()
			}

			// if only host2 is a monitor, we use the distance of (monitor1, host2), 
			//		read in monitor1->host1 and monitor1->monitor2, to estimate the latency 
			//		and loss/congestion captured
			else if (host2.equals(monitor2))    {
				for (int i=0; i < (NUMBER_OF_READINGS - SLIDINGWINDOWSIZE + 1); i++)  {
					if (loss[h1_num][h2_num][i]==1)  
						re.incActualLossOnlyCount();

					if (!monAbnormal[m1_num][i]) 
						if ( (loss[m2_num][m1_num][i]==1) || (loss[m1_num][h1_num][i]==1) )  {
							re.incEstimateLossOnlyCount();
							if (loss[h1_num][h2_num][i]==1) re.incBothLossCount();
						}
				}	// end of for()
			}

			// Else use dist of (monitor1, monitor2). read in monitor1->host1, 
			//	monitor1->monitor2, and monitor2->host2, and evaluate the loss/congestion captured.
			else {
				for (int i=0; i < (NUMBER_OF_READINGS - SLIDINGWINDOWSIZE + 1); i++)  {
					if (loss[h1_num][h2_num][i]==1)  
						re.incActualLossOnlyCount();

					if (monAbnormal[m1_num][i])  {
						if (!monAbnormal[m2_num][i])  {		// only m1 fails
							if (loss[m2_num][h2_num][i]==1)  {
								re.incEstimateLossOnlyCount();
								if (loss[h1_num][h2_num][i]==1) re.incBothLossCount();
							}
						}	// both fail, do nothing
					} else if (monAbnormal[m2_num][i])  {		// only m2 fails
							if (loss[m1_num][h1_num][i]==1)  {
								re.incEstimateLossOnlyCount();
								if (loss[h1_num][h2_num][i]==1) re.incBothLossCount();
							}
						} else {	// no monitor fail

							if ( (loss[m1_num][m2_num][i]==1) || (loss[m1_num][h1_num][i]==1) || (loss[m2_num][h2_num][i]==1) )  {
								re.incEstimateLossOnlyCount();
								if (loss[h1_num][h2_num][i]==1) re.incBothLossCount();
							}
						}
				}	// end of for()
			}
		}
		/*
		try {
		    //the slidingWinow constructor will perform all calculations and throw exceptions if error encounterd
		    
		    SlidingWindow slidingWindow = new SlidingWindow(host1, host2,monitorFilename, hostFilename, relativeError, hostReader, monitorReader,errorStream, naive, rttMatrix);
		    
		    actuals[hostMap.hash(host1)][hostMap.hash(host2)] = slidingWindow.getActual();
		    estimates[hostMap.hash(host1)][hostMap.hash(host2)] = slidingWindow.getEstimate();
	    
		    if (slidingWindow.getNumberOfReadings() == 0) 
			relativeError.setRelativeError(-1.0);
		    else 
			relativeError.setRelativeError(slidingWindow.getTotal()/slidingWindow.getNumberOfReadings());
	    
		    relativeError.setNumberOfReadings(slidingWindow.getNumberOfReadings());
		}
		*/

        // System.out.println(re.getActualLossOnlyCount()+"/" + re.getEstimateLossOnlyCount() + "/" + re.getBothLossCount());
	   	return re;
    }

    /** 
     * Note the change in the metric for calculation
     **/
	
    
    private double relativeErrorCalc(String estimate, String actual, RelativeError relativeError,
				     PrintWriter errorStream)
    {
	if(estimate.equals("zero"))
	    return 0.0;
	
        if(estimate.equals("loss"))
        {
            if(actual.equals("loss"))
		{
		    relativeError.incBothLossCount();
		    return 0.0;
            }
            else
            {
                relativeError.incEstimateLossOnlyCount();
                return -100.0;
            }
        }
        else
        {
            if(actual.equals("loss"))
            {
                relativeError.incActualLossOnlyCount();
                return -100.0;
            }
             else
             {
                try
                {
                    double est = (new Double(estimate)).doubleValue();
                    double act = (new Double(actual)).doubleValue();
                    // if actual is 0, make it 1
                    if(act<1) act = 1.0;
		    ///
		    /*
                    return (Math.abs(est-act))/((act<est) ? act : est);
		    */
		    return (Math.abs(log2(est) - log2(act)));
		    
		}
		
                catch(NumberFormatException e)
                {
                    errorStream.println("Number format exception. \n" + e.getMessage());
                    errorStream.println("Est='"+estimate+"'");
		    errorStream.println("Act='"+actual+"'");
		    throw new IllegalArgumentException();
                }
            }
        }
    }

    private double relativeErrorCalcNaive(String estimate, String actual, RelativeError relativeError,
					  PrintWriter errorStream)
    {
	if(estimate.equals("zero"))
	    return 0.0;

        if(estimate.equals("loss"))
        {
            if(actual.equals("loss"))
            {
                relativeError.incBothLossCount();
                return 0.0;
            }
            else
            {
                relativeError.incEstimateLossOnlyCount();
                return -100.0;
            }
        }
        else
        {
            if(actual.equals("loss"))
            {
                relativeError.incActualLossOnlyCount();
                return -100.0;
            }
             else
             {
                try
                {
                    double est = (new Double(estimate)).doubleValue();
                    double act = (new Double(actual)).doubleValue();
                    // if actual is 0, make it 1
                    if(act<1) act = 1.0;
		    ///
		    /*
                    return (Math.abs(est-act))/((act<est) ? act : est);
		    */
		    return (Math.abs(log2(est) - log2(act)));
		    
                }

                catch(NumberFormatException e)
                {
                    errorStream.println("Number format exception. \n" + e.getMessage());
                    errorStream.println("Est='"+estimate+"'");
		    errorStream.println("Act='"+actual+"'");
		    throw new IllegalArgumentException();
                }
            }
        }
    }
    public static double log2 (double x) 
    {
	return (Math.log(x) / Math.log(2));
    }
    
    
}















