package cluster;

import java.util.*;
import java.io.*;

/** 
 * Creates a <CODE>SlidingWindow </CODE> that will contain the readings that are read into the 2 <CODE> Reading </CODE> arrays
 * Will then perform the slidingwindow comparison and finally have the results in actual_total and estimate_total
 **/


public class SlidingWindow implements WNMMSConstants{
    
    //we will change the following values for any valid pair of estimate and actual values

    private double actual_total;
    private double estimate_total;
    private int numberOfReadings;
    private double total; //refers to the total error
    
    private Reading[] estimationReadings;
    private Reading[] actualReadings;
    
    private BufferedReader hostReader;
    private BufferedReader monitorReader;
    private PrintWriter errorStream;
    private boolean naive;
    private RTTMatrix rttMatrix;
    private RelativeError relativeError;
    private String monitorFilename;
    private String hostFilename;
    private Host host1; 
    private Host host2;

    
    public SlidingWindow(Host host1, Host host2, String monitorFilename, String hostFilename, RelativeError relativeError, BufferedReader hostReader, BufferedReader monitorReader, PrintWriter errorStream, boolean naive, RTTMatrix rttMatrix) throws IOException , IllegalArgumentException{
	
	this.host1 = host1;
	this.host2 = host2;
	this.monitorFilename = monitorFilename;
	this.hostFilename = hostFilename;
	this.relativeError = relativeError;
	this.hostReader = hostReader;
	this.monitorReader = monitorReader;
	this.errorStream = errorStream;
	this.naive = naive;
	this.rttMatrix = rttMatrix;
	
	estimationReadings = new Reading[NUMBER_OF_READINGS];
	actualReadings = new Reading[NUMBER_OF_READINGS];
	for (int i = 0; i< NUMBER_OF_READINGS; i++) {
	    
	    estimationReadings[i] = new Reading(); //initializes both array to array of readings
	    actualReadings[i] = new Reading();
	}
	
	produceValues(hostReader,"host");
	produceValues(monitorReader, "monitor");
	//print();
	///
	//System.out.println("monitorFilename is " + monitorFilename);
	///
	//System.out.println("hostFilename is " + hostFilename);
	performSlidingComparison();
	
    }

    /** Reads in all the values from the reader and updates the Reading arrays accordingly
     * will put each reading into the proper minute slot 
     * Reading object in the array will take average of the readings in case several readings map to the same min
     **/
    private void produceValues ( BufferedReader reader, String arg) throws IOException, IllegalArgumentException {
	
	String line;
	while ((line = reader.readLine()) != null) {
	    if (line.trim().equals("")) {
		errorStream.println("Error blank line for " + host1 + " vs " + host2 );
		continue;
	    }
	    String timeString = line.substring(0, line.indexOf(' '));
	    String valueString = line.substring(line.indexOf(' ') + 1);
		

	    ///
	    //System.out.println("timeSTring is " + timeString);
	    ///
	    //System.out.println("valueString is " + valueString);

	    //round timeString to approximate min
	    double time = Double.parseDouble(timeString);	  
	    Double timeInMin= new Double (time/60);
	    int floor = timeInMin.intValue();
	    
	    ///
	    //System.out.println("floor is " + floor);
	    if (floor < 0 || floor > NUMBER_OF_READINGS - 1) {
		errorStream.println("Error time value < 0 or > " + (NUMBER_OF_READINGS - 1) );
		continue;
	    }
	    
	    if (arg.equals("host")) {
		actualReadings[floor].addReading(valueString);
	    }
	    else if (arg.equals("monitor")){
		estimationReadings[floor].addReading(valueString);
		///
		//System.out.println(valueString);
	    }
	    else {
		throw new IllegalArgumentException("argument in produce values are either host or monitor");
	    }
	}
	
	
    }

    public void print() {
	System.out.println("monitorFilename is " + monitorFilename);
	System.out.println("hostFilename is " + hostFilename);
	System.out.println("**");
	for (int i = 0; i < NUMBER_OF_READINGS; i++) {
	    System.out.println("i = " + i + " estimate = " + estimationReadings[i].getValue() + " actual = " + actualReadings[i].getValue());
	    
	}
	System.out.println("***");
    }
    
    /**
     * The bulk of the SlidingWindow
     * Will take the gmean average over the slidingwindow of values
     * The estimation and actual total will be updated
     * The total error count so far will be incrememted accordingly
     **/

    private void performSlidingComparison() throws IOException {
	numberOfReadings = 0;
	actual_total = 0;
	estimate_total = 0;
	
	for (int i = 0 ; i < NUMBER_OF_READINGS; i++) {
	    if (!actualReadings[i].isValid()) {
		continue; // invalid entry
	    }
	    
	    boolean anyValidValues = false; 
	    int slides= 0;
	    double slidingTotal = 0;

	    //get the sliding average
	    // if by the end of the slide we still have no valid values
	    // we will ignore the value
	    
	    for (int j = 0 ; j <= SLIDINGWINDOWSIZE && (i-j) >= 0 ; j++) {
		if (estimationReadings[i-j].isValid()) {
		    anyValidValues = true;
		    slidingTotal += log10(estimationReadings[i-j].getValue());
		    ///
		    //System.out.println("log 10 is then " + log10(estimationReadings[i-j].getValue()));
		    slides++;
		}
	    }

	    if (!anyValidValues) continue; //no valid values 
	    double estimation = Math.pow(10, slidingTotal/slides);

	    
	    if (naive) 
		estimation = rttMatrix.get(host1, host2);
	    
	    ///
	    //System.out.println("slides = " + slides + "anyValidValues = " + anyValidValues + "estimation = " + estimation + "actual = " + actualReadings[i].getValue());
	    
	    
	    actual_total += actualReadings[i].getValue();
	    estimate_total += estimation;
	    numberOfReadings++; 
	    total += calcError(actualReadings[i].getValue(), estimation);
	}
	///
	//System.out.println("Numberof readings = "+ numberOfReadings);

	hostReader.close();
	monitorReader.close();
    }

    /**
     * The metric for calculation has been changed from the previous
     **/

    private double calcError(double act , double est) {
	if(act<1) act = 1.0;
	///
	/*
	  return (Math.abs(est-act))/((act<est) ? act : est);
	*/
	return (Math.abs(log2(est) - log2(act)));
    }



    private  double log2 (double x) 
    {
	return (Math.log(x) / Math.log(2));
    }
    
    public double getActual() {
	return actual_total/numberOfReadings;
    }
    
    public double getEstimate() {
	return estimate_total/numberOfReadings;
    }
    
    public double getTotal() {
	return total;
    }

    public int getNumberOfReadings() {
	return numberOfReadings;
    }
    
    public static double log10 (double x) 
    {
	return (Math.log(x) / Math.log(10));
    }
    
}
