package cluster;

import java.io.*;

/*************************************************************************************
* The <CODE>RelativeErrorMatrix</CODE> class is used to represent the matrix of
* relative errors between monitors and hosts on a specific date.
*
* 11/25/02 modify it to evaluate the ability of congestion/failure detection. 
**************************************************************************************/

public class RelativeErrorMatrix implements Serializable, WNMMSConstants
{

    protected RelativeError[][] relativeError;
    private MinimalCover hostCovering;
    protected String comparisonDate;
    protected String baseFilename;
    protected Histogram histogram;
    protected String intraEstimationMethod;
    private HostMap hostMap;
    private RTTMatrix rttMatrix;

    private double[][] estimates;
    private double[][] actuals;

    public RelativeErrorMatrix()
    {
    }

    /**
    * Creates a new <CODE>RelativeErrorMatrix</CODE> using the clustering represented by
    * <CODE>hostCovering</CODE> on the <CODE>comparisonDate</CODE>. The format of the
    * date is: 101.6.26 -> June 26, 2001
    * @param hostCovering
    *   a <CODE>MinimalCover</CODE> representing the partition of the hosts
    * @param comparisonDate
    *   the date which the relative errors are to be calculated
    **/
    public RelativeErrorMatrix(MinimalCover hostCovering, String comparisonDate, String intraEstimationMethod, boolean naive)
    {
        int hosts = hostCovering.getNumberOfHosts();
        this.relativeError = new RelativeError[hosts][hosts];
        this.baseFilename = ".." + SLASH + "data" + SLASH + "raw-data" + SLASH + comparisonDate + SLASH;
        this.comparisonDate = comparisonDate;
	this.intraEstimationMethod = intraEstimationMethod;
        this.hostCovering = hostCovering;
	this.hostMap = hostCovering.getHostMap();
	this.rttMatrix = hostCovering.getRTTMatrix();
	
	estimates = new double[hosts][hosts];
	actuals = new double[hosts][hosts];

	try
	{
	    PrintWriter errorStream = new PrintWriter(new FileOutputStream(comparisonDate + ".error"));
	    this.doRelativeErrorCalc(errorStream,naive);
	    errorStream.flush();
	    errorStream.close();
	}

	catch(FileNotFoundException e)
	{
	    System.out.println(e.getMessage());
	    System.exit(0);
	}
	this.histogram = new Histogram(this);
    }
    
    private void doRelativeErrorCalc(PrintWriter errorStream, boolean naive)
    {
		int hosts = hostCovering.getNumberOfHosts();

        for(int i=0;i<hosts;i++)
        {
            Host h_i = hostMap.getHost(i);

            for(int j=0;j<hosts;j++)
            {

                if (i == j)	
					// the parameter "inter" here doesn't matter, just initize the value to be 0
                    relativeError[i][j] = new RelativeError("inter");
                else
				{                
                    Host h_j = hostMap.getHost(j);
                    try {
                        relativeError[i][j] = findRelativeError(h_i,h_j,errorStream,naive);
                    } catch(FileNotFoundException e)  {
                        System.out.println(e.getMessage());
					}
                }
            }
            System.out.println(h_i);
        }
    }

    /**
    * Returns the comparison date used to calculate the relative errors.
    * @return
    *   the comparison date
    **/
    public String getComparisonDate()
    {
        return comparisonDate;
    }

    public MinimalCover getMinimalCover()
    {
	return hostCovering;
    }

    /**
    * Returns a string representing the relative error matrix.
    * @return
    *   a string
    **/
    public String toString()
    {
        int hosts = relativeError.length;
        StringBuffer result = new StringBuffer();

        for(int i=0;i<hosts;i++)
        {
            for(int j=0;j<hosts;j++)
            {
                result.append(hostMap.getHost(i)+" to "+hostMap.getHost(j)+": "+Format.format(relativeError[i][j].getRelativeError(),3)+"\n");
            }

            result.append("\n");
        }

        return result.toString();
    }

    /**************************************************************************
    * Returns the relative error between the estimated and actual RTT times 
    * between two hosts.
    * @param i
    *   an integer representing a host
    * @param j
    *   an integer representing a host
    * @return
    *   a double representing the relative error between the estimated and actual
    *   RTT between the two hosts
    ***************************************************************************/
    public double getRelativeError(Host i, Host j)
    {
	return relativeError[hostMap.hash(i)][hostMap.hash(j)].getRelativeError();
    }

    /**
    * Determines whether the relative error calculated for hosts i and j is an 
    * inter-cluster error, that is, hosts i and j are in the same cluster.
    * @param i
    *   an integer representing a host
    * @param j
    *   an integer representing a host
    * @return
    *   true if the error is an inter-cluster error, otherwise false
    **/
    public boolean isInterClusterError(int i, int j)
    {
	return relativeError[i][j].isInterClusterError();
    }

    /**
    * Determines whether the relative error calculated for hosts i and j is an 
    * intra-cluster error, that is, hosts i and j are in different clusters.
    * @param i
    *   an integer representing a host
    * @param j
    *   an integer representing a host
    * @return
    *   true if the error is an intra-cluster error, otherwise false
    **/
    public boolean isIntraClusterError(int i, int j)
    {
	return relativeError[i][j].isIntraClusterError();
    }

    /**
    * Returns the <CODE>Histogram</CODE> object associated with this relative error 
    * matrix.
    * @return
    *   the associated <CODE>Histogram</CODE> object
    **/
    public Histogram getHistogram()
    {
	return histogram;
    }

    public String getEstimatesVsActuals()
    {
	StringBuffer result = new StringBuffer();
	int hosts = hostCovering.getHostMap().getNumberOfHosts();

	result.append("EstimatesVsActuals:\n");

	double ac25=0;
	double ac50=0;
	double ac75=0;
	double ac100=0;

	double rc25=0;
	double rc50=0;
	double rc75=0;
	double rc100=0;

	for(int i=0;i<hosts;i++)
	    for(int j=0;j<hosts;j++)
	    {
		double are = Math.abs(estimates[i][j]-actuals[i][j])/min(estimates[i][j],actuals[i][j]);
		double rre = relativeError[i][j].getRelativeError();
		result.append("aggregate: "+hostMap.getHost(i)+" -> "+hostMap.getHost(j)+": est="+Format.format(estimates[i][j],1)+" act="+Format.format(actuals[i][j],1)+
			      " are: "+Format.format(are,2)+"\n");
		result.append("other relative error: "+rre+"\n");
		if(are<.25)
		    ac25++;
		if(are<.50)
		    ac50++;
		if(are<.75)
		    ac75++;
		if(are<1)
		    ac100++;

		if(rre<.25)
		    rc25++;
		if(rre<.50)
		    rc50++;
		if(rre<.75)
		    rc75++;
		if(rre<1)
		    rc100++;
	    }
	int t = 106*106;
	result.append("ac25= "+ac25/t+" ac50= "+ac50/t+" ac75= "+ac75/t+" ac100= "+ac100/t+"\n");
	result.append("rc25= "+rc25/t+" rc50= "+rc50/t+" rc75= "+rc75/t+" rc100= "+rc100/t+"\n");
	return result.toString();
    }

    private double min(double i, double j)
    {
	if(i>j)
	    return j;
	else
	    return i;
    }

    public int getNumberOfHosts()
    {
	return relativeError.length;
    }

    public int getLostReadings()
    {
	int total = 0;

	for(int i=0;i<relativeError.length;i++)
	    for(int j=0;j<relativeError.length;j++)
		total += relativeError[i][j].getLostReadings();

	return total;
    }

    public double getLostReadingRatio()
    {
	double den = 0;
	double num = 0;

	for(int i=0;i<relativeError.length;i++)
	    for(int j=0;j<relativeError.length;j++)
	    {
		num += relativeError[i][j].getLostReadings();
		den += relativeError[i][j].getNumberOfReadings();
	    }

	return num/(num+den);
    }

    public double getInterClusterErrorRatio()
    {
	double total = 0;

	for(int i=0;i<relativeError.length;i++)
	    for(int j=0;j<relativeError.length;j++)
	    {
		if(relativeError[i][j].isInterClusterError())
		    total++;
	    }

	return total/(relativeError.length*relativeError.length);
    }
	

    public HostMap getHostMap()
    {
	return hostMap;
    }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
        out.writeObject(relativeError);
        out.writeObject(hostCovering);
        out.writeObject(comparisonDate);
        out.writeObject(baseFilename);
	out.writeObject(histogram);
	out.writeObject(intraEstimationMethod);
	out.writeObject(hostMap);
	out.writeObject(rttMatrix);
	out.writeObject(actuals);
	out.writeObject(estimates);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
        relativeError = (RelativeError[][]) in.readObject();
        hostCovering = (MinimalCover) in.readObject();
        comparisonDate = (String) in.readObject();
        baseFilename = (String) in.readObject();
	histogram = (Histogram) in.readObject();
	intraEstimationMethod = (String) in.readObject();
	hostMap = (HostMap) in.readObject();
	rttMatrix = (RTTMatrix) in.readObject();
	actuals = (double[][]) in.readObject();
	estimates = (double[][]) in.readObject();
    }

    /** Note the change from chris' version.
     * A slidingwindow average is taken using <CODE>SlidingWindow </CODE> class
     **/
    
    private RelativeError findRelativeError(Host host1, Host host2, PrintWriter errorStream, boolean naive) throws FileNotFoundException
    {
        RelativeError relativeError = null;
        Host monitor1 = hostCovering.getMonitor(host1);
        Host monitor2 = hostCovering.getMonitor(host2);

        String hostFilename = baseFilename + host1.getName() + MIDDLE + host2.getName() + SLASH +
                                comparisonDate;
        String monitorFilename = null;
        File hostFile = new File(hostFilename);
        BufferedReader hostReader = new BufferedReader(new InputStreamReader(
									     new FileInputStream(hostFile)));

        double total = 0.0;
        int numberOfReadings = 0;

		double actual_total = 0.0;
		double estimate_total = 0.0;

        // handles the case the two monitors are in the same cluster
        if	(monitor1.equals(monitor2))
        {
	    	relativeError = new RelativeError("intra");
            String estimate = null;
	    	String actual = null;
	    	int estimateIndex=0;

            try
            {
                String hostLine;
				estimateIndex = 0; // used for indexing into the avgRTTMap for estimating intra errors

                while((hostLine = hostReader.readLine()) != null && estimateIndex < NUMBER_OF_READINGS) {
		    		if(hostLine.trim().equals("")) {
						errorStream.println("Error blank line in "+hostFilename);
						estimateIndex++;
						continue;
		    		}
          
		    		// if host1 is a monitor, then the distance is exactly measured
					// if host2 is a monitor, we use the distance of (host2 host1), i.e., the
					//	distance above as an estimation
					// Otherwise, we use the 
		    		// the average RTT between all cluster members and the
		    		// center. otherwise the estimate is double the above value.
		    		if  (host1.equals(monitor1) || host2.equals(monitor2))
						estimate = "zero";
		    		else 	{
						double e = (hostCovering.getCluster(monitor1)).findIntraClusterEstimate(comparisonDate,host1,host2,estimateIndex,intraEstimationMethod,errorStream);
						if(e < 0) {
			    			estimate = "loss";
			    			//			    System.out.println("loss in intracluster error");
			    			//System.out.println("m="+monitor1+" h1="+host1+" h2="+host2+" estindex="+estimateIndex+" compdate="+comparisonDate);
						} else
			    			estimate = "" + e;
		    		}
		
		    		if(naive)
						estimate = rttMatrix.get(host1,host2)+"";
    
		    		estimateIndex++;
		    		// old way for finding intra cluster estimate is below 
					// it was simply average RTT from center 
					//hostCovering.getCluster(monitor1).getAvgRTTFromCenter() + "";

                    actual = hostLine.substring(hostLine.indexOf(' ')+1);
                    double error = relativeErrorCalc(estimate,actual,relativeError,errorStream);

                    // error < 0 indicates loss at either monitor or host, but not both
                    if(error > -1) {
						if(!(estimate.equals("loss") || estimate.equals("zero")))
			    		{
			    			actual_total += (new Double(actual)).doubleValue();
			    			estimate_total += (new Double(estimate)).doubleValue();
			    		}
                        total += error;
                        numberOfReadings++;
                    }
                }

                hostReader.close();
            }

            catch(IOException e)
            {
		errorStream.println("File format exception.\n" + e.getMessage());
		System.exit(0);
            }
	    catch(Exception e)
	    {
		errorStream.println("hfn="+hostFilename+"\nmfn="+monitorFilename);
		errorStream.println("estimate="+estimate);
		errorStream.println("actual="+actual);
		errorStream.println("nor="+numberOfReadings);
		errorStream.println("estimateIndex="+estimateIndex);
		e.printStackTrace(errorStream);
	    }

	    
	    actuals[hostMap.hash(host1)][hostMap.hash(host2)] = actual_total/numberOfReadings;
	    estimates[hostMap.hash(host1)][hostMap.hash(host2)] = estimate_total/numberOfReadings;
	    
	    if(numberOfReadings == 0)
			relativeError.setRelativeError(-1.0);
	    else
			relativeError.setRelativeError(total/numberOfReadings);
	    
	    relativeError.setNumberOfReadings(numberOfReadings);
	    
	    return relativeError;
        }
        else    // else the two hosts are in different clusters
	    {
		relativeError = new RelativeError("inter");
		monitorFilename = baseFilename + monitor1.getName() + MIDDLE + monitor2.getName() +
		    SLASH + comparisonDate;
		
		File monitorFile = new File(monitorFilename);

		BufferedReader monitorReader = new BufferedReader(new InputStreamReader(
											new FileInputStream(monitorFile)));
	    
		
		
		
		try {
		    //the slidingWinow constructor will perform all calculations and throw exceptions if error encounterd
		    
		    SlidingWindow slidingWindow = new SlidingWindow(host1, host2,monitorFilename, hostFilename, relativeError, hostReader, monitorReader,errorStream, naive, rttMatrix);
		    
		    actuals[hostMap.hash(host1)][hostMap.hash(host2)] = slidingWindow.getActual();
		    estimates[hostMap.hash(host1)][hostMap.hash(host2)] = slidingWindow.getEstimate();
	    
		    if (slidingWindow.getNumberOfReadings() == 0) 
			relativeError.setRelativeError(-1.0);
		    else 
			relativeError.setRelativeError(slidingWindow.getTotal()/slidingWindow.getNumberOfReadings());
	    
		    relativeError.setNumberOfReadings(slidingWindow.getNumberOfReadings());
		    
	    
		}
		catch(IOException e)
		    {
			errorStream.println("File format exception.\n" + e.getMessage());
			System.exit(0);
		    }

		catch(IllegalArgumentException e)
		    {
			errorStream.println("Error in parsing file:");
			errorStream.println("hfn="+hostFilename+"\nmfn="+monitorFilename);
			errorStream.println("nor="+numberOfReadings);
		    }
		return relativeError;
		
	    }
    }

    /** 
     * Note the change in the metric for calculation
     **/
	
    
    private double relativeErrorCalc(String estimate, String actual, RelativeError relativeError,
				     PrintWriter errorStream)
    {
	if(estimate.equals("zero"))
	    return 0.0;
	
        if(estimate.equals("loss"))
        {
            if(actual.equals("loss"))
		{
		    relativeError.incBothLossCount();
		    return 0.0;
            }
            else
            {
                relativeError.incEstimateLossOnlyCount();
                return -100.0;
            }
        }
        else
        {
            if(actual.equals("loss"))
            {
                relativeError.incActualLossOnlyCount();
                return -100.0;
            }
             else
             {
                try
                {
                    double est = (new Double(estimate)).doubleValue();
                    double act = (new Double(actual)).doubleValue();
                    // if actual is 0, make it 1
                    if(act<1) act = 1.0;
		    ///
		    /*
                    return (Math.abs(est-act))/((act<est) ? act : est);
		    */
		    return (Math.abs(log2(est) - log2(act)));
		    
		}
		
                catch(NumberFormatException e)
                {
                    errorStream.println("Number format exception. \n" + e.getMessage());
                    errorStream.println("Est='"+estimate+"'");
		    errorStream.println("Act='"+actual+"'");
		    throw new IllegalArgumentException();
                }
            }
        }
    }

    private double relativeErrorCalcNaive(String estimate, String actual, RelativeError relativeError,
					  PrintWriter errorStream)
    {
	if(estimate.equals("zero"))
	    return 0.0;

        if(estimate.equals("loss"))
        {
            if(actual.equals("loss"))
            {
                relativeError.incBothLossCount();
                return 0.0;
            }
            else
            {
                relativeError.incEstimateLossOnlyCount();
                return -100.0;
            }
        }
        else
        {
            if(actual.equals("loss"))
            {
                relativeError.incActualLossOnlyCount();
                return -100.0;
            }
             else
             {
                try
                {
                    double est = (new Double(estimate)).doubleValue();
                    double act = (new Double(actual)).doubleValue();
                    // if actual is 0, make it 1
                    if(act<1) act = 1.0;
		    ///
		    /*
                    return (Math.abs(est-act))/((act<est) ? act : est);
		    */
		    return (Math.abs(log2(est) - log2(act)));
		    
                }

                catch(NumberFormatException e)
                {
                    errorStream.println("Number format exception. \n" + e.getMessage());
                    errorStream.println("Est='"+estimate+"'");
		    errorStream.println("Act='"+actual+"'");
		    throw new IllegalArgumentException();
                }
            }
        }
    }
    public static double log2 (double x) 
    {
	return (Math.log(x) / Math.log(2));
    }
    
    
}















