package cluster.stability;

import cluster.*;
import java.io.*;
import java.util.*;

public class Stability
{
    private static final String aggregationType = "gmean";
    private static final String baseFilename = ".." + File.separator + "data" +
         File.separator + aggregationType + "-data" + File.separator + aggregationType + "-";
    private static final String fileSuffix = ".dat";

    private int numberOfDays;
    private int hosts;
    private String startDateString;
    private double tolerance;
    private ClusterVector[] days;
    private FrequencyVector[] frequency;
    private HostMap hostmap;
    private boolean[] covered;
    private Cluster[] stableClusters;
    private HashSet coverset;
    private MinimalCover minimalCover;

    public Stability(String sitesFile, String startDateString, int numberOfDays, double tolerance, int hosts)
    {
	int startMonth = (new Integer(startDateString.substring(0,2))).intValue() - 1;
	int startDay = (new Integer(startDateString.substring(2,4))).intValue();
	int startYear = 2000 + (new Integer(startDateString.substring(4,6))).intValue();
	GregorianCalendar iteratorDate = new GregorianCalendar(startYear,startMonth,startDay);

	days = new ClusterVector[numberOfDays];
	this.hosts = hosts;
        hostmap = new HostMap(sitesFile,hosts);

    	for(int i=0;i<numberOfDays;i++)
    	{
	    String currentDate = formatDate(iteratorDate.get(Calendar.MONTH)+1) +
		formatDate(iteratorDate.get(Calendar.DAY_OF_MONTH)) +
		formatDate(iteratorDate.get(Calendar.YEAR) % 2000);
	    String dataFile = baseFilename + currentDate + fileSuffix;

	    RTTMatrix rttMatrix = new RTTMatrix(dataFile,hostmap,aggregationType,hosts);
	    LimitRadiusClustering c = new LimitRadiusClustering(rttMatrix,tolerance,"none",0.0,0);
	    days[i] = new ClusterVector(c.getClusters(),currentDate,rttMatrix.getHostMap());
	    iteratorDate.add(Calendar.DAY_OF_MONTH,1);
	}

	frequency = new FrequencyVector[hosts];

	for(int i=0;i<hosts;i++)
	{
	    Host center = hostmap.getHost(i);
	    frequency[i] = new FrequencyVector(days,center);
	}

	stableClusters = new Cluster[hosts];
	for(int i=0;i<hosts;i++)
	{
	    stableClusters[i] = new Cluster(hostmap.getHost(i));
	    Iterator iterator = frequency[i].getHosts().iterator();
	    while(iterator.hasNext())
	    {
		Host host = (Host) iterator.next();
		if(frequency[i].getFrequency(host) >= .5)
		    stableClusters[i].add(host);
	    }
	}
		
	covered = new boolean[hosts];
	for(int i=0;i<hosts;i++) {
	    covered[i] = false;
	}

	coverset = greedyMinCover(stableClusters);
	minimalCover = new MinimalCover(coverset,stableClusters,frequency,hostmap,"none",0.0,0);
    }

    public RelativeErrorMatrix getRelativeErrorMatrix(String date, String intraEstimationMethod)
    {
	return new RelativeErrorMatrix(minimalCover,date,intraEstimationMethod,false);
    }

    public MinimalCover getMinimalCover()
    {
	return minimalCover;
    }

    public Cluster[] getClusters()
    {
	return stableClusters;
    }

    public String toString()
    {
	StringBuffer result = new StringBuffer();

	for(int i=0;i<frequency.length;i++)
	    result.append(frequency[i] + "\n");

	return result.toString();
    }

    public String daysToString()
    {
	StringBuffer result = new StringBuffer();

	for(int i=0;i<hosts;i++)
	    for(int j=0;j<days.length;j++)
		result.append("day "+j+": "+days[j].getCluster(i) + "\n");

	return result.toString();
    }

    private HashSet greedyMinCover(Cluster[] clusters)
    {
	HashSet picked = new HashSet();
	HashSet result = new HashSet();

	for(int j=0;j<hosts;j++) {
	    int victim = victimize(picked,clusters);

	    if(victim == -1)
		continue;

	    picked.add(hostmap.getHost(victim));
	    Cluster t = clusters[victim];

	    Iterator iterator = t.iterator();
	    while(iterator.hasNext())
	    {
		covered[hostmap.hash((Host) iterator.next())] = true;
	    }

	    result.add(hostmap.getHost(victim));
	}

	return result;
    }

  // Find the next host which has the most uncovered hosts
    private int victimize(HashSet pick, Cluster[] clusters) 
    {
	int max = 0;
	int picked = -1;
	for(int i=0;i<hosts;i++) {
	    if(pick.contains(hostmap.getHost(i)))
		continue;
	    Cluster test = clusters[i];
	    int count = 0;

	    Iterator iterator = test.iterator();
	    while(iterator.hasNext())
	    {
		Host intey = (Host) iterator.next();
		if(!covered[hostmap.hash(intey)])
		    count++;
	    }

	    // if the host isn't already covered, or wasn't already
	    // accounted for, increase the count because a host
	    // can cover itself
	    if(!covered[i] && !test.contains(hostmap.getHost(i)))
		count++;

	    if(count > max) {
		picked = i;
		max = count;
	    }
	}
	//forme = forme + max;

	return picked;
    }

    private String formatDate(int field)
    {
	if(field<10)
	    return "0" + field;
	else
	    return "" + field;
    }
}



















