% [xG] = compute_xG(Gbar, Rbar, bbar, param)
%
% Compute the loss rate vector xG from Gbar, Rbar, and bbar.
% Gbar and Rbar are produced through rowselect, and bbar is measured.
% The param structure must be compatible with the one passed to
% rowselect (see description of the options in rowselect).
%
% Parameters:
%
%   layout (default 'full'): storage layout for the R factor.  Can
%     be 'full' or 'packed'.  The 'full' format is an ordinary
%     column major layout; the 'packed' format is a blocked
%     column-major packed layout.  If kb is the block size, then
%     kb elements are stored from each of columns 1:kb; 2*kb
%     elements are stored from columns 2*kb+1:3*kb; and so forth.
%
%   verbose (default 1): zero for minimal printed output, one for
%     more printed output.
%
%   blockR (default 100): blocking factor used in the packed layout
%     of R (see the description of the layout parameter).

function [xG, stat_rank] = compute_xG(G, R, b, param)

if  nargin < 4,                      param = [];                end
if ~isfield(param, 'layout'),        param.layout  = 'full';    end
if ~isfield(param, 'verbose'),       param.verbose = 1;         end
if ~isfield(param, 'blockR'),        param.blockR  = 100;       end
if ~strcmp(param.layout, 'packed'),  param.blockR  = 0;         end

if isa(R, 'single'), b1 = single(b); end
y1 = trsolve(R, b1, 'T', param.blockR);
y2 = trsolve(R, y1, 'N', param.blockR);
if isa(R, 'single'), y2 = double(y2); end
xG = G'*y2;

% A few steps of iterative refinement for sanity's sake
rnew = G * xG - b;
for k = 1:15
  r = rnew;
  if isa(R, 'single'), b2 = single(r); end
  y1 = trsolve(R, b2,  'T', param.blockR);
  y2 = trsolve(R, y1, 'N', param.blockR);
  if isa(R, 'single'), y2 = double(y2); end
  xG = xG - G'*y2;
  rnew = G * xG - b;
  if (norm(rnew) / norm(r) > 0.5)
    break;
  end
end

fprintf('Rank');
stat_rank = length( y2 );
%stat_rank = rank( G );











