/*
 * function v = pathv(to, p, labels, ne, I)
 * 
 * Generate path vectors v to specified destinations
 * Inputs:
 *   to      - path destinations
 *   p       - predecessor vector from Dijkstra
 *   labels  - edge identifier adjacency matrix: 
 *             labels(i,j) = id of edge between node i and node j
 *   ne      - number of edges
 *
 * If I is passed, it is used to mark nodes used.
 *
 * If to has length one, this code should be equivalent to the
 * following Matlab code:
 *
 *   n = [to];
 *   while (p(to) > 0)
 *     to = p(to);
 *     n  = [to n];
 *   end
 *
 *   v = spalloc(ne,1, length(n));
 *   for k = 1:length(n)-1
 *     id = A(n(k), n(k+1));
 *     v(id) = 1;
 *   end
 */

#include <mex.h>

/* Integer comparison for qsort */
int int_compare(const void* ap, const void* bp)
{
    int a = *(int*) ap;
    int b = *(int*) bp;
    if (a < b)
        return -1;
    if (a > b)
        return 1;
    return 0;
}

int get_edge_num(const mxArray* A, int src, int dest)
{
    int i, j;

    /* Unpack A matrix */
    int*    ir = mxGetIr(A);
    int*    jc = mxGetJc(A);
    double* Ar = mxGetPr(A);

    /* Adjust indices to zero base */
    --src;
    --dest;

    /* Find the edge in the adjacency matrix */
    for (i = jc[src]; i < jc[src+1]; ++i) {
        if (ir[i] == dest)
	    return (int) Ar[i];
    }

    printf("No edge found\n");
    return 0;
}

void mexFunction(int nlhs, mxArray** plhs,
                 int nrhs, const mxArray** prhs)
{
    int i, j;
    double* to;
    double* p;
    int     nto;
    int     nnz;
    int     ne;
    int     nv;
    int*    ir;
    int*    jc;
    double* vpr;
    double* node_used;

    /* Process arguments */

    if (nrhs < 4)
        mexErrMsgTxt("Wrong number of arguments");

    if (!mxIsNumeric(prhs[0]))
        mexErrMsgTxt("Source vector must be positive integers");
    to  = mxGetPr(prhs[0]);
    nto = mxGetM(prhs[0]) * mxGetN(prhs[0]);

    if (!mxIsNumeric(prhs[1]))
        mexErrMsgTxt("Predecessor vector must be numeric");
    p   = mxGetPr(prhs[1]);

    if (!mxIsSparse(prhs[2]))
        mexErrMsgTxt("Adjacency matrix must be sparse");

    if (!mxIsNumeric(prhs[3]) || mxGetM(prhs[3]) * mxGetN(prhs[3]) != 1)
        mexErrMsgTxt("Number of edges must be a scalar");
    ne  = (int) *mxGetPr(prhs[3]);
    if (ne <= 0)
        mexErrMsgTxt("Number of edges must be positive");

    /* Count nonzeros */ 
    nnz = 0;
    for (i = 0; i < nto; ++i) {
        int dest = to[i];
	while (p[dest-1] > 0) {
	    dest = p[dest-1];
	    ++nnz;
	}
    }

    /* Build path list */
    plhs[0] = mxCreateSparse(ne, nto, nnz, mxREAL);
    jc  = mxGetJc(plhs[0]);
    ir  = mxGetIr(plhs[0]);
    vpr = mxGetPr(plhs[0]);
    nnz = 0;

    if (nrhs == 5)
        node_used = mxGetPr(prhs[4]);
    else
	node_used = NULL;
    
    jc[0] = 0;
    for (i = 0; i < nto; ++i) {
        int dest = to[i];
	int plen = 0;
	while (p[dest-1] > 0) {
	    int e = get_edge_num(prhs[2], dest, p[dest-1]);

	    if (node_used) node_used[ dest-1 ] = 1;
	    dest = p[dest-1];
	    if (node_used) node_used[ dest-1 ] = 1;

	    vpr[nnz]  = 1;
	    ir[nnz]   = e-1;
	    nnz++;
	    plen++;
	}
	qsort(ir+(nnz-plen), plen, sizeof(int), int_compare);
	jc[i+1] = jc[i] + plen;
    }
}
