#include <mex.h>

/* 
 * function [e] = qrchoose(Vp, tol)
 *
 * Choose linearly independent columns from Vp using QR with column
 * pivoting.  Consider any columns for which the diagonal R entry is < tol
 * to be linearly dependent.  Return a vector e of indices to a linearly
 * independent set of vectors.
 *
 * This function overwrites the argument Vp.  Otherwise, it is
 * functionally equivalent to the following Matlab code:
 *
 * [q,r,e] = qr(Vp);
 * e       = e' * (1:length(e))';
 * e       = e(find( abs(diag(r)) > tol ));
 * V       = V(:,e);
 *
 */

/*
 * SUBROUTINE DGEQP3( M, N, A, LDA, JPVT, TAU,  WORK,  LWORK, INFO )
 *  -- QR with column pivoting.  Level 3 BLAS.
 */
void dgeqp3_(int* m, int* n, double* a, int* lda, int* jpvt, double* tau,
             double* work, int* lwork, int* info);

void mexFunction(int nlhs, mxArray** plhs,
                 int nrhs, const mxArray** prhs)
{
    int i, k, n, lwork, info;
    double twork;
    double tol;
    double* work;
    double* a;
    int*    jpvt;

    if (nrhs != 2)
        mxErrMsgTxt("Too few arguments");
    if (!mxIsNumeric(prhs[0]) || mxGetM(prhs[0]) != mxGetN(prhs[0]))
        mxErrMsgTxt("Argument must be a square matrix");
    if (!mxIsNumeric(prhs[1]) || mxGetM(prhs[1]) != mxGetN(prhs[1]))
        mxErrMsgTxt("Argument must be a scalar");

    n   = mxGetN(prhs[0]);
    a   = mxMalloc(n*n*sizeof(double));
    memcpy(a, mxGetPr(prhs[0]), n*n*sizeof(double));
    tol = *mxGetPr(prhs[1]);

    /* Workspace query and allocations */
    lwork = -1;
    dgeqp3_(&n, &n, NULL, &n, NULL, NULL, &twork, &lwork, &info);
    lwork = (int) twork;
    work  = mxMalloc((n + lwork) * sizeof(double));
    jpvt  = mxMalloc( n * sizeof(int) );
    memset(jpvt, 0, n*sizeof(int));

    /* Factor */
    dgeqp3_(&n, &n, a, &n, jpvt, work, work + n, &lwork, &info);

    /* Mark selected columns */
    k = 0;
    for (i = 0; i < n; ++i) {
        if (a[i*(n+1)] > tol || a[i*(n+1)] < -tol) {
            work[k++] = jpvt[i];
        }
    }

    /* Build return vector */
    plhs[0] = mxCreateDoubleMatrix(k,1, mxREAL);
    memcpy(mxGetPr(plhs[0]), work, k*sizeof(double));

    /* Free workspace */
    mxFree(a);
    mxFree(jpvt);
    mxFree(work);
}
