/* 
 * R = resizeR(R, len, precision, layout, blockR)
 *
 * Expand the triangular factor R as needed to hold len columns.
 * The precision can be 'single' or 'double'; the layout is 'full'
 * or 'packed' (actually a block packed format); and blockR is the 
 * blocking size used if R is packed.
 */

#include <mex.h>

void mexFunction(int nlhs, mxArray** plhs,
                 int nrhs, const mxArray** prhs)
{
    int is_double;  /* One for double precision, zero for float */
    int is_full;    /* One for full layout, zero for blocked    */
    int blockR;     /* Block size                               */
    int len;        /* Extension length                         */
    int space_need; /* Space needed                             */

    char stmp[15];  /* Temporary string buffer */


    /* -- Process arguments -- */

    if (nrhs != 5)
        mexErrMsgTxt("Wrong number of arguments");

    len       = (int) *mxGetPr(prhs[1]);
    blockR    = (int) *mxGetPr(prhs[4]);

    mxGetString(prhs[2], stmp, sizeof(stmp));
    is_double = ( strcmp(stmp, "single") != 0 );

    mxGetString(prhs[3], stmp, sizeof(stmp));
    is_full   = ( strcmp(stmp, "full")   == 0 );


    /* -- Exit if R is already sufficiently large -- */

    if (is_full) {

        if (len <= mxGetN(prhs[0])) {
            plhs[0] = (mxArray*) prhs[0];
            return;
        }

    } else {

        int block_len  = (len + blockR - 1) / blockR;
        int block_need = (block_len * (block_len + 1)) / 2;
        space_need = block_need * blockR;

        if (space_need <= mxGetN(prhs[0])) {
            plhs[0] = (mxArray*) prhs[0];
            return;
        }

    }


    /* -- If R too small, grow R and copy contents -- */

    if (is_full) {

        int old_m = mxGetM(prhs[0]);
        int m_2x  = old_m * 1.4;
        int new_m = (m_2x < len) ? len : m_2x;
        int esize = is_double ? sizeof(double) : sizeof(float);
        int etype = is_double ? mxDOUBLE_CLASS : mxSINGLE_CLASS;
        int i;

        plhs[0] = mxCreateNumericMatrix(new_m, new_m, etype, mxREAL);

        for (i = 0; i < old_m; ++i) {

            void* Ri_old = is_double ? 
                (void*) (          mxGetPr(prhs[0]) + i*old_m ) :
                (void*) ( (float*) mxGetPr(prhs[0]) + i*old_m );

            void* Ri_new = is_double ? 
                (void*) (          mxGetPr(plhs[0]) + i*new_m ) :
                (void*) ( (float*) mxGetPr(plhs[0]) + i*new_m );

            memcpy(Ri_new, Ri_old, (i+1) * esize);

        }

    } else {

        int old_n = mxGetN(prhs[0]);
        int n_2x  = old_n * 2;
        int new_n = (n_2x < space_need) ? space_need : n_2x;
        int esize = is_double ? sizeof(double) : sizeof(float);
        int etype = is_double ? mxDOUBLE_CLASS : mxSINGLE_CLASS;

        plhs[0] = mxCreateNumericMatrix(blockR, space_need, etype, mxREAL);
        memcpy(mxGetPr(plhs[0]), mxGetPr(prhs[0]), 
               mxGetM(prhs[0]) * mxGetN(prhs[0]) * esize);

    }
}
