% [I, R] = rowselect(G, param)
%
% Selects a set of rows which span the row space of G.  The resulting
% reduced matrix is Gbar = G(I,:).  This routine also produces an
% upper triangular factor R which can be used to compute the minimal
% norm solution to systems of the form Gbar * x = bbar.
%
% The parameters governing the row selection are passed in the
% param structure.  Options which are not specified explicitly have
% a default value.
%
% Parameters:
%
%   precision (default 'double'): can be 'single' or 'double' to
%     determine the precision used for storage and computation of
%     R.
%
%   layout (default 'full'): storage layout for the R factor.  Can
%     be 'full' or 'packed'.  The 'full' format is an ordinary
%     column major layout; the 'packed' format is a blocked
%     column-major packed layout.  If kb is the block size, then
%     kb elements are stored from each of columns 1:kb; 2*kb
%     elements are stored from columns 2*kb+1:3*kb; and so forth.
%
%   tol (default 1e-5): tolerance used for determining linear
%     dependence of rows.  If the component of a row v normal to
%     the space spanned by the rows already examined is <
%     tol*norm(v), then v is considered linearly dependent on the
%     previous rows.  Otherwise, we consider it a linearly
%     independent direction.
%
%   verbose (default 1): zero for minimal printed output, one for
%     more printed output.
%
%   blockG (default 100): number of rows of G to be examined
%     simultaneously.  When blockG is large, more work is done in
%     fast level 3 BLAS operations, but more scratch memory is used
%     as well.
%
%   blockR (default 100): blocking factor used in the packed layout
%     of R (see the description of the layout parameter).

function [I, R] = rowselect(G, param)

if  nargin < 2,                   param = [];                  end
if ~isfield(param, 'precision'),  param.precision = 'double';  end
if ~isfield(param, 'layout'),     param.layout    = 'full';    end
if ~isfield(param, 'tol'),        param.tol       = 1e-5;      end
if ~isfield(param, 'verbose'),    param.verbose   = 1;         end
if ~isfield(param, 'blockG'),     param.blockG    = 100;       end
if ~isfield(param, 'blockR'),     param.blockR    = 100;       end

V  = G';
Vb = [];
I  = [];
R  = [];
m  = size(G,1);
n  = size(G,2);
mblocks = ceil(m / param.blockG);

R = [];

for i = 1:mblocks

  % --- Get a block of rows of G
  k    = length(I);
  J    = ( (i-1)*param.blockG+1 : min(i*param.blockG, m) );
  Vi   = V(:,J);
  
  % --- Resize R as needed and choose l.i. rows of G block
  R    = resizeR(R, max(k+length(J), 1000), ...
		 param.precision, param.layout, param.blockR);
  Jsub = addpaths(Vb, Vi, R, param.tol, param.layout, ...
		  param.blockR);

  % --- Diagnostic output
  if param.verbose
    fprintf('\rProcessed block %d / %d; k = %d    ', i, mblocks, ...
	    k+length(Jsub));
  end

  % --- Update list of rows kept
  I    = [I, J(Jsub)];
  Vb   = [Vb, Vi(:,Jsub)];

end

if param.verbose, fprintf('\n'); end
