% [V, R, Vinfo, k, percent, tall] = 
%   scanadd(nold, nnew, V, R, Vinfo, adj, A, ne, verb);
%
% Try to compute an orthonormal basis V for the range space of
% all pairs of shortest-path vectors between the nodes in n.
% Return also the rank k of the generated basis.
%
% Inputs:
%   n     - list of nodes used
%   V     - path list so far
%   R     - associated normalization matrix
%   Vinfo - path info so far
%   adj  - adjacency matrix of edge weights
%   A    - adjacency matrix of edge indices
%   ne   - number of edges (compute from A if not provided)
%   verb - verbosity flag (default is 1, verbose)
%
% Outputs:
%   V       - updated reduced path matrix 
%   Vinfo   - updated [source; destination; distance] for columns of V
%   k       - rank (number of columns in V)
%   percent - percentage of total paths kept
%   tall    - total time

function [V, R, Vinfo, k, percent, tall] = ...
  scanaddf(nold, nnew, V, R, Vinfo, adj, A, ne, verb);

if nargin < 9, verb = 1; end

mold  = length(nold);
mnew  = length(nnew);
nconsidered = mold*(mold-1)/2;

ki     = 0;
Vi     = [];
Vinfoi = [];

if verb, fprintf('i:\tk\t%%\tt_step\tt_all\n'); end

t0 = clock;
for i = 1:mnew-1
  t1 = clock;

  % Compute paths to extend basis
  [d, p] = dijkstra(adj, nnew(i));

  % Add paths to old nodes
  if mold > 0
    Vi = [Vi, pathv(nold, p, A, ne)];
    for j = 1:mold
      Vinfoi = [Vinfoi, [nnew(i); nold(j); d(nold(j))]];   
    end
  end

  % Add paths to new nodes
  Vi = [Vi, pathv(nnew(i+1:mnew), p, A, ne)];
  for j = i+1:mnew
    Vinfoi = [Vinfoi, [nnew(i); nnew(j); d(nnew(j))]];
  end

  nconsidered = nconsidered + mold + mnew-i;

  % Compute index vector e of new independent paths:
  %  - Gram-Schmidt to eliminate V directions from Vi
  %  - Rank revealing QR on remainder
  %
  if size(Vi,2) > 200 | i == mnew-1 

    % Update path record
    if isa(R, 'single')
      e = addpaths(V, R, Vi, 1e-3);
    else
      e = addpathsd(V, R, Vi, 1e-3);
    end
    V = [V, Vi(:,e)];

    Vinfo   = [ Vinfo, Vinfoi(:,e) ];
    Vi      = [];
    Vinfoi  = [];

  end

  % Print current info
  if verb
    fprintf('\r                                           ');
    fprintf('\r%d:\t%d\t%3.2f\t%3.2f\t%g', i, size(V,2), ...
            100*size(V,2) / nconsidered, ...
            etime(clock,t1), etime(clock,t0));
  end

end

% Compute statistics
percent = 100*size(V,2) / nconsidered;
k       = size(V,2);
tall    = etime(clock,t0);

if verb, fprintf('\n'); end
