/*
 * function [x] = trsolve(R, b, trans, blockR)
 *
 * Solve the system R*x = b, where R is upper triangular.
 * Trans is 'N' for Rx = b, 'T' for R^T x = b, and blockR
 * is the block size if a block packed format is used,
 * and zero otherwise.
 */

#include <mex.h>

/*
 * SUBROUTINE TRTRS( UPLO, TRANS, DIAG, N, NRHS, A, LDA,  B,
 *                   LDB, INFO )
 *
 * LAPACK routine -- perform a triangular solve.
 */
void dtrtrs_(char* uplo, char* trans, char* diag,
             int* n, int* nrhs,
             double* a, int* lda,
             double* b, int* ldb,
             int* info);

/*
 * SUBROUTINE GEMM ( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA,
 *                   B, LDB, BETA, C, LDC )
 *
 * BLAS routine -- perform a matrix-matrix multiplication.
 */
void dgemm_(char* transA, char* transB, 
            int* m, int* n, int* k, double* alpha,
            double* A, int* ldA,
            double* B, int* ldB, double* beta, 
            double* C, int* ldC);

/* 
 * Blocked backsolve:
 */
static void block_trs(double* R, int k, int kb, double* b, int m, char trans)
{
    int n, J, info;
    double one     = 1;
    double neg_one = -1;

    if (trans != 'T' && trans != 't') {

        /* Ordinary blocked backsolve */
        n = k;
        for (J = (k-1)/kb; J >= 0; --J) {
            
            int Jkb    = J*kb;
            int Jkbp   = Jkb + kb;
            int blockn = n-Jkb;
        
            dtrtrs_("U", "N", "N", &blockn, &m,
                    R    + Jkb + (Jkb*Jkbp)/2,  &Jkbp,
                    b    + Jkb,                 &k,
                    &info);
        
            dgemm_("N", "N", &Jkb, &m, &blockn,
                   &neg_one,
                   R    + (Jkb*Jkbp)/2,  &Jkbp,
                   b    + Jkb,           &k,
                   &one,
                   b,   &k);

            n = Jkb;
            
        }

    } else {

        /* Transposed blocked backsolve */
        for (J = 0; J <= (k-1)/kb; ++J) {
            
            int Jkb    = J*kb;
            int Jkbp   = Jkb + kb;
            int blockn = (Jkbp < k) ? kb : (k - Jkb);

            dgemm_("T", "N", &blockn, &m, &Jkb,
                   &neg_one,
                   R    + (Jkb*Jkbp)/2,  &Jkbp,
                   b,                    &k,
                   &one,
                   b    + Jkb,           &k);
        
            dtrtrs_("U", "T", "N", &blockn, &m,
                    R    + Jkb + (Jkb*Jkbp)/2,  &Jkbp,
                    b    + Jkb,                 &k,
                    &info);

        }

    }
}

void mextrsolved(int nlhs, mxArray** plhs,
                 int nrhs, const mxArray** prhs)
{
    double* R;
    double* b;
    double* x;
    int    ldR;
    int    k;
    int    kb;
    int    info;
    int    m = 1;
    char   trans[15];

    if (nrhs != 4)
        mexErrMsgTxt("Wrong number of arguments");

    R    = (double*) mxGetPr(prhs[0]);
    b    = (double*) mxGetPr(prhs[1]);
    x    = (double*) mxGetPr(plhs[0]);
    mxGetString(prhs[2], trans, sizeof(trans));

    ldR  = mxGetM(prhs[0]);
    k    = mxGetM(prhs[1]);
    kb   = (int) *mxGetPr(prhs[3]);

    memcpy(x, b, k * sizeof(double));
    if (kb)  block_trs(R, k, kb, x, m, *trans);
    else     dtrtrs_("U", trans, "N", &k, &m, R, &ldR, x, &k, &info);
}
